/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* PURPOSE: Common include file for ZigBee
*/
#ifndef ZB_COMMON_H
#define ZB_COMMON_H 1

#include "zboss_api.h"
#include "zb_config.h"
#include "zb_types.h"
#include "zb_errors.h"
#include "zb_debug.h"
#include "zb_g_context.h"
#include "zb_trace.h"
#include "zb_hash.h"
#ifdef ZB_TH_ENABLED
#include "zb_th_internal.h"
#else
#include "zb_th_dummy.h"
#endif /* ZB_TH_ENABLED */
#include "zb_common_u.h"

#ifdef ZB_USE_NVRAM
#include "zb_nvram.h"
#endif /* ZB_USE_NVRAM */

/*! \addtogroup init_api */
/*! @{ */

void zb_globals_init(void);

//TODO Document function. Obsolete?
void zb_usbc_serial_init(void);

//TODO Document function
int usbc_has_rx_data(void);

//TODO Document function
zb_uint16_t usbc_serial_data_tx(zb_uint8_t* buf, zb_uint32_t len);

/**
   Receive data via UART over USB.

   This routine is to be executed _not_ from the interrupt connext.
   This routine does not wait for a data.

   @param Buf - data buffer
   @param Len - buffer size

   @return number of bytes copied to Buf. 0 if no data copied.
 */
zb_uint_t usbc_serial_data_rx (zb_uint8_t* Buf, zb_uint32_t Len);

#if defined ENABLE_USB_SERIAL_IMITATOR
#define USB_SERIAL_IMITATOR_RX_PIPE "usbc_rx"
#define USB_SERIAL_IMITATOR_TX_PIPE "usbc_tx"
#endif /* defined ENABLE_USB_SERIAL_IMITATOR */

#define ZB_RESET_NETWORK_SETTINGS() zb_reset_network_settings()

/*! @} */

/** @cond internals_doc */
zb_ret_t zb_schedule_callback(zb_callback_t func, zb_uint8_t param,
                              zb_bool_t use_2_param, zb_uint16_t user_param, zb_bool_t is_prior);

/**
   Schedule single-param callback execution.

   Schedule execution of function `func' in the main scheduler loop.

   @param func - function to execute
   @param param - callback parameter - usually, but not always ref to packet buffer

   @return RET_OK or RET_OVERFLOW.

   See sched sample
 */
#define ZB_SCHEDULE_CALLBACK(func, param) zb_schedule_callback(func, param, ZB_FALSE, 0, ZB_FALSE)

/**
   Schedule two-param callback execution.
   Schedule execution of function `func' in the main scheduler loop.

   @param func - function to execute
   @param param - zb_uint8_t callback parameter - usually, but not always ref to
   packet buffer
   @param user_param - zb_uint16_t user parameter - usually, but not
   always short address

   @return RET_OK or RET_OVERFLOW.
   See sched sample
 */
#define ZB_SCHEDULE_CALLBACK2(func, param, user_param) zb_schedule_callback((zb_callback_t)(func),  param, ZB_TRUE, user_param, ZB_FALSE)

zb_ret_t zb_schedule_alarm(zb_callback_t func, zb_uint8_t param, zb_time_t timeout_bi);

/**
   Schedule alarm - callback to be executed after timeout.

   Function will be called via scheduler after timeout expired (maybe, plus some
   additional time).
   Timer resolution depends on implementation.
   Same callback can be scheduled for execution more then once.

   @param func - function to call via scheduler
   @param param - parameter to pass to the function
   @param timeout_bi - timeout, in beacon intervals
   @return RET_OK or RET_OVERFLOW

   See any sample
 */
#define ZB_SCHEDULE_ALARM(func, param, timeout_bi) zb_schedule_alarm(func, param, timeout_bi)

/**
   Cancel scheduled alarm.

   This function cancel previously scheduled alarm. Function is identified by
   the pointer.

   @param func - function to cancel
   @param param - parameter to cancel. \see ZB_ALARM_ANY_PARAM. \see ZB_ALARM_ALL_CB
   @return RET_OK or error code

   See reporting_srv sample
 */
#define ZB_SCHEDULE_ALARM_CANCEL ZB_SCHEDULE_APP_ALARM_CANCEL

/**
   Cancel scheduled alarm and get buffer.

   This function cancel previously scheduled alarm and returns buffer ref associated with alarm.
   Function is identified by the pointer.

   @param func - function to cancel
   @param param - parameter to cancel. \see ZB_ALARM_ANY_PARAM. \see ZB_ALARM_ALL_CB
   @param p_param - [out] pointer of ref buffer from cancelled flag: free buffer if its alarm will be cancel
   @return RET_OK or error code

   @b Example:
   @code
   {
     zb_uint8_t cancelled_param;

     ZB_SCHEDULE_ALARM_CANCEL_AND_GET_BUF(my_func1, ZB_ALARM_ANY_PARAM, &cancelled_param);
     my_func1(cancelled_param);
   }
   @endcode

   See reporting_srv sample
 */
#define ZB_SCHEDULE_APP_ALARM_CANCEL_AND_GET_BUF(func, param, p_param) zb_schedule_alarm_cancel((func), (param), p_param)

/**
   Cancel scheduled alarm and get buffer.

   This function cancel previously scheduled alarm and returns buffer ref associated with alarm.
   Function is identified by the pointer.

   @param func - function to cancel
   @param param - parameter to cancel. \see ZB_ALARM_ANY_PARAM. \see ZB_ALARM_ALL_CB
   @param p_param - [out] pointer of ref buffer from cancelled flag: free buffer if its alarm will be cancel
   @return RET_OK or error code

   @b Example:
   @code
   {
     zb_uint8_t cancelled_param;

     ZB_SCHEDULE_ALARM_CANCEL_AND_GET_BUF(my_func1, ZB_ALARM_ANY_PARAM, &cancelled_param);
     my_func1(cancelled_param);
   }
   @endcode

   See reporting_srv sample
 */
#define ZB_SCHEDULE_ALARM_CANCEL_AND_GET_BUF ZB_SCHEDULE_APP_ALARM_CANCEL_AND_GET_BUF

/** @endcond */ /* internals_doc */

/*! \internal \addtogroup ZB_BASE */
/*! @{ */

/**
   Set Informational Bases refaults.

   @param rx_pipe - rx pipe name (for Unix) or node number (for ns build
                            in 8051 simulator)
 */
void zb_ib_set_defaults(zb_char_t *rx_pipe);

/**
   Resets network settings in ZG->nwk, APS
 */
void zb_reset_network_settings(void);

#ifdef ZB_PRODUCTION_CONFIG

/**
  Calculates crc from production configuration block using

  VP: it is not zb_crc32().

  @param prod_cfg - pointer to configuration block
  @param prod_cfg_size - block size

  @return crc32
 */
zb_uint32_t zb_calculate_production_config_crc(zb_uint8_t *prod_cfg, zb_uint8_t prod_cfg_size);
#endif /* ZB_PRODUCTION_CONFIG */

enum zb_dev_specific_e
{
  /*! Standard device */
  ZB_STANDARD_SPECIFIC_DEVICE = 0x00,
  /*! Manufactured device */
  ZB_MANUFACTURED_SPECIFIC_DEVICE = 0x01
};


/* Definitions for multiple tests in the same FW image / same Linux binary */
#ifdef ZB_MULTI_TEST
extern zb_int_t g_argc;
extern char** g_argv;
#define ZB_ARGV g_argv
#define ZB_ARGC g_argc
#undef MAIN
#undef ARGV_UNUSED
#define ARGV_UNUSED
#undef MAIN_RETURN
#define MAIN_RETURN(x)
#define TN_CAT0(x,y) x ## y
#define TN_CAT(x,y) TN_CAT0(x,y)
#define MAIN void TN_CAT(ZB_TEST_NAME,_main)
#define ZB_ZDO_STARTUP_COMPLETE void TN_CAT(ZB_TEST_NAME,_zb_zdo_startup_complete)
#define ZGPD_STARTUP_COMPLETE void TN_CAT(ZB_TEST_NAME,_zgpd_startup_complete)

/* List of MAC primitives' callbacks */
#define ZB_MLME_ASSOCIATE_INDICATION void TN_CAT(ZB_TEST_NAME,_zb_mlme_associate_indication)
#define ZB_MLME_ASSOCIATE_CONFIRM void TN_CAT(ZB_TEST_NAME,_zb_mlme_associate_confirm)
#define ZB_MLME_BEACON_NOTIFY_INDICATION void TN_CAT(ZB_TEST_NAME,_zb_mlme_beacon_notify_indication)
#define ZB_MLME_COMM_STATUS_INDICATION void TN_CAT(ZB_TEST_NAME,_zb_mlme_comm_status_indication)
#define ZB_MLME_ORPHAN_INDICATION void TN_CAT(ZB_TEST_NAME,_zb_mlme_orphan_indication)
#define ZB_MLME_RESET_CONFIRM void TN_CAT(ZB_TEST_NAME,_zb_mlme_reset_confirm)
#define ZB_MLME_SCAN_CONFIRM void TN_CAT(ZB_TEST_NAME,_zb_mlme_scan_confirm)
#define ZB_MLME_START_CONFIRM void TN_CAT(ZB_TEST_NAME,_zb_mlme_start_confirm)
#define ZB_MLME_POLL_CONFIRM void TN_CAT(ZB_TEST_NAME,_zb_mlme_poll_confirm)
#define ZB_MLME_PURGE_CONFIRM void TN_CAT(ZB_TEST_NAME,_zb_mlme_purge_confirm)
#define ZB_MCPS_DATA_INDICATION void TN_CAT(ZB_TEST_NAME,_zb_mcps_data_indication)
#define ZB_MCPS_DATA_CONFIRM void TN_CAT(ZB_TEST_NAME,_zb_mcps_data_confirm)
#define ZB_MLME_DUTY_CYCLE_MODE_INDICATION void TN_CAT(ZB_TEST_NAME,_zb_mlme_duty_cycle_mode_indication)

#else  /* ZB_MULTI_TEST */

#define ZB_ZDO_STARTUP_COMPLETE void zb_zdo_startup_complete
#define ZGPD_STARTUP_COMPLETE void zgpd_startup_complete

  /* List of MAC primitives' callbacks */
#define ZB_MLME_ASSOCIATE_INDICATION void zb_mlme_associate_indication
#define ZB_MLME_ASSOCIATE_CONFIRM void zb_mlme_associate_confirm
#define ZB_MLME_BEACON_NOTIFY_INDICATION void zb_mlme_beacon_notify_indication
#define ZB_MLME_COMM_STATUS_INDICATION void zb_mlme_comm_status_indication
#define ZB_MLME_ORPHAN_INDICATION void zb_mlme_orphan_indication
#define ZB_MLME_RESET_CONFIRM void zb_mlme_reset_confirm
#define ZB_MLME_SCAN_CONFIRM void zb_mlme_scan_confirm
#define ZB_MLME_START_CONFIRM void zb_mlme_start_confirm
#define ZB_MLME_POLL_CONFIRM void zb_mlme_poll_confirm
#define ZB_MLME_PURGE_CONFIRM void zb_mlme_purge_confirm
#define ZB_MCPS_DATA_INDICATION void zb_mcps_data_indication
#define ZB_MCPS_DATA_CONFIRM void zb_mcps_data_confirm
#define ZB_MLME_DUTY_CYCLE_MODE_INDICATION void zb_mlme_duty_cycle_mode_indication

#define ZB_ARGV argv
#define ZB_ARGC argc

#endif  /* ZB_MULTI_TEST */

#ifndef ZB_CONFIGURABLE_MEM
#define ZB_RESYNC_CFG_MEM()
#else /* ZB_CONFIGURABLE_MEM */
void zb_init_configurable_mem(int clear);
#define ZB_RESYNC_CFG_MEM() zb_init_configurable_mem(0)
#endif /* ZB_CONFIGURABLE_MEM */

#ifdef ZB_INTERRUPT_SAFE_ALARMS
#define ZB_ALARM_INT_DISABLE() ZB_OSIF_GLOBAL_LOCK()
#define ZB_ALARM_INT_ENABLE() ZB_OSIF_GLOBAL_UNLOCK()
#else /* ZB_INTERRUPT_SAFE_ALARMS */
#define ZB_ALARM_INT_DISABLE()
#define ZB_ALARM_INT_ENABLE()
#endif /* ZB_INTERRUPT_SAFE_ALARMS */

#ifdef ZB_INTERRUPT_SAFE_CALLBACKS
#define ZB_CB_INT_DISABLE() ZB_OSIF_GLOBAL_LOCK()
#define ZB_CB_INT_ENABLE() ZB_OSIF_GLOBAL_UNLOCK()
#else /* ZB_INTERRUPT_SAFE_CALLBACKS */
#define ZB_CB_INT_DISABLE()
#define ZB_CB_INT_ENABLE()
#endif /* ZB_INTERRUPT_SAFE_CALLBACKS */

/* Note: disabling interrupts during buffers operations conflicts with trace from buffers subsystem!
   Must switch off trace from buffers.
   Normally we do not use from interrupt handlers buffer pool API which can touch global buffer pool data structures,
   so unlikely that feature is required.
 */
#ifdef ZB_INTERRUPT_SAFE_BUFS
#define ZB_BUF_INT_DISABLE() ZB_OSIF_GLOBAL_LOCK()
#define ZB_BUF_INT_ENABLE() ZB_OSIF_GLOBAL_UNLOCK()
#else /* ZB_INTERRUPT_SAFE_BUFS */
#define ZB_BUF_INT_DISABLE()
#define ZB_BUF_INT_ENABLE()
#endif /* ZB_INTERRUPT_SAFE_BUFS */

#if !defined ZB_LEGACY_BUFS
#ifdef zb_buf_t
#undef zb_buf_t
#undef ZB_BUF_ALLOC_LEFT
#undef ZB_BUF_ALLOC_RIGHT
#undef ZB_BUF_BEGIN_FROM_REF
#undef ZB_BUF_BEGIN
#undef ZB_BUF_CLEAR_PARAM
#undef ZB_BUF_COPY
#undef ZB_BUF_CUT_LEFT2
#undef ZB_BUF_CUT_LEFT
#undef ZB_BUF_CUT_RIGHT
#undef ZB_BUF_FROM_REF
#undef ZB_BUF_GET_ARRAY_PTR
#undef ZB_BUF_INITIAL_ALLOC
#undef ZB_BUF_LEN_BY_REF
#undef ZB_BUF_LEN
#undef ZB_BUF_OFFSET
#undef ZB_BUF_REUSE
#undef ZB_FREE_BUF
#undef ZB_FREE_BUF_BY_REF
#undef ZB_GET_BUF_PARAM
#undef ZB_GET_BUF_TAIL
#undef ZB_GET_IN_BUF
#undef ZB_GET_IN_BUF_DELAYED
#undef ZB_GET_OUT_BUF
#undef ZB_GET_OUT_BUF_DELAYED2
#undef ZB_GET_OUT_BUF_DELAYED
#undef ZB_REF_FROM_BUF
#undef ZB_REF_FROM_BUF_PTR
#undef ZB_SET_BUF_PARAM_PTR
#undef ZB_SET_BUF_PARAM
#undef ZB_SWITCH_BUF

#endif /*zb_buf_t */

#define ZB_BUF_ALLOC_LEFT(zbbuf, size, ptr)           No old buf API
#define ZB_BUF_ALLOC_RIGHT(zbbuf, size, ptr)          No old buf API
#define ZB_BUF_BEGIN_FROM_REF(_ref)                   No old buf API
#define ZB_BUF_BEGIN(zbbuf)                           No old buf API
#define ZB_BUF_CLEAR_PARAM(zbbuf)                     No old buf API
#define ZB_BUF_COPY(dst_buf, src_buf)                 No old buf API
#define ZB_BUF_CUT_LEFT2(zbbuf, size)                 No old buf API
#define ZB_BUF_CUT_LEFT(zbbuf, size, ptr)             No old buf API
#define ZB_BUF_CUT_RIGHT(zbbuf, size)                 No old buf API
#define ZB_BUF_FROM_REF(ref)                          No old buf API
#define ZB_BUF_GET_ARRAY_PTR(zbbuf)                   No old buf API
#define ZB_BUF_INITIAL_ALLOC(zbbuf, size, ptr)        No old buf API
#define ZB_BUF_LEN_BY_REF(param)                      No old buf API
#define ZB_BUF_LEN(zbbuf)                             No old buf API
#define ZB_BUF_OFFSET(zbbuf)                          No old buf API
#define ZB_BUF_REUSE(zbbuf)                           No old buf API
#define ZB_FREE_BUF(buf)                              No old buf API
#define ZB_FREE_BUF_BY_REF(param)                     No old buf API
#define ZB_GET_BUF_PARAM(zbbuf, type)                 No old buf API
#define ZB_GET_BUF_TAIL(zbbuf, size)                  No old buf API
#define ZB_GET_IN_BUF()                               No old buf API
#define ZB_GET_IN_BUF_DELAYED(callback)               No old buf API
#define ZB_GET_OUT_BUF()                              No old buf API
#define ZB_GET_OUT_BUF_DELAYED2(callback, user_param) No old buf API
#define ZB_GET_OUT_BUF_DELAYED(callback)              No old buf API
#define ZB_REF_FROM_BUF(buf)                          No old buf API
#define ZB_REF_FROM_BUF_PTR(ptr)                      No old buf API
#define ZB_SET_BUF_PARAM_PTR(zbbuf, param, type)      No old buf API
#define ZB_SET_BUF_PARAM(zbbuf, param, type)          No old buf API
#define ZB_SWITCH_BUF(buf, to_in)                     No old buf API
#endif /* !defined ZB_LEGACY_BUFS */

#endif /* ZB_COMMON_H */
