/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*  PURPOSE: NCP High level transport (adapters layer) implementation for the host side: ZDO category
*/
#define ZB_TRACE_FILE_ID 17511

#include "ncp_host_hl_transport_internal_api.h"
#include "ncp_host_soc_state.h"

#define ADAPTER_ZDO_TRAN_TABLE_SIZE ZDO_TRAN_TABLE_SIZE

typedef ZB_PACKED_PRE struct ncp_host_zdo_cb_ent_s
{
  zb_callback_t cb;
  zb_uint8_t    zdp_tsn; /* ZDP Transaction sequence number */
  zb_uint8_t    ncp_tsn; /* NCP Transaction sequence number */
  zb_bufid_t    buffer;  /* Request buffer to use in response handler */
} ZB_PACKED_STRUCT ncp_host_zdo_cb_ent_t;

typedef ZB_PACKED_PRE struct ncp_host_zdo_adapter_ctx_s
{
  ncp_host_zdo_cb_ent_t zdo_cb[ADAPTER_ZDO_TRAN_TABLE_SIZE];
  zb_uint8_t zdp_tsn; /* ZDP Transaction sequence number */
} ZB_PACKED_STRUCT ncp_host_zdo_adapter_ctx_t;

ncp_host_zdo_adapter_ctx_t g_ncp_host_zdo_adapter_ctx;

static void ncp_host_zdo_cb_reset(void)
{
  zb_uint8_t i;
  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_cb_reset", (FMT__0));

  for (i = 0; i < ADAPTER_ZDO_TRAN_TABLE_SIZE; i++)
  {
    g_ncp_host_zdo_adapter_ctx.zdo_cb[i].cb = NULL;
    g_ncp_host_zdo_adapter_ctx.zdo_cb[i].zdp_tsn = ZB_ZDO_INVALID_TSN;
    g_ncp_host_zdo_adapter_ctx.zdo_cb[i].ncp_tsn = ZB_ZDO_INVALID_TSN;
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_cb_reset", (FMT__0));
}

void ncp_host_zdo_adapter_init_ctx(void)
{
  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_adapter_init_ctx", (FMT__0));

  g_ncp_host_zdo_adapter_ctx.zdp_tsn = 0;
  ncp_host_zdo_cb_reset();

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_adapter_init_ctx", (FMT__0));
}

static void increment_zdp_tsn(void)
{
  g_ncp_host_zdo_adapter_ctx.zdp_tsn++;
  if (g_ncp_host_zdo_adapter_ctx.zdp_tsn == 0xFF)
  {
    g_ncp_host_zdo_adapter_ctx.zdp_tsn = 0;
  }
}

static void get_cb_ent_by_ncp_tsn(zb_uint8_t ncp_tsn, ncp_host_zdo_cb_ent_t **cb_ent)
{
  zb_uint8_t i;

  *cb_ent = NULL;

  for (i = 0; i < ADAPTER_ZDO_TRAN_TABLE_SIZE; i++)
  {
    if (g_ncp_host_zdo_adapter_ctx.zdo_cb[i].ncp_tsn == ncp_tsn)
    {
      *cb_ent = &g_ncp_host_zdo_adapter_ctx.zdo_cb[i];
    }
  }
}

static void ncp_host_zdo_adapter_register_cb(zb_callback_t cb, zb_bufid_t buffer, zb_uint8_t ncp_tsn, zb_uint8_t *zdp_tsn)
{
  zb_uint8_t i = 0;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_adapter_register_cb", (FMT__0));

  /* Even if there are no callback to register, return new zdp_tsn */
  *zdp_tsn = g_ncp_host_zdo_adapter_ctx.zdp_tsn;
  increment_zdp_tsn();

  if (cb != NULL)
  {
    for (i = 0; i < ADAPTER_ZDO_TRAN_TABLE_SIZE; i++)
    {
      if (g_ncp_host_zdo_adapter_ctx.zdo_cb[i].cb == NULL
          && g_ncp_host_zdo_adapter_ctx.zdo_cb[i].zdp_tsn == ZB_ZDO_INVALID_TSN
          && g_ncp_host_zdo_adapter_ctx.zdo_cb[i].ncp_tsn == ZB_ZDO_INVALID_TSN)
      {
        g_ncp_host_zdo_adapter_ctx.zdo_cb[i].cb = cb;
        g_ncp_host_zdo_adapter_ctx.zdo_cb[i].ncp_tsn = ncp_tsn;
        g_ncp_host_zdo_adapter_ctx.zdo_cb[i].zdp_tsn = *zdp_tsn;
        g_ncp_host_zdo_adapter_ctx.zdo_cb[i].buffer = buffer;

        TRACE_MSG(TRACE_TRANSPORT2, "set zdo_cb entry in ncp_host_zdo_adapter_ctx to %p by %hd index",
            (FMT__P_H, cb, i));

        break;
      }
    }
    if (i == ADAPTER_ZDO_TRAN_TABLE_SIZE)
    {
      TRACE_MSG(TRACE_ERROR, "No free space for ZDO cb in ZDO adapters, zdp_tsn 0x%hx, ncp_tsn 0x%hx",
                (FMT__H_H, zdp_tsn, ncp_tsn));
      ZB_ASSERT(0);
    }
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_adapter_register_cb", (FMT__0));
}


static void ncp_host_zdo_adapter_clean_cb_ent(ncp_host_zdo_cb_ent_t *cb_ent)
{
  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_adapter_clean_cb_ent", (FMT__0));

  if (cb_ent != NULL)
  {
    cb_ent->cb = NULL;
    cb_ent->zdp_tsn = ZB_ZDO_INVALID_TSN;
    cb_ent->ncp_tsn = ZB_ZDO_INVALID_TSN;
    cb_ent->buffer = ZB_BUF_INVALID;
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_adapter_clean_cb_ent", (FMT__0));
}

static void ncp_host_zdo_adapter_call_cb_for_blocking_req(zb_bufid_t buf, ncp_host_zdo_cb_ent_t *cb_ent)
{
  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_adapter_call_cb_by_for_blocking_req", (FMT__0));

  if (cb_ent != NULL)
  {
    if (cb_ent->cb)
    {
      cb_ent->cb(buf);
    }
    ncp_host_zdo_adapter_clean_cb_ent(cb_ent);
  }
  else
  {
    zb_buf_free(buf);
  }

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_adapter_call_cb_by_for_blocking_req", (FMT__0));
}

static void ncp_host_zdo_adapter_call_cb_for_nonblocking_req(zb_bufid_t buf, ncp_host_zdo_cb_ent_t *cb_ent)
{
  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_adapter_call_cb_for_nonblocking_req", (FMT__0));

  if (cb_ent != NULL)
  {
    cb_ent->cb(buf);
  }
  else
  {
    zb_buf_free(buf);
  }

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_adapter_call_cb_for_nonblocking_req", (FMT__0));
}

zb_uint8_t zb_zdo_ieee_addr_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_ieee_addr_req_param_t *req_param = ZB_BUF_GET_PARAM(param, zb_zdo_ieee_addr_req_param_t);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_ieee_addr_req", (FMT__0));

  ret = ncp_host_zdo_ieee_addr_request(&ncp_tsn, req_param);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_ieee_addr_req", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zdo_mgmt_leave_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_mgmt_leave_param_t *req_param = ZB_BUF_GET_PARAM(param, zb_zdo_mgmt_leave_param_t);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zdo_mgmt_leave_req", (FMT__0));

  ret = ncp_host_zdo_mgmt_leave_request(&ncp_tsn, req_param);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zdo_mgmt_leave_req", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t ncp_host_zdo_nwk_addr_request_adapter(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_nwk_addr_req_param_t *req_param = ZB_BUF_GET_PARAM(param, zb_zdo_nwk_addr_req_param_t);

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_nwk_addr_request_adapter", (FMT__0));

  ret = ncp_host_zdo_nwk_addr_request(&ncp_tsn, req_param);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_nwk_addr_request_adapter", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zb_zdo_power_desc_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_power_desc_req_t *req = (zb_zdo_power_desc_req_t *)zb_buf_begin(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_power_desc_req", (FMT__0));

  ret = ncp_host_zdo_power_desc_request(&ncp_tsn, req->nwk_addr);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_power_desc_req", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zb_zdo_node_desc_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_node_desc_req_t *req = (zb_zdo_node_desc_req_t *)zb_buf_begin(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_node_desc_req_adapter", (FMT__0));

  ret = ncp_host_zdo_node_desc_request(&ncp_tsn, req->nwk_addr);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_node_desc_req_adapter", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zb_zdo_simple_desc_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_simple_desc_req_t *req = (zb_zdo_simple_desc_req_t *)zb_buf_begin(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_simple_desc_req", (FMT__0));

  ret = ncp_host_zdo_simple_desc_request(&ncp_tsn, req->nwk_addr, req->endpoint);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_simple_desc_req", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zb_zdo_active_ep_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_active_ep_req_t *req = (zb_zdo_active_ep_req_t *)zb_buf_begin(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_active_ep_req", (FMT__0));

  ret = ncp_host_zdo_active_ep_request(&ncp_tsn, req->nwk_addr);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_active_ep_req", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zb_zdo_match_desc_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_match_desc_param_t *req = (zb_zdo_match_desc_param_t *)zb_buf_begin(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_match_desc_req", (FMT__0));

  ret = ncp_host_zdo_match_desc_request(&ncp_tsn, req);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_match_desc_req", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zb_zdo_bind_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_bind_req_param_t *bind_param = ZB_BUF_GET_PARAM(param, zb_zdo_bind_req_param_t);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_bind_req", (FMT__0));

  ret = ncp_host_zdo_bind_unbind_request(&ncp_tsn, bind_param, ZB_TRUE);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_bind_req", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zb_zdo_unbind_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_bind_req_param_t *bind_param = ZB_BUF_GET_PARAM(param, zb_zdo_bind_req_param_t);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_unbind_req", (FMT__0));

  ret = ncp_host_zdo_bind_unbind_request(&ncp_tsn, bind_param, ZB_FALSE);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_unbind_req", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zb_zdo_mgmt_permit_joining_req(zb_bufid_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_mgmt_permit_joining_req_param_t *req_param
      = ZB_BUF_GET_PARAM(param, zb_zdo_mgmt_permit_joining_req_param_t);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_mgmt_permit_joining_req", (FMT__0));

  ret = ncp_host_zdo_permit_joining_request(&ncp_tsn, req_param->dest_addr,
    req_param->permit_duration, req_param->tc_significance);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_mgmt_permit_joining_req", (FMT__0));

  return zdp_tsn;
}

zb_uint8_t zb_zdo_system_server_discovery_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;
  zb_zdo_system_server_discovery_param_t *req_param =
      ZB_BUF_GET_PARAM(param, zb_zdo_system_server_discovery_param_t);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_system_server_discovery_req", (FMT__0));

  ret = ncp_host_zdo_system_server_discovery_request(&ncp_tsn, req_param->server_mask);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, ZB_BUF_INVALID, ncp_tsn, &zdp_tsn);
  zb_buf_free(param);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_system_server_discovery_req", (FMT__0));

  return zdp_tsn;
}

void ncp_host_handle_zdo_ieee_addr_response(zb_ret_t status, zb_uint8_t ncp_tsn,
                                            zb_zdo_ieee_addr_resp_t *resp,
                                            zb_bool_t is_extended,
                                            zb_zdo_ieee_addr_resp_ext_t *resp_ext,
                                            zb_zdo_ieee_addr_resp_ext2_t *resp_ext2,
                                            zb_uint16_t *addr_list)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_uint8_t *ptr;
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_ieee_addr_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);

  ZB_ASSERT(cb_ent != NULL);
  resp->tsn = cb_ent->zdp_tsn;

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->status = ERROR_GET_CODE(status);
      ptr = zb_buf_initial_alloc(buf, sizeof(zb_zdo_nwk_addr_resp_head_t));
      ZB_MEMCPY(ptr, resp, sizeof(zb_zdo_nwk_addr_resp_head_t));
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->status = ZB_ZDP_STATUS_SUCCESS;
    ptr = zb_buf_initial_alloc(buf, sizeof(zb_zdo_ieee_addr_resp_t));
    ZB_MEMCPY(ptr, resp, sizeof(zb_zdo_ieee_addr_resp_t));

    if (is_extended)
    {
      ptr = zb_buf_alloc_right(buf, sizeof(zb_zdo_ieee_addr_resp_ext_t)
                               + sizeof(zb_zdo_ieee_addr_resp_ext2_t)
                               + sizeof(zb_uint16_t) * resp_ext->num_assoc_dev);

      ZB_MEMCPY(ptr, resp_ext, sizeof(zb_zdo_ieee_addr_resp_ext_t));
      ptr = ptr + sizeof(zb_zdo_ieee_addr_resp_ext_t);
      ZB_MEMCPY(ptr, resp_ext2, sizeof(zb_zdo_ieee_addr_resp_ext2_t));
      ptr = ptr + sizeof(zb_zdo_ieee_addr_resp_ext2_t);
      ZB_MEMCPY(ptr, addr_list, sizeof(zb_uint16_t) * resp_ext->num_assoc_dev);
    }
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_ieee_addr_response ", (FMT__0));
}

void ncp_host_handle_zdo_nwk_addr_response(zb_ret_t status, zb_uint8_t ncp_tsn,
                                           zb_zdo_nwk_addr_resp_head_t *resp,
                                           zb_bool_t is_extended,
                                           zb_zdo_nwk_addr_resp_ext_t *resp_ext,
                                           zb_zdo_nwk_addr_resp_ext2_t *resp_ext2,
                                           zb_uint16_t *addr_list)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_uint8_t *ptr;
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_nwk_addr_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->status = ERROR_GET_CODE(status);
      ptr = zb_buf_initial_alloc(buf, sizeof(zb_zdo_nwk_addr_resp_head_t));
      ZB_MEMCPY(ptr, resp, sizeof(zb_zdo_nwk_addr_resp_head_t));
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->status = ZB_ZDP_STATUS_SUCCESS;
    ptr = zb_buf_initial_alloc(buf, sizeof(zb_zdo_nwk_addr_resp_head_t));
    ZB_MEMCPY(ptr, resp, sizeof(zb_zdo_nwk_addr_resp_head_t));

    if (is_extended)
    {
      ptr = zb_buf_alloc_right(buf, sizeof(zb_zdo_nwk_addr_resp_ext_t)
                               + sizeof(zb_zdo_nwk_addr_resp_ext2_t)
                               + sizeof(zb_uint16_t) * resp_ext->num_assoc_dev);

      ZB_MEMCPY(ptr, resp_ext, sizeof(zb_zdo_nwk_addr_resp_ext_t));
      ptr = ptr + sizeof(zb_zdo_nwk_addr_resp_ext_t);
      ZB_MEMCPY(ptr, resp_ext2, sizeof(zb_zdo_nwk_addr_resp_ext2_t));
      ptr = ptr + sizeof(zb_zdo_nwk_addr_resp_ext2_t);
      ZB_MEMCPY(ptr, addr_list, sizeof(zb_uint16_t) * resp_ext->num_assoc_dev);
    }
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_nwk_addr_response ", (FMT__0));
}

void ncp_host_handle_zdo_power_descriptor_response(zb_ret_t status, zb_uint8_t ncp_tsn,
                                                   zb_uint16_t nwk_addr, zb_uint16_t power_descriptor)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_zdo_power_desc_resp_t *resp = (zb_zdo_power_desc_resp_t *)zb_buf_begin(buf);
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_power_descriptor_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->hdr.tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->hdr.status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->hdr.status = ZB_ZDP_STATUS_SUCCESS;
    resp->hdr.nwk_addr = nwk_addr;
    resp->power_desc.power_desc_flags = power_descriptor;
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_power_descriptor_response ", (FMT__0));
}

void ncp_host_handle_zdo_node_descriptor_response(zb_ret_t status, zb_uint8_t ncp_tsn,
                                                  zb_uint16_t nwk_addr, zb_af_node_desc_t *node_desc)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_zdo_node_desc_resp_t *resp = (zb_zdo_node_desc_resp_t *)zb_buf_begin(buf);
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_node_descriptor_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->hdr.tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->hdr.status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->hdr.status = ZB_ZDP_STATUS_SUCCESS;
    resp->hdr.nwk_addr = nwk_addr;
    ZB_MEMCPY(&resp->node_desc, node_desc, sizeof(zb_af_node_desc_t));
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_node_descriptor_response ", (FMT__0));
}

void ncp_host_handle_zdo_simple_descriptor_response(zb_ret_t status, zb_uint8_t ncp_tsn,
                                                    zb_uint16_t nwk_addr,
                                                    zb_af_simple_desc_1_1_t *simple_desc,
                                                    zb_uint16_t *app_cluster_list_ptr)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_zdo_simple_desc_resp_t *resp = (zb_zdo_simple_desc_resp_t *)zb_buf_begin(buf);
  zb_uint8_t i;
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_simple_descriptor_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->hdr.tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->hdr.status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->hdr.status = ZB_ZDP_STATUS_SUCCESS;
    resp->hdr.length = sizeof(zb_af_simple_desc_1_1_t) + sizeof(zb_uint16_t)
        * (resp->simple_desc.app_input_cluster_count + resp->simple_desc.app_output_cluster_count - 2);
    resp->hdr.nwk_addr = nwk_addr;

    ZB_MEMCPY(&resp->simple_desc, simple_desc, sizeof(zb_af_simple_desc_1_1_t));
    ZB_MEMCPY(&resp->simple_desc.app_cluster_list, app_cluster_list_ptr, sizeof(zb_uint16_t)
              * (resp->simple_desc.app_input_cluster_count + resp->simple_desc.app_output_cluster_count));

    TRACE_MSG(TRACE_TRANSPORT3, "simple_desc: input_app_cluster_list:", (FMT__0));
    for(i = 0; i < resp->simple_desc.app_input_cluster_count; i++)
    {
      TRACE_MSG(TRACE_TRANSPORT3, " 0x%x", (FMT__D, resp->simple_desc.app_cluster_list[i]));
    }

    TRACE_MSG(TRACE_TRANSPORT3, "simple_desc: output_app_cluster_list:", (FMT__0));
    for(; i < resp->simple_desc.app_output_cluster_count + resp->simple_desc.app_input_cluster_count; i++)
    {
      TRACE_MSG(TRACE_TRANSPORT3, " 0x%x", (FMT__D, resp->simple_desc.app_cluster_list[i]));
    }
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_simple_descriptor_response ", (FMT__0));
}

void ncp_host_handle_zdo_active_ep_response(zb_ret_t status, zb_uint8_t ncp_tsn, zb_uint16_t nwk_addr,
                                            zb_uint8_t ep_count, zb_uint8_t *ep_list)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_zdo_ep_resp_t *resp = (zb_zdo_ep_resp_t *)zb_buf_begin(buf);
  zb_uint8_t *ep_list_ptr = (zb_uint8_t *)(resp + 1);
  zb_uint8_t i;
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_active_ep_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->status = ZB_ZDP_STATUS_SUCCESS;
    resp->ep_count = ep_count;
    ZB_MEMCPY(ep_list_ptr, ep_list, sizeof(zb_uint8_t) * ep_count);

    resp->nwk_addr = nwk_addr;

    TRACE_MSG(TRACE_TRANSPORT3, "endpoint list: ", (FMT__0));
    for(i = 0; i < ep_count; i++)
    {
      TRACE_MSG(TRACE_TRANSPORT3, " 0x%hx", (FMT__H, ep_list_ptr[i]));
    }
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_active_ep_response ", (FMT__0));
}

void ncp_host_handle_zdo_match_desc_response(zb_ret_t status, zb_uint8_t ncp_tsn, zb_uint16_t nwk_addr,
                                             zb_uint8_t match_ep_count, zb_uint8_t *match_ep_list)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_zdo_match_desc_resp_t *resp = (zb_zdo_match_desc_resp_t *)zb_buf_begin(buf);
  zb_uint8_t *match_ep_list_ptr = (zb_uint8_t *)(resp + 1);
  zb_uint8_t i;
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_match_desc_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->status = ZB_ZDP_STATUS_SUCCESS;
    resp->match_len = match_ep_count;
    resp->nwk_addr = nwk_addr;

    ZB_MEMCPY(match_ep_list_ptr, match_ep_list, sizeof(zb_uint8_t) * match_ep_count);

    TRACE_MSG(TRACE_TRANSPORT3, "match endpoint list: ", (FMT__0));
    for(i = 0; i < match_ep_count; i++)
    {
      TRACE_MSG(TRACE_TRANSPORT3, " 0x%hx", (FMT__H, match_ep_list_ptr[i]));
    }
  }

  ncp_host_zdo_adapter_call_cb_for_nonblocking_req(buf, cb_ent);

  /* Upon receiving ZB_ZDP_STATUS_TIMEOUT remove match desc callback from the context*/
  if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO && ERROR_GET_CODE(status) == ZB_ZDP_STATUS_TIMEOUT)
  {
    TRACE_MSG(TRACE_TRANSPORT3, "received ZB_ZDP_STATUS_TIMEOUT from ncp, removes match desc cb", (FMT__0));
    ncp_host_zdo_adapter_clean_cb_ent(cb_ent);
  }

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_match_desc_response ", (FMT__0));
}

void ncp_host_handle_zdo_system_server_discovery_response(zb_ret_t status, zb_uint8_t ncp_tsn,
                                                          zb_uint16_t server_mask)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_zdo_system_server_discovery_resp_t *resp = (zb_zdo_system_server_discovery_resp_t *)zb_buf_begin(buf);
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_system_server_discovery_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->status = ZB_ZDP_STATUS_SUCCESS;
    resp->server_mask = server_mask;
  }

  ncp_host_zdo_adapter_call_cb_for_nonblocking_req(buf, cb_ent);

  /* Upon receiving ZB_ZDP_STATUS_TIMEOUT remove match desc callback from the context*/
  if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO && ERROR_GET_CODE(status) == ZB_ZDP_STATUS_TIMEOUT)
  {
    TRACE_MSG(TRACE_TRANSPORT3, "received ZB_ZDP_STATUS_TIMEOUT from ncp, removes system srv disc cb",
              (FMT__0));

    ncp_host_zdo_adapter_clean_cb_ent(cb_ent);
  }

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_system_server_discovery_response ", (FMT__0));
}

void ncp_host_handle_zdo_bind_response(zb_ret_t status, zb_uint8_t ncp_tsn)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_zdo_bind_resp_t *resp = (zb_zdo_bind_resp_t*)zb_buf_begin(buf);
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_bind_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->status = ZB_ZDP_STATUS_SUCCESS;
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_bind_response ", (FMT__0));
}

void ncp_host_handle_zdo_unbind_response(zb_ret_t status, zb_uint8_t ncp_tsn)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  zb_zdo_bind_resp_t *resp = (zb_zdo_bind_resp_t*)zb_buf_begin(buf);
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_unbind_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->status = ZB_ZDP_STATUS_SUCCESS;
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_unbind_response ", (FMT__0));
}

void ncp_host_handle_zdo_permit_joining_response(zb_ret_t status, zb_uint8_t ncp_tsn)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;
  zb_zdo_mgmt_permit_joining_resp_t *resp = (zb_zdo_mgmt_permit_joining_resp_t*)zb_buf_begin(buf);

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_permit_joining_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->status = ZB_ZDP_STATUS_SUCCESS;
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_permit_joining_response ", (FMT__0));
}


void ncp_host_handle_zdo_mgmt_leave_response(zb_ret_t status, zb_uint8_t ncp_tsn)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;
  zb_zdo_mgmt_leave_res_t *resp = (zb_zdo_mgmt_leave_res_t*)zb_buf_begin(buf);

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_mgmt_leave_response ", (FMT__0));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);
  if (cb_ent != NULL)
  {
    resp->tsn = cb_ent->zdp_tsn;
  }

  if (status != RET_OK)
  {
    /* Pass only ZDO errors to the application layer */
    if (ERROR_GET_CATEGORY(status) == ERROR_CATEGORY_ZDO)
    {
      resp->status = ERROR_GET_CODE(status);
    }
    else
    {
      TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
                (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
      ZB_ASSERT(0);
    }
  }
  else
  {
    resp->status = ZB_ZDP_STATUS_SUCCESS;
    ncp_host_state_set_joined(ZB_FALSE);
  }

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_mgmt_leave_response ", (FMT__0));
}


void ncp_host_handle_zdo_device_annce_indication(zb_uint16_t nwk_addr, zb_ieee_addr_t ieee_addr,
                                                 zb_uint8_t capability)
{
  zb_bufid_t param;
  zb_zdo_signal_device_annce_params_t *dev_annce_params;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_device_annce_indication ", (FMT__0));

  param = zb_buf_get(ZB_TRUE, 0);
  ZB_ASSERT(param != ZB_BUF_INVALID);

  dev_annce_params = (zb_zdo_signal_device_annce_params_t*)zb_app_signal_pack(param,
    ZB_ZDO_SIGNAL_DEVICE_ANNCE,
    RET_OK,
    sizeof(zb_zdo_signal_device_annce_params_t));

  dev_annce_params->device_short_addr = nwk_addr;
  ZB_IEEE_ADDR_COPY(dev_annce_params->ieee_addr, ieee_addr);
  dev_annce_params->capability = capability;

  ZB_SCHEDULE_CALLBACK(zboss_signal_handler, param);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_device_annce_indication", (FMT__0));
}

#ifdef ZB_JOIN_CLIENT

zb_ret_t zdo_initiate_rejoin(zb_bufid_t buf, zb_uint8_t *ext_pan_id,
                             zb_channel_page_t *channels_list, zb_bool_t secure_rejoin)
{
  zb_ret_t status = RET_ERROR;

  TRACE_MSG(TRACE_ZDO2, ">> zdo_initiate_rejoin", (FMT__0));

  status = ncp_host_zdo_rejoin_request(NULL, ext_pan_id, channels_list, secure_rejoin);
  ZB_ASSERT(status == RET_OK);

  zb_buf_free(buf);

  TRACE_MSG(TRACE_ZDO2, "<< zdo_initiate_rejoin", (FMT__0));

  return RET_OK;
}


void ncp_host_handle_zdo_rejoin_response(zb_ret_t status)
{
  zb_bufid_t buf = zb_buf_get(ZB_TRUE, 0);

  TRACE_MSG(TRACE_ZDO2, ">> ncp_host_handle_zdo_rejoin_response, status 0x%x", (FMT__D, status));

  ZB_ASSERT(buf != ZB_BUF_INVALID);

  if (buf != ZB_BUF_INVALID)
  {
    if (status == RET_OK)
    {
      ncp_host_state_set_joined(ZB_TRUE);
      zdo_commissioning_handle_dev_annce_sent_event(buf);
    }
    else if (status == ERROR_CODE(ERROR_CATEGORY_ZDO, ZB_ZDP_STATUS_NOT_AUTHORIZED))
    {
      zdo_commissioning_authentication_failed(buf);
    }
    else
    {
      zdo_commissioning_join_failed(buf);
    }
  }

  TRACE_MSG(TRACE_ZDO2, "<< ncp_host_handle_zdo_rejoin_response", (FMT__0));
}

#endif /* ZB_JOIN_CLIENT */


void zb_zdo_startup_complete_int(zb_uint8_t param)
{
  zb_zdo_app_signal_type_t sig = zb_get_app_signal(param, NULL);
  TRACE_MSG(TRACE_ERROR, "> zb_zdo_startup_complete_int param %hd status %hd signal %hd",
            (FMT__H_H_H, param, zb_buf_get_status(param), sig));

  ZB_ASSERT((param) && (param != ZB_UNDEFINED_BUFFER));

  /* Send signal to the app */
  zboss_signal_handler(param);

  TRACE_MSG(TRACE_ZDO1, "< zb_zdo_startup_complete_int", (FMT__0));
}


void zb_af_register_device_ctx(zb_af_device_ctx_t *device_ctx)
{
  TRACE_MSG(TRACE_ZDO2, ">> zb_af_register_device_ctx", (FMT__0));

  ZB_ASSERT(!ncp_host_state_get_zboss_started());

  zb_zcl_register_device_ctx(device_ctx);

  ZB_ASSERT(zb_zcl_check_cluster_list());

  TRACE_MSG(TRACE_ZDO2, "<< zb_af_register_device_ctx", (FMT__0));
}


void zb_zdo_pim_get_long_poll_interval_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_get_long_poll_interval_req", (FMT__0));

  ZB_ASSERT(param != ZB_BUF_INVALID);

  ret = ncp_host_nwk_get_long_poll_interval(&ncp_tsn);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, param, ncp_tsn, &zdp_tsn);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_get_long_poll_interval_req", (FMT__0));
}


void zb_zdo_pim_get_in_fast_poll_flag_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_get_in_fast_poll_flag_req", (FMT__0));

  ZB_ASSERT(param != ZB_BUF_INVALID);

  ret = ncp_host_nwk_get_in_fast_poll_flag(&ncp_tsn);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, param, ncp_tsn, &zdp_tsn);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_get_in_fast_poll_flag_req", (FMT__0));
}


void ncp_host_handle_nwk_get_long_poll_interval_response(zb_ret_t status, zb_uint8_t ncp_tsn, zb_uint32_t interval)
{
  zb_bufid_t buf;
  zb_zdo_pim_get_long_poll_interval_resp_t *resp = NULL;
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_nwk_get_long_poll_interval_response ", (FMT__0));

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);

  buf = cb_ent->buffer;
  ZB_ASSERT(buf != ZB_BUF_INVALID);

  resp = ZB_BUF_GET_PARAM(buf, zb_zdo_pim_get_long_poll_interval_resp_t);

  if (status != RET_OK)
  {
    TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
      (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
    ZB_ASSERT(0);
  }

  resp->interval = interval;

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_nwk_get_long_poll_interval_response ", (FMT__0));
}


void ncp_host_handle_nwk_get_in_fast_poll_flag_response(zb_ret_t status, zb_uint8_t ncp_tsn, zb_uint8_t in_fast_poll)
{
  zb_bufid_t buf;
  zb_zdo_pim_get_in_fast_poll_flag_resp_t *resp = NULL;
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_nwk_get_in_fast_poll_flag_response ", (FMT__0));

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);

  buf = cb_ent->buffer;
  ZB_ASSERT(buf != ZB_BUF_INVALID);

  resp = ZB_BUF_GET_PARAM(buf, zb_zdo_pim_get_in_fast_poll_flag_resp_t);

  if (status != RET_OK)
  {
    TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
      (FMT__H_H, ERROR_GET_CATEGORY(status), ERROR_GET_CODE(status)));
    ZB_ASSERT(0);
  }

  resp->in_fast_poll = in_fast_poll;

  ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_nwk_get_in_fast_poll_flag_response ", (FMT__0));
}


/* Stop Fast Poll */
void zb_zdo_pim_stop_fast_poll_extended_req(zb_uint8_t param, zb_callback_t cb)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t zdp_tsn = ZB_ZDO_INVALID_TSN;
  zb_uint8_t ncp_tsn;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_stop_fast_poll_extended_req", (FMT__0));

  ZB_ASSERT(param != ZB_BUF_INVALID);

  ret = ncp_host_nwk_stop_fast_poll(&ncp_tsn);
  ZB_ASSERT(ret == RET_OK);

  ncp_host_zdo_adapter_register_cb(cb, param, ncp_tsn, &zdp_tsn);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_stop_fast_poll_extended_req, ret %d", (FMT__D, ret));
}


void zb_zdo_pim_stop_fast_poll(zb_uint8_t param)
{
  zb_ret_t ret = RET_BUSY;
  zb_uint8_t ncp_tsn;

  ZVUNUSED(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_stop_fast_poll", (FMT__0));

  ret = ncp_host_nwk_stop_fast_poll(&ncp_tsn);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_stop_fast_poll, ret %d", (FMT__D, ret));
}


void ncp_host_handle_nwk_stop_fast_poll_response(zb_ret_t status_code, zb_uint8_t ncp_tsn,
  zb_stop_fast_poll_result_t stop_result)
{
  zb_bufid_t buf;
  zb_zdo_pim_stop_fast_poll_extended_resp_t *resp = NULL;
  ncp_host_zdo_cb_ent_t *cb_ent = NULL;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_nwk_start_fast_poll_response, status_code %d, tsn %hd, stop_result %hd",
    (FMT__D_H_H, status_code, ncp_tsn, stop_result));

  get_cb_ent_by_ncp_tsn(ncp_tsn, &cb_ent);

  if (status_code != RET_OK)
  {
    TRACE_MSG(TRACE_ERROR, "Unexpected type of error, error category = %hd, error status = %hd ",
      (FMT__H_H, ERROR_GET_CATEGORY(status_code), ERROR_GET_CODE(status_code)));
    ZB_ASSERT(0);
  }

  if (cb_ent->cb != NULL)
  {
    buf = cb_ent->buffer;
    ZB_ASSERT(buf != ZB_BUF_INVALID);

    resp = ZB_BUF_GET_PARAM(buf, zb_zdo_pim_stop_fast_poll_extended_resp_t);

    resp->stop_result = stop_result;

    ncp_host_zdo_adapter_call_cb_for_blocking_req(buf, cb_ent);
  }

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_nwk_start_fast_poll_response", (FMT__0));
}



void zb_zdo_pim_set_long_poll_interval(zb_time_t ms)
{
  zb_ret_t ret = RET_BUSY;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_set_long_poll_interval", (FMT__0));

  ret = ncp_host_nwk_set_long_poll_interval(ms);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_set_long_poll_interval, ret %d", (FMT__D, ret));
}


void zb_zdo_pim_set_fast_poll_interval(zb_time_t ms)
{
  zb_ret_t ret = RET_BUSY;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_set_fast_poll_interval", (FMT__0));

  ret = ncp_host_nwk_set_fast_poll_interval(ms);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_set_fast_poll_interval, ret %d", (FMT__D, ret));
}


/* Start Fast Poll */
void zb_zdo_pim_start_fast_poll(zb_uint8_t param)
{
  zb_ret_t ret = RET_BUSY;

  ZVUNUSED(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_start_fast_poll", (FMT__0));

  ret = ncp_host_nwk_start_fast_poll();
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_start_fast_poll, ret %d", (FMT__D, ret));
}


void zb_zdo_pim_start_turbo_poll_packets(zb_uint8_t n_packets)
{
  zb_ret_t ret = RET_BUSY;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_start_turbo_poll_packets", (FMT__0));

  ret = ncp_host_nwk_start_turbo_poll_packets(n_packets);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_start_turbo_poll_packets, ret %d", (FMT__D, ret));
}


void zb_zdo_pim_start_turbo_poll_continuous(zb_time_t turbo_poll_timeout_ms)
{
  zb_ret_t ret = RET_BUSY;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_start_turbo_poll_continuous", (FMT__0));

  ret = ncp_host_nwk_start_turbo_poll_continuous(turbo_poll_timeout_ms);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_start_turbo_poll_continuous, ret %d", (FMT__D, ret));
}


void zb_zdo_pim_turbo_poll_continuous_leave(zb_uint8_t param)
{
  zb_ret_t ret = RET_BUSY;

  ZVUNUSED(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_turbo_poll_continuous_leave", (FMT__0));

  ret = ncp_host_nwk_turbo_poll_continuous_leave();
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_turbo_poll_continuous_leave, ret %d", (FMT__D, ret));
}


void zb_zdo_turbo_poll_packets_leave(zb_uint8_t param)
{
  zb_ret_t ret = RET_BUSY;

  ZVUNUSED(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_turbo_poll_packets_leave", (FMT__0));

  ret = ncp_host_nwk_turbo_poll_packets_leave();
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_turbo_poll_packets_leave, ret %d", (FMT__D, ret));
}


void zb_zdo_pim_set_fast_poll_timeout(zb_time_t ms)
{
  zb_ret_t ret = RET_BUSY;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_set_fast_poll_timeout", (FMT__0));

  ret = ncp_host_nwk_set_fast_poll_timeout(ms);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_set_fast_poll_timeout, ret %d", (FMT__D, ret));
}


void zb_zdo_pim_permit_turbo_poll(zb_uint8_t permit)
{
  zb_ret_t ret = RET_BUSY;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_permit_turbo_poll", (FMT__0));

  ret = ncp_host_nwk_permit_turbo_poll(permit);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_permit_turbo_poll, ret %d", (FMT__D, ret));
}


void zb_zdo_pim_start_poll(zb_uint8_t param)
{
  zb_ret_t ret = RET_BUSY;

  ZVUNUSED(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_start_poll", (FMT__0));

  ret = ncp_host_nwk_start_poll();
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_start_poll, ret %d", (FMT__D, ret));
}


void zb_zdo_pim_stop_poll(zb_uint8_t param)
{
  zb_ret_t ret = RET_BUSY;

  ZVUNUSED(param);

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_zdo_pim_stop_poll", (FMT__0));

  ret = ncp_host_nwk_start_poll();
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_zdo_pim_stop_poll, ret %d", (FMT__D, ret));
}

