/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*  PURPOSE: ZBOSS NCP Host startup sequence.
*/

#define ZB_TRACE_FILE_ID 94
#include "zb_common.h"
#include "ncp_host_hl_transport_internal_api.h"
#include "ncp_host_soc_state.h"

#define DELAY_BEFORE_RESET (3 * ZB_TIME_ONE_SECOND)

typedef enum ncp_startup_stage_e
{
  STARTUP_STAGE_SETTING_PARAMS,
  STARTUP_STAGE_READING_SOC_STATE,
  STARTUP_STAGE_DONE
} ncp_startup_stage_t;

typedef enum ncp_setting_param_step_e
{
  PARAM_STEP_RESET,
  /* TODO: check NCP version? */
  PARAM_STEP_LONG_ADDR,
  PARAM_STEP_SECUR_MATERIAL,
  PARAM_STEP_USE_EXTENDED_PAN_ID,
  PARAM_STEP_PAN_ID,
  PARAM_STEP_RX_ON_WHEN_IDLE,
  PARAM_STEP_DEVICE_TYPE,
  PARAM_STEP_CHANNEL_LIST,
  PARAM_STEP_PERMIT_CONTROL4_NETWORK,
  PARAM_STEP_KEEPALIVE_MODE,
  PARAM_STEP_KEEPALIVE_TIMEOUT,
  PARAM_STEP_ED_TIMEOUT,
  PARAM_STEP_SETTING_SIMPLE_DESCRIPTORS,
  /* TODO: add setting max children */
  /* TODO: add setting node and power descriptors */
  PARAM_STEP_DONE
} ncp_setting_param_step_t;

typedef enum ncp_reading_soc_state_step_e
{
  SOC_STATE_STEP_AUTHENTICATED,
  SOC_STATE_STEP_PARENT_ADDRESS,
  SOC_STATE_STEP_EXTENDED_PAN_ID,
  SOC_STATE_STEP_PAN_ID,
  SOC_STATE_STEP_COORDINATOR_VERSION,
  SOC_STATE_STEP_NETWORK_ADDRESS,
  SOC_STATE_STEP_CURRENT_PAGE_AND_CHANNEL,
  SOC_STATE_STEP_TRUST_CENTER_ADDRESS,
  SOC_STATE_STEP_TCLK,
  SOC_STATE_STEP_ADDRESS_TABLE,
  SOC_STATE_STEP_APS_KEYS_TABLE,
  SOC_STATE_STEP_GROUP_TABLE,
  SOC_STATE_STEP_DONE
} ncp_reading_soc_state_step_t;

typedef struct ncp_host_startup_state_s
{
  ncp_startup_stage_t current_stage;
  zb_uint_t current_step;
  zb_uindex_t simple_desc_index;
  zb_bool_t autostart;
  zb_uindex_t start_secur_material_index;
  zb_bool_t soc_initial_reset_resp_processed;
} ncp_host_startup_state_t;

static ncp_host_startup_state_t startup_ctx;

static void startup_dispatch(void);
static void startup_next_step(zb_uint8_t param);

static void startup_next_step(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO2, ">> startup_next_step", (FMT__0));
  ZVUNUSED(param);

  TRACE_MSG(TRACE_ZDO2, "current stage: %d", (FMT__D, startup_ctx.current_stage));
  TRACE_MSG(TRACE_ZDO2, "current step: %d", (FMT__D, startup_ctx.current_step));

  ZB_ASSERT(startup_ctx.current_stage < STARTUP_STAGE_DONE);

  startup_ctx.current_step++;
  switch(startup_ctx.current_stage)
  {
    case STARTUP_STAGE_SETTING_PARAMS:
      if (startup_ctx.current_step >= PARAM_STEP_DONE)
      {
        startup_ctx.current_stage++;
        startup_ctx.current_step = 0;
      }
      break;
    case STARTUP_STAGE_READING_SOC_STATE:
      if (startup_ctx.current_step >= SOC_STATE_STEP_DONE)
      {
        startup_ctx.current_stage++;
        startup_ctx.current_step = 0;
      }
      break;
    default:
      TRACE_MSG(TRACE_ERROR, "incorrect stage", (FMT__0));
      ZB_ASSERT(0);
  }

  TRACE_MSG(TRACE_ZDO2, "new stage: %d", (FMT__D, startup_ctx.current_stage));
  TRACE_MSG(TRACE_ZDO2, "new step: %d", (FMT__D, startup_ctx.current_step));

  startup_dispatch();

  TRACE_MSG(TRACE_ZDO2, "<< startup_next_step", (FMT__0));
}


static void soc_reset(zb_uint8_t nvram_erase)
{
  TRACE_MSG(TRACE_ZDO2, ">> soc_reset, nvram_erase %d", (FMT__D, nvram_erase));

  if (nvram_erase)
  {
    ncp_host_reset_nvram_erase();
  }
  else if (!startup_ctx.soc_initial_reset_resp_processed)
  {
    ncp_host_reset();
  }
  else
  {
    TRACE_MSG(TRACE_ZDO2, "initial Reset Response is received, reset is not needed", (FMT__0));
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }

  TRACE_MSG(TRACE_ZDO2, "<< soc_reset", (FMT__0));
}


void ncp_host_handle_reset_response(zb_ret_t status_code, zb_bool_t is_solicited)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_reset_response, status: 0x%x, is_solicited %d",
    (FMT__D_H, status_code, is_solicited));

  ZB_ASSERT(status_code == RET_OK);

  /* We can get unsolicited reset response when SoC starts */
  ZB_ASSERT(is_solicited
            || (startup_ctx.current_stage <= STARTUP_STAGE_SETTING_PARAMS
                && startup_ctx.current_step <= PARAM_STEP_RESET));

  if (is_solicited)
  {
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }
  else if (!startup_ctx.soc_initial_reset_resp_processed) {
    startup_ctx.soc_initial_reset_resp_processed = ZB_TRUE;
  }
}


static void soc_set_long_addr(const zb_ieee_addr_t long_addr)
{
  zb_ret_t ret = RET_ERROR;
  TRACE_MSG(TRACE_ZDO2, ">> soc_set_long_addr, addr: " TRACE_FORMAT_64,
            (FMT__A, TRACE_ARG_64(long_addr)));

  ZB_ASSERT(ZB_IEEE_ADDR_IS_VALID(long_addr));

  ret = ncp_host_set_local_ieee_addr(0, long_addr);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_long_addr", (FMT__0));
}


void ncp_host_handle_set_local_ieee_addr_response(zb_ret_t status_code)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_set_local_ieee_addr_response, status: 0x%x",
            (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


void ncp_host_handle_set_end_device_timeout_response(zb_ret_t status_code)
{
  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_set_end_device_timeout_response, status_code %d",
    (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  if (!ncp_host_state_get_zboss_started())
  {
    ZB_ASSERT(startup_ctx.current_step == PARAM_STEP_ED_TIMEOUT);
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_set_end_device_timeout_response", (FMT__0));
}


void ncp_host_handle_set_keepalive_timeout_response(zb_ret_t status_code)
{
  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_set_keepalive_timeout_response, status_code %d",
    (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  if (!ncp_host_state_get_zboss_started())
  {
    ZB_ASSERT(startup_ctx.current_step == PARAM_STEP_KEEPALIVE_TIMEOUT);
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_set_keepalive_timeout_response", (FMT__0));
}


void ncp_host_handle_nwk_set_keepalive_mode_response(zb_ret_t status_code)
{
  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_nwk_set_keepalive_mode_response, status_code %d",
    (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  if (!ncp_host_state_get_zboss_started())
  {
    ZB_ASSERT(startup_ctx.current_step == PARAM_STEP_KEEPALIVE_MODE);
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_nwk_set_keepalive_mode_response", (FMT__0));
}

static void soc_set_next_secur_material(zb_uint8_t param)
{
  zb_ret_t ret = RET_ERROR;
  zb_uint8_t key_idx = 0;
  zb_uint8_t nwk_key[ZB_CCM_KEY_SIZE];

  ZVUNUSED(param);

  TRACE_MSG(TRACE_ZDO2, ">> soc_set_next_secur_material", (FMT__0));

  for (key_idx = startup_ctx.start_secur_material_index; key_idx < ZB_SECUR_N_SECUR_MATERIAL; key_idx++)
  {
    if (ncp_host_state_get_nwk_key(nwk_key, key_idx))
    {
      ret = ncp_host_set_nwk_key(nwk_key, key_idx);
      ZB_ASSERT(ret == RET_OK);

      break;
    }
  }

  /* If there is not nwk key to set on SoC, just continue startup procedure */
  if (key_idx == ZB_SECUR_N_SECUR_MATERIAL)
  {
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_next_secur_material", (FMT__0));
}


void ncp_host_handle_set_nwk_key_response(zb_ret_t status_code)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_set_nwk_key_response, status: 0x%x", (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  startup_ctx.start_secur_material_index++;

  if (startup_ctx.start_secur_material_index == ZB_SECUR_N_SECUR_MATERIAL)
  {
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }
  else
  {
    ZB_SCHEDULE_CALLBACK(soc_set_next_secur_material, 0);
  }
}


static void soc_set_use_extended_pan_id(const zb_ext_pan_id_t ext_pan_id)
{
  TRACE_MSG(TRACE_ZDO2, ">> soc_set_use_extended_pan_id, ext_pan_id:" TRACE_FORMAT_64,
            (FMT__A, TRACE_ARG_64(ext_pan_id)));

  if (ZB_EXTPANID_IS_ZERO(ext_pan_id))
  {
    TRACE_MSG(TRACE_ZDO2, "use_extended_pan_id is not set, proceed to the next step", (FMT__0));
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }
  else
  {
    /* TODO: implement the corresponding NCP call */
    ZB_ASSERT(0);
  }

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_use_extended_pan_id", (FMT__0));
}


static void soc_set_pan_id(zb_uint16_t pan_id)
{
  zb_ret_t ret = RET_ERROR;
  TRACE_MSG(TRACE_ZDO2, "<< soc_set_pan_id, pan_id: 0x%x", (FMT__D, pan_id));

  if (!pan_id)
  {
    TRACE_MSG(TRACE_ZDO2, "pan_id is not set, proceed to the next step", (FMT__0));
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }
  else
  {
    ret = ncp_host_set_zigbee_pan_id(pan_id);
    ZB_ASSERT(ret == RET_OK);
  }

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_pan_id", (FMT__0));
}


void ncp_host_handle_set_zigbee_pan_id_response(zb_ret_t status_code)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_set_zigbee_pan_id_response, status: 0x%x",
            (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_set_device_type(zb_nwk_device_type_t device_type)
{
  zb_ret_t ret = RET_ERROR;
  TRACE_MSG(TRACE_ZDO2, ">> soc_set_device_type, device_type: %d", (FMT__D, device_type));

  ret = ncp_host_set_zigbee_role(device_type);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_device_type", (FMT__0));
}

void ncp_host_handle_set_zigbee_role_response(zb_ret_t status_code)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_set_device_type_response, status: 0x%x",
            (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_set_channel_list(zb_channel_list_t channel_list)
{
  zb_ret_t status = RET_ERROR;
  zb_uint32_t channel_mask_page_0;
  /* TODO: support pages other than page 0 */

  TRACE_MSG(TRACE_ZDO2, ">> soc_set_channel_list", (FMT__0));

  channel_mask_page_0 = zb_channel_page_list_get_mask(channel_list, ZB_CHANNEL_LIST_PAGE0_IDX);
  TRACE_MSG(TRACE_ZDO2, "page 0 channel mask: 0x%x", (FMT__D, channel_mask_page_0));

  status = ncp_host_set_zigbee_channel_mask(0, channel_mask_page_0);
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_channel_list", (FMT__0));
}


void ncp_host_handle_set_zigbee_channel_mask_response(zb_ret_t status_code)
{
  /* TODO: add channel page */
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_set_zigbee_channel_mask_response, status: 0x%x",
            (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_set_rx_on_when_idle(zb_bool_t rx_on)
{
  zb_ret_t status = RET_ERROR;

  TRACE_MSG(TRACE_ZDO2, ">> soc_set_rx_on_when_idle, rx_on %d", (FMT__D, rx_on));

  status = ncp_host_set_rx_on_when_idle(rx_on);
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_rx_on_when_idle", (FMT__0));
}


void ncp_host_handle_set_rx_on_when_idle_response(zb_ret_t status_code)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_set_rx_on_when_idle_response, status_code 0x%x",
            (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}

static void soc_set_permit_control4_network(zb_bool_t permit)
{
  TRACE_MSG(TRACE_ZDO2, ">> soc_set_permit_control4_network, permit_on %d", (FMT__D, permit));

  /* TODO: use NCP call when it is implemented */

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_permit_control4_network", (FMT__0));
}


/* Set Keepalive Timeout */
void soc_set_keepalive_timeout(zb_uint_t timeout)
{
  zb_ret_t ret = RET_BUSY;

  TRACE_MSG(TRACE_TRANSPORT2, ">> soc_set_keepalive_timeout", (FMT__0));

  ret = ncp_host_set_keepalive_timeout(timeout);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< soc_set_keepalive_timeout, ret %d", (FMT__D, ret));
}


/* Set End Device Timeout */
void soc_set_ed_timeout(zb_uint_t timeout)
{
  zb_ret_t ret = RET_BUSY;

  TRACE_MSG(TRACE_TRANSPORT2, ">> zb_set_ed_timeout", (FMT__0));

  ret = ncp_host_set_end_device_timeout(timeout);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< zb_set_ed_timeout, ret %d", (FMT__D, ret));
}


/* Set Keepalive Mode */
void soc_set_keepalive_mode(nwk_keepalive_supported_method_t mode)
{
  zb_ret_t ret = RET_BUSY;

  TRACE_MSG(TRACE_TRANSPORT2, ">> soc_set_keepalive_mode", (FMT__0));

  ret = ncp_host_nwk_set_keepalive_mode(mode);
  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_TRANSPORT2, "<< soc_set_keepalive_mode, ret %d", (FMT__D, ret));
}


static void soc_set_simple_descriptor(zb_af_simple_desc_1_1_t *dsc)
{
  TRACE_MSG(TRACE_ZDO2, ">> soc_set_simple_descriptor", (FMT__0));

  TRACE_MSG(TRACE_ZDO3,
            "dsc %p ep %hd prof_id 0x%x dev_id 0x%x dev_ver 0x%hx in_clu_c %hd out_clu_c %hd",
            (FMT__P_H_D_D_D_H_H, dsc, dsc->endpoint, dsc->app_profile_id, dsc->app_device_id,
             dsc->app_device_version, dsc->app_input_cluster_count, dsc->app_output_cluster_count));

  /* TODO: support cluster filtering, see ZB_FILTER_OUT_CLUSTERS */

  ncp_host_af_set_simple_descriptor(dsc->endpoint,
                                    dsc->app_profile_id,
                                    dsc->app_device_id,
                                    dsc->app_device_version,
                                    dsc->app_input_cluster_count,
                                    dsc->app_output_cluster_count,
                                    dsc->app_cluster_list,
                                    &dsc->app_cluster_list[dsc->app_input_cluster_count]);

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_simple_descriptor", (FMT__0));
}

static void soc_set_next_simple_descriptor(void);

void ncp_host_handle_af_set_simple_descriptor_response(zb_ret_t status_code)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_af_set_simple_descriptor_response, status: 0x%x",
            (FMT__D, status_code));

  ZB_ASSERT(status_code == RET_OK);

  startup_ctx.simple_desc_index++;
  soc_set_next_simple_descriptor();
}


static void soc_set_next_simple_descriptor(void)
{
  zb_af_endpoint_desc_t *ep = NULL;

  TRACE_MSG(TRACE_ZDO2, ">> soc_set_next_simple_descriptor", (FMT__0));
  TRACE_MSG(TRACE_ZDO2, "current desc index: %d", (FMT__D, startup_ctx.simple_desc_index));

  if (ZCL_CTX().device_ctx == NULL)
  {
    TRACE_MSG(TRACE_ZDO2, "device context is not registered, move to next step", (FMT__0));
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }
  else if (startup_ctx.simple_desc_index == ZCL_CTX().device_ctx->ep_count)
  {
    TRACE_MSG(TRACE_ZDO2, "all descriptors are set", (FMT__0));
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }
  else
  {
    ep = ZCL_CTX().device_ctx->ep_desc_list[startup_ctx.simple_desc_index];

    zb_zcl_init_endpoint(ep);
    soc_set_simple_descriptor(ep->simple_desc);
  }

  TRACE_MSG(TRACE_ZDO2, "<< soc_set_next_simple_descriptor", (FMT__0));
}


static void setting_param_dispatch(void)
{
  TRACE_MSG(TRACE_ZDO2, ">> setting_param_dispatch, step %d",
            (FMT__D, startup_ctx.current_step));

  switch(startup_ctx.current_step)
  {
    case PARAM_STEP_RESET:
      /* wait in case SoC is going to send an unsolicited reset response */
      ZB_SCHEDULE_ALARM(soc_reset, ncp_host_state_get_nvram_erase(), DELAY_BEFORE_RESET);
      break;
    case PARAM_STEP_LONG_ADDR:
    {
      zb_ieee_addr_t long_addr;
      ncp_host_state_get_long_addr(long_addr);
      soc_set_long_addr(long_addr);
      break;
    }
    case PARAM_STEP_SECUR_MATERIAL:
    {
      soc_set_next_secur_material(0);
      break;
    }
    case PARAM_STEP_USE_EXTENDED_PAN_ID:
    {
      zb_ext_pan_id_t use_ext_pan_id;
      ncp_host_state_get_use_extended_pan_id(use_ext_pan_id);
      soc_set_use_extended_pan_id(use_ext_pan_id);
      break;
    }
    case PARAM_STEP_PAN_ID:
      soc_set_pan_id(ncp_host_state_get_pan_id());
      break;
    case PARAM_STEP_DEVICE_TYPE:
      soc_set_device_type(ncp_host_state_get_device_type());
      break;
    case PARAM_STEP_CHANNEL_LIST:
    {
      zb_channel_list_t channel_list;
      ncp_host_state_get_channel_list(channel_list);
      soc_set_channel_list(channel_list);
      break;
    }
    case PARAM_STEP_RX_ON_WHEN_IDLE:
      soc_set_rx_on_when_idle(ncp_host_state_get_rx_on_when_idle());
      break;
    case PARAM_STEP_PERMIT_CONTROL4_NETWORK:
      soc_set_permit_control4_network(ncp_host_state_get_permit_control4_network());
      break;
    case PARAM_STEP_KEEPALIVE_MODE:
      if (ncp_host_state_keepalive_mode_is_set())
      {
        soc_set_keepalive_mode(ncp_host_state_get_keepalive_mode());
      }
      else
      {
        ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
      }
      break;
    case PARAM_STEP_KEEPALIVE_TIMEOUT:
      if (ncp_host_state_keepalive_timeout_is_set())
      {
        soc_set_keepalive_timeout(ncp_host_state_get_keepalive_timeout());
      }
      else
      {
        ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
      }
      break;
    case PARAM_STEP_ED_TIMEOUT:
      if (ncp_host_state_ed_timeout_is_set())
      {
        soc_set_ed_timeout(ncp_host_state_get_ed_timeout());
      }
      else
      {
        ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
      }
      break;
    case PARAM_STEP_SETTING_SIMPLE_DESCRIPTORS:
      soc_set_next_simple_descriptor();
      break;
    default:
      TRACE_MSG(TRACE_ERROR, "incorrect setting param step", (FMT__0));
      ZB_ASSERT(0);
  }

  TRACE_MSG(TRACE_ZDO2, "<< setting_param_dispatch", (FMT__0));
}


static void soc_get_authenticated_status(void)
{
  zb_ret_t status = RET_ERROR;

  TRACE_MSG(TRACE_ZDO2, ">> soc_get_authenticated_status", (FMT__0));

  status = ncp_host_get_authentication_status();
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_authenticated_status", (FMT__0));
}


void ncp_host_handle_get_authentication_status_response(zb_ret_t status_code, zb_bool_t authenticated)
{
  TRACE_MSG(TRACE_ZDO2,
            "ncp_host_handle_get_authentication_status_response, status_code 0x%x, auth %d",
            (FMT__D_D, status_code, authenticated));

  ZB_ASSERT(status_code == RET_OK);

  if (status_code == RET_OK)
  {
    ncp_host_state_set_authenticated(authenticated);
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_get_parent_address(void)
{
  zb_ret_t status = RET_ERROR;

  TRACE_MSG(TRACE_ZDO2, ">> soc_get_parent_address", (FMT__0));

  status = ncp_host_get_parent_address();
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_parent_address", (FMT__0));
}


void ncp_host_handle_get_parent_address_response(zb_ret_t status_code, zb_uint16_t parent_address)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_get_parent_address_response, status 0x%x, parent 0x%x",
            (FMT__D_D, status_code, parent_address));

  ZB_ASSERT(status_code == RET_OK);

  if (status_code == RET_OK)
  {
    ncp_host_state_set_parent_short_address(parent_address);
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_get_extended_pan_id(void)
{
  zb_ret_t status = RET_ERROR;
  TRACE_MSG(TRACE_ZDO2, ">> soc_get_extended_pan_id", (FMT__0));

  status = ncp_host_get_extended_pan_id();
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_extended_pan_id", (FMT__0));
}


void ncp_host_handle_get_extended_pan_id_response(zb_ret_t status_code,
                                                  const zb_ieee_addr_t ext_pan_id)
{
  TRACE_MSG(TRACE_ZDO2,
            "ncp_host_handle_get_extended_pan_id_response, status 0x%x, ext_pan_id " TRACE_FORMAT_64,
            (FMT__D_A, status_code, TRACE_ARG_64(ext_pan_id)));

  ZB_ASSERT(status_code == RET_OK);

  if (status_code == RET_OK)
  {
    ncp_host_state_set_extended_pan_id(ext_pan_id);
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_get_pan_id(void)
{
  zb_ret_t status = RET_ERROR;
  TRACE_MSG(TRACE_ZDO2, ">> soc_get_pan_id", (FMT__0));

  status = ncp_host_get_zigbee_pan_id();
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_pan_id", (FMT__0));
}


void ncp_host_handle_get_zigbee_pan_id_response(zb_ret_t status_code, zb_uint16_t pan_id)
{
  TRACE_MSG(TRACE_ZDO2,
            "ncp_host_handle_get_zigbee_pan_id_response, status 0x%x, pan_id 0x%x",
            (FMT__D_D, status_code, pan_id));

  ZB_ASSERT(status_code == RET_OK);

  if (status_code == RET_OK)
  {
    ncp_host_state_set_pan_id(pan_id);
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_get_coordinator_version(void)
{
  zb_ret_t status = RET_ERROR;
  TRACE_MSG(TRACE_ZDO2, ">> soc_get_coordinator_version", (FMT__0));

  status = ncp_host_get_coordinator_version();
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_coordinator_version", (FMT__0));
}


void ncp_host_handle_get_coordinator_version_response(zb_ret_t status_code, zb_uint8_t version)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_get_coordinator_version_response, status 0x%x, version %d",
            (FMT__D_D, status_code, version));

  ZB_ASSERT(status_code == RET_OK);

  if (status_code == RET_OK)
  {
    ncp_host_state_set_coordinator_version(version);
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_get_network_address(void)
{
  zb_ret_t status = RET_ERROR;

  TRACE_MSG(TRACE_ZDO2, ">> soc_get_network_address", (FMT__0));

  status = ncp_host_get_short_address();
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_network_address", (FMT__0));
}


void ncp_host_handle_get_short_address_response(zb_ret_t status_code, zb_uint16_t address)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_get_short_address_response, status 0x%x, address 0x%x",
            (FMT__D_D, status_code, address));

  ZB_ASSERT(status_code == RET_OK);

  if (status_code == RET_OK)
  {
    ncp_host_state_set_short_address(address);
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_get_current_page_and_channel(void)
{
  zb_ret_t status = RET_ERROR;

  TRACE_MSG(TRACE_ZDO2, ">> soc_get_current_page_and_channel", (FMT__0));

  status = ncp_host_get_zigbee_channel();
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_current_page_and_channel", (FMT__0));
}


void ncp_host_handle_get_zigbee_channel_response(zb_ret_t status_code,
                                                 zb_uint8_t page,
                                                 zb_uint8_t channel)
{
  TRACE_MSG(TRACE_ZDO2,
            "ncp_host_handle_get_zigbee_channel_response, status 0x%x, page %d, channel %d",
            (FMT__D_D_D, status_code, page, channel));

  ZB_ASSERT(status_code == RET_OK);

  if (status_code == RET_OK)
  {
    ncp_host_state_set_current_page(page);
    ncp_host_state_set_current_channel(channel);
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_get_trust_center_address(void)
{
  zb_ret_t status = RET_ERROR;

  TRACE_MSG(TRACE_ZDO2, ">> soc_get_trust_center_address", (FMT__0));

  status = ncp_host_get_trust_center_address();
  ZB_ASSERT(status == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_trust_center_address", (FMT__0));
}


void ncp_host_handle_get_trust_center_address_response(zb_ret_t status_code,
                                                       const zb_ieee_addr_t address)
{
  TRACE_MSG(TRACE_ZDO2,
            "ncp_host_handle_get_trust_center_address_response, status 0x%x, address " TRACE_FORMAT_64,
            (FMT__D_A, status_code, TRACE_ARG_64(address)));

  ZB_ASSERT(status_code == RET_OK);

  if (status_code == RET_OK)
  {
    zb_aib_set_trust_center_address(address);
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_get_tclk(void)
{
  zb_ieee_addr_t trust_center_address;
  zb_ret_t status = RET_ERROR;

  TRACE_MSG(TRACE_ZDO2, ">> soc_get_tclk", (FMT__0));

  if (!zb_aib_trust_center_address_is_unknown()
      && !zb_aib_trust_center_address_is_zero())
  {
    zb_aib_get_trust_center_address(trust_center_address);
    status = ncp_host_get_aps_key_by_ieee(trust_center_address);
    ZB_ASSERT(status == RET_OK);
  }
  else
  {
    TRACE_MSG(TRACE_ZDO2, "trust center address is unknown, skip this step", (FMT__0));
    ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
  }

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_tclk", (FMT__0));
}


void ncp_host_handle_get_aps_key_by_ieee_response(zb_ret_t status_code, const zb_uint8_t *key)
{
  TRACE_MSG(TRACE_ZDO2, "ncp_host_handle_get_aps_key_by_ieee_response, status 0x%x",
            (FMT__D, status_code));

  /* TODO: check key source and type, key check depends on commissioning type */
  if (status_code == RET_OK)
  {
    TRACE_MSG(TRACE_ZDO2, "key: " TRACE_FORMAT_128, (FMT__AA, TRACE_ARG_128(key)));
    ncp_host_state_set_tclk_valid(ZB_TRUE);
  }
  else
  {
    ncp_host_state_set_tclk_valid(ZB_FALSE);
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);
}


static void soc_get_address_table(void)
{
  TRACE_MSG(TRACE_ZDO2, ">> soc_get_address_table", (FMT__0));

  /* TODO: request address table from SoC */

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_address_table", (FMT__0));
}


static void soc_get_aps_keys_table(void)
{
  TRACE_MSG(TRACE_ZDO2, ">> soc_get_aps_keys_table", (FMT__0));

  /* TODO: request aps keys table from SoC */

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_aps_keys_table", (FMT__0));
}


static void soc_get_group_table(void)
{
  zb_ret_t ret = RET_ERROR;

  TRACE_MSG(TRACE_ZDO2, ">> soc_get_group_table", (FMT__0));

  ret = ncp_host_aps_get_group_table_request();

  ZB_ASSERT(ret == RET_OK);

  TRACE_MSG(TRACE_ZDO2, "<< soc_get_group_table", (FMT__0));
}


void ncp_host_handle_aps_get_group_table_response(zb_ret_t status_code,
                                                  zb_aps_group_table_ent_t* table,
                                                  zb_size_t entry_cnt)
{
  zb_aps_group_table_t *group_table = NULL;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_handle_aps_get_group_table_request, status 0x%x",
            (FMT__D, status_code));

  group_table = ncp_host_state_get_group_table();

  if (status_code == RET_OK)
  {
    ZB_MEMCPY(group_table, table, sizeof(*table) * entry_cnt);
    group_table->n_groups = entry_cnt;
  }

  ZB_SCHEDULE_CALLBACK(startup_next_step, 0);

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_handle_aps_get_group_table_request", (FMT__0));
}


static void reading_soc_state_dispatch(void)
{
  TRACE_MSG(TRACE_ZDO2, ">> reading_soc_state_dispatch, step %d",
            (FMT__D, startup_ctx.current_step));

  switch(startup_ctx.current_step)
  {
    case SOC_STATE_STEP_AUTHENTICATED:
      soc_get_authenticated_status();
      break;
    case SOC_STATE_STEP_PARENT_ADDRESS:
      soc_get_parent_address();
      break;
    case SOC_STATE_STEP_EXTENDED_PAN_ID:
      soc_get_extended_pan_id();
      break;
    case SOC_STATE_STEP_PAN_ID:
      soc_get_pan_id();
      break;
    case SOC_STATE_STEP_COORDINATOR_VERSION:
      soc_get_coordinator_version();
      break;
    case SOC_STATE_STEP_NETWORK_ADDRESS:
      soc_get_network_address();
      break;
    case SOC_STATE_STEP_CURRENT_PAGE_AND_CHANNEL:
      soc_get_current_page_and_channel();
      break;
    case SOC_STATE_STEP_TRUST_CENTER_ADDRESS:
      soc_get_trust_center_address();
      break;
    case SOC_STATE_STEP_TCLK:
      soc_get_tclk();
      break;
    case SOC_STATE_STEP_ADDRESS_TABLE:
      soc_get_address_table();
      break;
    case SOC_STATE_STEP_APS_KEYS_TABLE:
      soc_get_aps_keys_table();
      break;
    case SOC_STATE_STEP_GROUP_TABLE:
      soc_get_group_table();
      break;

    default:
      TRACE_MSG(TRACE_ERROR, "incorrect reading soc state step", (FMT__0));
      ZB_ASSERT(0);
  }

  TRACE_MSG(TRACE_ZDO2, "<< reading_soc_state_dispatch", (FMT__0));
}


static void startup_done(void)
{
  TRACE_MSG(TRACE_ZDO2, ">> startup_done", (FMT__0));

  ncp_host_state_set_zboss_started(ZB_TRUE);

  if (startup_ctx.autostart)
  {
    zboss_start_continue();
  }
  else
  {
    zb_buf_get_out_delayed(zb_send_no_autostart_signal);
  }

  TRACE_MSG(TRACE_ZDO2, "<< startup_done", (FMT__0));
}

static void startup_dispatch(void)
{
  TRACE_MSG(TRACE_ZDO2, ">> startup_dispatch, stage %d", (FMT__D, startup_ctx.current_stage));
  switch(startup_ctx.current_stage)
  {
    case STARTUP_STAGE_SETTING_PARAMS:
      setting_param_dispatch();
      break;
    case STARTUP_STAGE_READING_SOC_STATE:
      reading_soc_state_dispatch();
      break;
    case STARTUP_STAGE_DONE:
      startup_done();
      break;
    default:
      TRACE_MSG(TRACE_ERROR, "incorrect startup stage", (FMT__0));
      ZB_ASSERT(0);
  }
  TRACE_MSG(TRACE_ZDO2, "<< startup_dispatch", (FMT__0));
}

static void begin_startup(void)
{
  TRACE_MSG(TRACE_ZDO2, ">> begin_startup", (FMT__0));

  /* All params are zeroed on start, so just call dispatcher here */
  startup_dispatch();

  TRACE_MSG(TRACE_ZDO2, "<< begin_startup", (FMT__0));
}


zb_ret_t zboss_start(void)
{
  zb_ret_t ret = RET_ERROR;
  ZB_ASSERT(!ncp_host_state_get_zboss_started());

  startup_ctx.autostart = ZB_TRUE;
  ret = zboss_start_no_autostart();

  /* Default FFD descriptors will be set at device.
     1. call zboss_start_no_autostart
     2. start commissioning_cb
  */
  return ret;
}


void zboss_start_continue(void)
{
  TRACE_MSG(TRACE_ZDO2, ">> zboss_start_continue", (FMT__0));

  zb_buf_get_out_delayed(zdo_commissioning_start);

  TRACE_MSG(TRACE_ZDO2, "<< zboss_start_continue", (FMT__0));
}


/*
   load production config - config is on the dev side - not needed?
   set node descriptor - Set Node Descriptor
   After setting Node Descriptor and other stuff send manually ZB_ZDO_SIGNAL_SKIP_STARTUP?
 */
zb_ret_t zboss_start_no_autostart(void)
{
  zb_ret_t ret = RET_ERROR;
  ZB_ASSERT(!ncp_host_state_get_zboss_started());

  ret = ncp_host_state_validate_start_state();

  if (ret == RET_OK)
  {
    begin_startup();
  }
  else
  {
    TRACE_MSG(TRACE_ERROR, "invalid start params", (FMT__0));
  }

  return ret;
}
