/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*  PURPOSE: NCP High level transport implementation for the host side: SECUR category
*/

#define ZB_TRACE_FILE_ID 18

#include "ncp_host_hl_proto.h"

static void handle_secur_add_ic_response(ncp_hl_response_header_t* response,
                                         zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  ZVUNUSED(len);

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_secur_ic_add_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  ncp_host_handle_secur_add_ic_response(error_code);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_secur_ic_add_response", (FMT__0));
}

void ncp_host_handle_secur_response(void* data, zb_uint16_t len)
{
  ncp_hl_response_header_t* response_header = (ncp_hl_response_header_t*)data;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_secur_response", (FMT__0));

  switch(response_header->hdr.call_id)
  {
    case NCP_HL_SECUR_ADD_IC:
      handle_secur_add_ic_response(response_header, len);
      break;
    default:
      /* TODO: implement handlers for other responses! */
      ZB_ASSERT(0);
  }

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_secur_response", (FMT__0));
}

static void handle_secur_tclk_indication(ncp_hl_ind_header_t* indication, zb_uint16_t len)
{
  ncp_host_hl_rx_buf_handle_t body;
  zb_ieee_addr_t trust_center_address;
  zb_uint8_t key_type;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_secur_tclk_indication", (FMT__0));

  ncp_host_hl_init_indication_body(indication, len, &body);

  ncp_host_hl_buf_get_u64addr(&body, trust_center_address);
  ncp_host_hl_buf_get_u8(&body, &key_type);

  TRACE_MSG(TRACE_TRANSPORT3, "trust_center_address" TRACE_FORMAT_64 " key_type: 0x%x",
            (FMT__A_D, TRACE_ARG_64(trust_center_address), key_type));

  ncp_host_handle_secur_tclk_indication(RET_OK, trust_center_address, key_type);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_secur_tclk_indication", (FMT__0));
}

static void handle_secur_tclk_exchange_failed_indication(ncp_hl_ind_header_t* indication,
                                                         zb_uint16_t len)
{
  ncp_host_hl_rx_buf_handle_t body;
  zb_uint8_t status_category;
  zb_uint8_t status_code;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_secur_tclk_exchange_failed_indication", (FMT__0));

  ncp_host_hl_init_indication_body(indication, len, &body);

  ncp_host_hl_buf_get_u8(&body, &status_category);
  ncp_host_hl_buf_get_u8(&body, &status_code);

  ncp_host_handle_secur_tclk_indication(ERROR_CODE(status_category, status_code), NULL, 0);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_secur_tclk_exchange_failed_indication", (FMT__0));
}

void ncp_host_handle_secur_indication(void* data, zb_uint16_t len)
{
  ncp_hl_ind_header_t* indication_header = (ncp_hl_ind_header_t*)data;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_secur_indication", (FMT__0));

  switch(indication_header->call_id)
  {
    case NCP_HL_SECUR_TCLK_IND:
      handle_secur_tclk_indication(indication_header, len);
      break;
    case NCP_HL_SECUR_TCLK_EXCHANGE_FAILED_IND:
      handle_secur_tclk_exchange_failed_indication(indication_header, len);
      break;
    default:
      /* TODO: implement handlers for other indications! */
      ZB_ASSERT(0);
    }

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_secur_indication", (FMT__0));
}

zb_ret_t ncp_host_secur_add_ic(zb_ieee_addr_t address, zb_uint8_t ic_type, zb_uint8_t *ic)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_secur_add_ic, address " TRACE_FORMAT_64 ", ic_type %d ",
            (FMT__A_D, TRACE_ARG_64(address), ic_type));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_SECUR_ADD_IC, &body, NULL))
  {
    ncp_host_hl_buf_put_u64addr(&body, address);
    ncp_host_hl_buf_put_array(&body, ic, ZB_IC_SIZE_BY_TYPE(ic_type) + 2);
    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_secur_add_ic, ret %d", (FMT__D, ret));

  return ret;
}
