/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*  PURPOSE: NCP High level transport implementation for the host side: ZDO category
*/

#define ZB_TRACE_FILE_ID 15

#include "ncp_host_hl_proto.h"


static void handle_zdo_dev_annce_indication(ncp_hl_ind_header_t* indication, zb_uint16_t len)
{
  ncp_host_hl_rx_buf_handle_t body;
  zb_uint16_t nwk_addr;
  zb_ieee_addr_t ieee_addr;
  zb_uint8_t capability;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_dev_annce_indication ",(FMT__0));

  TRACE_MSG(TRACE_TRANSPORT3, ">> parse handle_zdo_dev_annce_indication", (FMT__0));

  ncp_host_hl_init_indication_body(indication, len, &body);
  ncp_host_hl_buf_get_u16(&body, &nwk_addr);
  ncp_host_hl_buf_get_u64addr(&body, ieee_addr);
  ncp_host_hl_buf_get_u8(&body, &capability);

  TRACE_MSG(TRACE_TRANSPORT3, "<< parse handle_zdo_dev_annce_indication, nwk_addr 0x%x, long_addr " TRACE_FORMAT_64 " mac_capabilities %hd",
            (FMT__D_A_H, nwk_addr, TRACE_ARG_64(ieee_addr), capability));

  ncp_host_handle_zdo_device_annce_indication(nwk_addr, ieee_addr, capability);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_dev_annce_indication", (FMT__0));
}

static void handle_zdo_ieee_addr_response(ncp_hl_response_header_t* response,
                                          zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;
  ncp_host_hl_rx_buf_handle_t body;
  zb_zdo_ieee_addr_resp_t resp;
  zb_bool_t is_extended = ZB_FALSE;
  zb_zdo_ieee_addr_resp_ext_t resp_ext;
  zb_zdo_ieee_addr_resp_ext2_t resp_ext2;
  zb_uint16_t *addr_list;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_ieee_addr_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  if (error_code == RET_OK)
  {
    TRACE_MSG(TRACE_TRANSPORT3, ">> parse handle_zdo_ieee_addr_response", (FMT__0));

    ncp_host_hl_init_response_body(response, len, &body);
    ncp_host_hl_buf_get_u64addr(&body, resp.ieee_addr_remote_dev);
    ncp_host_hl_buf_get_u16(&body, &resp.nwk_addr_remote_dev);

    TRACE_MSG(TRACE_TRANSPORT3, " ieee_addr = " TRACE_FORMAT_64 " nwk_addr = 0x%x",
              (FMT__A_D, TRACE_ARG_64(resp.ieee_addr_remote_dev), resp.nwk_addr_remote_dev));

    if (len > sizeof(ncp_hl_response_header_t) + sizeof(zb_uint16_t) + sizeof(zb_ieee_addr_t))
    {
      is_extended = ZB_TRUE;
      ncp_host_hl_buf_get_u8(&body, &resp_ext.num_assoc_dev);

      TRACE_MSG(TRACE_TRANSPORT3, "num_assoc_dev = %h", (FMT__H, resp_ext.num_assoc_dev));

      if(resp_ext.num_assoc_dev != 0)
      {
        ncp_host_hl_buf_get_u8(&body, &resp_ext2.start_index);
        TRACE_MSG(TRACE_TRANSPORT3, "start_index = %h", (FMT__H, resp_ext2.start_index));

        ncp_host_hl_buf_get_ptr(&body, (zb_uint8_t **)&addr_list, resp_ext.num_assoc_dev);
      }
    }

    TRACE_MSG(TRACE_TRANSPORT3, "<< parse handle_zdo_ieee_addr_response", (FMT__0));
  }

  ncp_host_handle_zdo_ieee_addr_response(error_code, ncp_tsn,
                                         &resp, is_extended,
                                         &resp_ext, &resp_ext2, addr_list);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_ieee_addr_response", (FMT__0));
}

static void handle_zdo_nwk_addr_response(ncp_hl_response_header_t* response,
                                         zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;
  ncp_host_hl_rx_buf_handle_t body;
  zb_zdo_nwk_addr_resp_head_t resp;
  zb_bool_t is_extended = ZB_FALSE;
  zb_zdo_nwk_addr_resp_ext_t resp_ext;
  zb_zdo_nwk_addr_resp_ext2_t resp_ext2;
  zb_uint16_t *addr_list;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_nwk_addr_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  if (error_code == RET_OK)
  {
    TRACE_MSG(TRACE_TRANSPORT3, ">> parse handle_zdo_nwk_addr_response", (FMT__0));

    ncp_host_hl_init_response_body(response, len, &body);
    ncp_host_hl_buf_get_u64addr(&body, resp.ieee_addr);
    ncp_host_hl_buf_get_u16(&body, &resp.nwk_addr);

    TRACE_MSG(TRACE_TRANSPORT3, " ieee_addr = " TRACE_FORMAT_64 " nwk_addr = 0x%x",
              (FMT__A_D, TRACE_ARG_64(resp.ieee_addr), resp.nwk_addr));

    if (len > sizeof(ncp_hl_response_header_t) + sizeof(zb_uint16_t) + sizeof(zb_ieee_addr_t))
    {
      is_extended = ZB_TRUE;
      ncp_host_hl_buf_get_u8(&body, &resp_ext.num_assoc_dev);

      TRACE_MSG(TRACE_TRANSPORT3, "num_assoc_dev = %h", (FMT__H, resp_ext.num_assoc_dev));

      if(resp_ext.num_assoc_dev != 0)
      {
        ncp_host_hl_buf_get_u8(&body, &resp_ext2.start_index);
        TRACE_MSG(TRACE_TRANSPORT3, "start_index = %h", (FMT__H, resp_ext2.start_index));

        ncp_host_hl_buf_get_ptr(&body, (zb_uint8_t **)&addr_list, resp_ext.num_assoc_dev);
      }
    }

    TRACE_MSG(TRACE_TRANSPORT3, "<< parse handle_zdo_nwk_addr_response", (FMT__0));
  }

  ncp_host_handle_zdo_nwk_addr_response(error_code, ncp_tsn,
                                        &resp, is_extended,
                                        &resp_ext, &resp_ext2, addr_list);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_nwk_addr_response", (FMT__0));
}

static void handle_zdo_power_descriptor_response(ncp_hl_response_header_t* response,
                                                 zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;
  ncp_host_hl_rx_buf_handle_t body;
  zb_uint16_t power_descriptor;
  zb_uint16_t nwk_addr;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_power_descriptor_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  if (error_code == RET_OK)
  {
    TRACE_MSG(TRACE_TRANSPORT3, ">> parse handle_zdo_power_descriptor_response", (FMT__0));

    ncp_host_hl_init_response_body(response, len, &body);
    ncp_host_hl_buf_get_u16(&body, &power_descriptor);
    ncp_host_hl_buf_get_u16(&body, &nwk_addr);

    TRACE_MSG(TRACE_TRANSPORT3, ">> power_descriptor = 0x%x ", (FMT__D, power_descriptor));

    TRACE_MSG(TRACE_TRANSPORT3, "<< parse handle_zdo_power_descriptor_response", (FMT__0));
  }

  ncp_host_handle_zdo_power_descriptor_response(error_code, ncp_tsn, nwk_addr, power_descriptor);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_power_descriptor_response", (FMT__0));
}

static void handle_zdo_node_descriptor_response(ncp_hl_response_header_t* response,
                                                zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;
  ncp_host_hl_rx_buf_handle_t body;
  zb_af_node_desc_t node_desc;
  zb_uint16_t nwk_addr;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_node_descriptor_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  if (error_code == RET_OK)
  {
    TRACE_MSG(TRACE_TRANSPORT3, ">> parse handle_zdo_node_descriptor_response", (FMT__0));

    ncp_host_hl_init_response_body(response, len, &body);

    ncp_host_hl_buf_get_u16(&body, &node_desc.node_desc_flags);
    TRACE_MSG(TRACE_TRANSPORT3, "node_desc.node_desc_flags 0x%x", (FMT__D, node_desc.node_desc_flags));

    ncp_host_hl_buf_get_u8(&body, &node_desc.mac_capability_flags);
    TRACE_MSG(TRACE_TRANSPORT3, "node_desc.mac_capability_flags 0x%hx",
              (FMT__H, node_desc.mac_capability_flags));

    ncp_host_hl_buf_get_u16(&body, &node_desc.manufacturer_code);
    TRACE_MSG(TRACE_TRANSPORT3, "node_desc.manufacturer_code 0x%x", (FMT__D, node_desc.manufacturer_code));

    ncp_host_hl_buf_get_u8(&body, &node_desc.max_buf_size);
    TRACE_MSG(TRACE_TRANSPORT3, "node_desc.max_buf_size %hd", (FMT__H, node_desc.max_buf_size));

    ncp_host_hl_buf_get_u16(&body, &node_desc.max_incoming_transfer_size);
    TRACE_MSG(TRACE_TRANSPORT3, "node_desc.max_incoming_transfer_size %d",
              (FMT__D, node_desc.max_incoming_transfer_size));

    ncp_host_hl_buf_get_u16(&body, &node_desc.server_mask);
    TRACE_MSG(TRACE_TRANSPORT3, "node_desc.server_mask 0x%x", (FMT__D, node_desc.server_mask));

    ncp_host_hl_buf_get_u16(&body, &node_desc.max_outgoing_transfer_size);
    TRACE_MSG(TRACE_TRANSPORT3, "node_desc.max_outgoing_transfer_size %d",
              (FMT__D, node_desc.max_outgoing_transfer_size));

    ncp_host_hl_buf_get_u8(&body, &node_desc.desc_capability_field);
    TRACE_MSG(TRACE_TRANSPORT3, "node_desc.desc_capability_field 0x%hx",
              (FMT__H, node_desc.desc_capability_field));

    ncp_host_hl_buf_get_u16(&body, &nwk_addr);
    TRACE_MSG(TRACE_TRANSPORT3, "nwk_addr 0x%x", (FMT__D, nwk_addr));

    TRACE_MSG(TRACE_TRANSPORT3, "<< parse handle_zdo_node_descriptor_response", (FMT__0));
  }

  ncp_host_handle_zdo_node_descriptor_response(error_code, ncp_tsn, nwk_addr, &node_desc);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_node_descriptor_response", (FMT__0));
}

static void handle_zdo_simple_descriptor_response(ncp_hl_response_header_t* response,
                                                  zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;
  ncp_host_hl_rx_buf_handle_t body;
  zb_af_simple_desc_1_1_t simple_desc;
  zb_uint16_t *app_cluster_list_ptr;
  zb_uint16_t nwk_addr;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_simple_descriptor_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  if (error_code == RET_OK)
  {
    TRACE_MSG(TRACE_TRANSPORT3, ">> parse handle_zdo_simple_descriptor_response", (FMT__0));

    ncp_host_hl_init_response_body(response, len, &body);

    ncp_host_hl_buf_get_u8(&body, &simple_desc.endpoint);
    TRACE_MSG(TRACE_TRANSPORT3, "simple_desc.endpoint %hd", (FMT__H, simple_desc.endpoint));

    ncp_host_hl_buf_get_u16(&body, &simple_desc.app_profile_id);
    TRACE_MSG(TRACE_TRANSPORT3, "simple_desc.app_profile_id 0x%x", (FMT__D, simple_desc.app_profile_id));

    ncp_host_hl_buf_get_u16(&body, &simple_desc.app_device_id);
    TRACE_MSG(TRACE_TRANSPORT3, "simple_desc.app_device_id 0x%x", (FMT__D, simple_desc.app_device_id));

    ncp_host_hl_buf_get_u8(&body, (zb_uint8_t *)(&simple_desc.app_device_id + 1));
    TRACE_MSG(TRACE_TRANSPORT3, "simple_desc.app_device_version 0x%hx",
              (FMT__H, simple_desc.app_device_version));

    ncp_host_hl_buf_get_u8(&body, &simple_desc.app_input_cluster_count);
    TRACE_MSG(TRACE_TRANSPORT3, "simple_desc.app_input_cluster_count %hd",
              (FMT__H, simple_desc.app_input_cluster_count));

    ncp_host_hl_buf_get_u8(&body, &simple_desc.app_output_cluster_count);
    TRACE_MSG(TRACE_TRANSPORT3, "simple_desc.app_output_cluster_count %hd",
              (FMT__H, simple_desc.app_output_cluster_count));

    ncp_host_hl_buf_get_ptr(&body, (zb_uint8_t **)&app_cluster_list_ptr,
                            sizeof(zb_uint16_t) *
                            (simple_desc.app_input_cluster_count + simple_desc.app_output_cluster_count));

    ncp_host_hl_buf_get_u16(&body, &nwk_addr);
    TRACE_MSG(TRACE_TRANSPORT3, "nwk_addr 0x%x", (FMT__D, nwk_addr));

    TRACE_MSG(TRACE_TRANSPORT3, "<< parse handle_zdo_simple_descriptor_response", (FMT__0));
  }

  ncp_host_handle_zdo_simple_descriptor_response(error_code, ncp_tsn, nwk_addr,
                                                 &simple_desc, app_cluster_list_ptr);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_simple_descriptor_response", (FMT__0));
}

static void handle_zdo_active_ep_response(ncp_hl_response_header_t* response,
                                          zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;
  ncp_host_hl_rx_buf_handle_t body;
  zb_uint8_t ep_count;
  zb_uint8_t *ep_list;
  zb_uint16_t nwk_addr;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_active_ep_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  if (error_code == RET_OK)
  {
    TRACE_MSG(TRACE_TRANSPORT3, ">> parse handle_zdo_active_ep_response", (FMT__0));

    ncp_host_hl_init_response_body(response, len, &body);

    ncp_host_hl_buf_get_u8(&body, &ep_count);
    TRACE_MSG(TRACE_TRANSPORT3, ">> endpoint count = %hd ", (FMT__H, ep_count));

    ncp_host_hl_buf_get_ptr(&body, &ep_list, ep_count);

    ncp_host_hl_buf_get_u16(&body, &nwk_addr);
    TRACE_MSG(TRACE_TRANSPORT3, "nwk_addr 0x%x", (FMT__D, nwk_addr));

    TRACE_MSG(TRACE_TRANSPORT3, "<< parse handle_zdo_active_ep_response", (FMT__0));
  }

  ncp_host_handle_zdo_active_ep_response(error_code, ncp_tsn, nwk_addr, ep_count, ep_list);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_active_ep_response", (FMT__0));
}

static void handle_zdo_match_desc_response(ncp_hl_response_header_t* response,
                                           zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;
  ncp_host_hl_rx_buf_handle_t body;
  zb_uint8_t match_ep_count;
  zb_uint8_t *match_ep_list;
  zb_uint16_t nwk_addr;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_match_desc_response, status_code %d",
            (FMT__D, error_code));

  if (error_code == RET_OK)
  {
    TRACE_MSG(TRACE_TRANSPORT3, ">> parse handle_zdo_match_desc_response", (FMT__0));

    ncp_host_hl_init_response_body(response, len, &body);

    ncp_host_hl_buf_get_u8(&body, &match_ep_count);
    TRACE_MSG(TRACE_TRANSPORT3, ">> match endpoint count = %hd ", (FMT__H, match_ep_count));

    ncp_host_hl_buf_get_ptr(&body, &match_ep_list, match_ep_count);

    ncp_host_hl_buf_get_u16(&body, &nwk_addr);
    TRACE_MSG(TRACE_TRANSPORT3, "nwk_addr 0x%x", (FMT__D, nwk_addr));

    TRACE_MSG(TRACE_TRANSPORT3, "<< parse handle_zdo_match_desc_response", (FMT__0));
  }

  ncp_host_handle_zdo_match_desc_response(error_code, ncp_tsn, nwk_addr, match_ep_count, match_ep_list);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_match_ep_response", (FMT__0));
}

static void handle_zdo_bind_response(ncp_hl_response_header_t* response,
                                     zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;

  ZVUNUSED(len);

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_bind_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  ncp_host_handle_zdo_bind_response(error_code, ncp_tsn);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_bind_response", (FMT__0));
}

static void handle_zdo_unbind_response(ncp_hl_response_header_t* response,
                                       zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;

  ZVUNUSED(len);

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_unbind_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  ncp_host_handle_zdo_unbind_response(error_code, ncp_tsn);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_unbind_response", (FMT__0));
}


static void handle_zdo_permit_joining_response(ncp_hl_response_header_t* response, zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;

  ZVUNUSED(len);

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_permit_joining_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  ncp_host_handle_zdo_permit_joining_response(error_code, ncp_tsn);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_permit_joining_response", (FMT__0));
}


static void handle_zdo_mgmt_leave_response(ncp_hl_response_header_t* response, zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;

  ZVUNUSED(len);

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_mgmt_leave_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  ncp_host_handle_zdo_mgmt_leave_response(error_code, ncp_tsn);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_mgmt_leave_response", (FMT__0));
}


void handle_zdo_rejoin_response(ncp_hl_response_header_t* response, zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);

  ZVUNUSED(len);

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_rejoin_response, status_code %d",
            (FMT__D, error_code));

  ncp_host_mark_blocking_request_finished();

  ncp_host_handle_zdo_rejoin_response(error_code);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_rejoin_response", (FMT__0));
}

void handle_zdo_system_server_discovery_response(ncp_hl_response_header_t* response, zb_uint16_t len)
{
  zb_ret_t error_code = ERROR_CODE(response->status_category, response->status_code);
  zb_uint8_t ncp_tsn = response->tsn;
  ncp_host_hl_rx_buf_handle_t body;
  zb_uint16_t server_mask;

  TRACE_MSG(TRACE_TRANSPORT3, ">> handle_zdo_system_server_discovery_response, status_code %d",
            (FMT__D, error_code));

  if (error_code == RET_OK)
  {
    ncp_host_hl_init_response_body(response, len, &body);

    ncp_host_hl_buf_get_u16(&body, &server_mask);
    TRACE_MSG(TRACE_TRANSPORT3, "server_mask 0x%x", (FMT__D, server_mask));
  }

  ncp_host_handle_zdo_system_server_discovery_response(error_code, ncp_tsn, server_mask);

  TRACE_MSG(TRACE_TRANSPORT3, "<< handle_zdo_system_server_discovery_response", (FMT__0));
}


void ncp_host_handle_zdo_response(void* data, zb_uint16_t len)
{
  ncp_hl_response_header_t* response_header = (ncp_hl_response_header_t*)data;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_response", (FMT__0));

  switch(response_header->hdr.call_id)
  {
    case NCP_HL_ZDO_IEEE_ADDR_REQ:
      handle_zdo_ieee_addr_response(response_header, len);
      break;
    case NCP_HL_ZDO_NWK_ADDR_REQ:
      handle_zdo_nwk_addr_response(response_header, len);
      break;
    case NCP_HL_ZDO_POWER_DESC_REQ:
      handle_zdo_power_descriptor_response(response_header, len);
      break;
    case NCP_HL_ZDO_NODE_DESC_REQ:
      handle_zdo_node_descriptor_response(response_header, len);
      break;
    case NCP_HL_ZDO_SIMPLE_DESC_REQ:
      handle_zdo_simple_descriptor_response(response_header, len);
      break;
    case NCP_HL_ZDO_ACTIVE_EP_REQ:
      handle_zdo_active_ep_response(response_header, len);
      break;
    case NCP_HL_ZDO_MATCH_DESC_REQ:
      handle_zdo_match_desc_response(response_header, len);
      break;
    case NCP_HL_ZDO_BIND_REQ:
      handle_zdo_bind_response(response_header, len);
      break;
    case NCP_HL_ZDO_UNBIND_REQ:
      handle_zdo_unbind_response(response_header, len);
      break;
    case NCP_HL_ZDO_PERMIT_JOINING_REQ:
      handle_zdo_permit_joining_response(response_header, len);
      break;
    case NCP_HL_ZDO_MGMT_LEAVE_REQ:
      handle_zdo_mgmt_leave_response(response_header, len);
      break;
    case NCP_HL_ZDO_REJOIN:
      handle_zdo_rejoin_response(response_header, len);
      break;
    case NCP_HL_ZDO_SYSTEM_SRV_DISCOVERY_REQ:
      handle_zdo_system_server_discovery_response(response_header, len);
      break;
    default:
      /* TODO: implement handlers for other responses! */
      ZB_ASSERT(0);
    }

   TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_response", (FMT__0));
}

void ncp_host_handle_zdo_indication(void* data, zb_uint16_t len)
{
  ncp_hl_ind_header_t* indication_header = (ncp_hl_ind_header_t*)data;

  TRACE_MSG(TRACE_TRANSPORT3, ">> ncp_host_handle_zdo_indication", (FMT__0));

  switch(indication_header->call_id)
  {
    case NCP_HL_ZDO_DEV_ANNCE_IND:
      handle_zdo_dev_annce_indication(indication_header, len);
      break;
    default:
      /* TODO: implement handlers for other responses! */
      ZB_ASSERT(0);
    }

  TRACE_MSG(TRACE_TRANSPORT3, "<< ncp_host_handle_zdo_indication", (FMT__0));
}

zb_ret_t ncp_host_zdo_ieee_addr_request(zb_uint8_t *ncp_tsn, zb_zdo_ieee_addr_req_param_t *req_param)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_ieee_addr_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_ZDO_IEEE_ADDR_REQ, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, req_param->dst_addr);
    ncp_host_hl_buf_put_u16(&body, req_param->nwk_addr);
    ncp_host_hl_buf_put_u8(&body, req_param->request_type);
    ncp_host_hl_buf_put_u8(&body, req_param->start_index);
    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_ieee_addr_request, ret %d", (FMT__D, ret));

  return ret;
}

zb_ret_t ncp_host_zdo_nwk_addr_request(zb_uint8_t *ncp_tsn, zb_zdo_nwk_addr_req_param_t *req_param)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_nwk_addr_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_ZDO_NWK_ADDR_REQ, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, req_param->dst_addr);
    ncp_host_hl_buf_put_u64addr(&body, req_param->ieee_addr);
    ncp_host_hl_buf_put_u8(&body, req_param->request_type);
    ncp_host_hl_buf_put_u8(&body, req_param->start_index);

    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_nwk_addr_request, ret %d", (FMT__D, ret));

  return ret;
}

zb_ret_t ncp_host_zdo_power_desc_request(zb_uint8_t *ncp_tsn, zb_uint16_t nwk_addr)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_power_desc_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_ZDO_POWER_DESC_REQ, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, nwk_addr);
    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_power_desc_request, ret %d", (FMT__D, ret));

  return ret;
}

zb_ret_t ncp_host_zdo_node_desc_request(zb_uint8_t *ncp_tsn, zb_uint16_t nwk_addr)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_node_desc_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_ZDO_NODE_DESC_REQ, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, nwk_addr);
    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_node_desc_request, ret %d", (FMT__D, ret));

  return ret;
}

zb_ret_t ncp_host_zdo_simple_desc_request(zb_uint8_t *ncp_tsn, zb_uint16_t nwk_addr, zb_uint8_t endpoint)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_simple_desc_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_ZDO_SIMPLE_DESC_REQ, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, nwk_addr);
    ncp_host_hl_buf_put_u8(&body, endpoint);
    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_simple_desc_request, ret %d", (FMT__D, ret));

  return ret;
}

zb_ret_t ncp_host_zdo_active_ep_request(zb_uint8_t *ncp_tsn, zb_uint16_t nwk_addr)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_active_ep_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_ZDO_ACTIVE_EP_REQ, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, nwk_addr);
    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_active_ep_request, ret %d", (FMT__D, ret));

  return ret;
}

zb_ret_t ncp_host_zdo_permit_joining_request(zb_uint8_t *ncp_tsn, zb_uint16_t short_addr,
                                             zb_uint8_t permit_duration, zb_uint8_t tc_significance)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_permit_joining_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_ZDO_PERMIT_JOINING_REQ, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, short_addr);
    ncp_host_hl_buf_put_u8(&body, permit_duration);
    ncp_host_hl_buf_put_u8(&body, tc_significance);
    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_permit_joining_request, ret %d", (FMT__D, ret));

  return ret;
}

zb_ret_t ncp_host_zdo_match_desc_request(zb_uint8_t *ncp_tsn, zb_zdo_match_desc_param_t *req)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;
  zb_uint8_t i;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_match_desc_request", (FMT__0));

  if (!ncp_host_get_buf_for_nonblocking_request(NCP_HL_ZDO_MATCH_DESC_REQ, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, req->addr_of_interest);
    TRACE_MSG(TRACE_TRANSPORT3, "addr_of_interest 0x%x", (FMT__D, req->addr_of_interest));

    ncp_host_hl_buf_put_u16(&body, req->profile_id);
    TRACE_MSG(TRACE_TRANSPORT3, "app_profile_id 0x%x", (FMT__D, req->profile_id));

    ncp_host_hl_buf_put_u8(&body, req->num_in_clusters);
    TRACE_MSG(TRACE_TRANSPORT3, "num_in_clusters %hd", (FMT__H, req->num_in_clusters));

    ncp_host_hl_buf_put_u8(&body, req->num_out_clusters);
    TRACE_MSG(TRACE_TRANSPORT3, "num_out_clusters %hd", (FMT__H, req->num_out_clusters));

    TRACE_MSG(TRACE_TRANSPORT3, "in_clusters: ", (FMT__0));
    for (i = 0; i < req->num_in_clusters; i++)
    {
      ncp_host_hl_buf_put_u16(&body, req->cluster_list[i]);
      TRACE_MSG(TRACE_TRANSPORT3, " 0x%x", (FMT__D, req->cluster_list[i]));
    }

    TRACE_MSG(TRACE_TRANSPORT3, "out_clusters: ", (FMT__0));
    for (; i < req->num_in_clusters + req->num_out_clusters; i++)
    {
      ncp_host_hl_buf_put_u16(&body, req->cluster_list[i]);
      TRACE_MSG(TRACE_TRANSPORT3, " 0x%x", (FMT__D, req->cluster_list[i]));
    }

    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_match_desc_request, ret %d", (FMT__D, ret));

  return ret;
}


zb_ret_t ncp_host_zdo_bind_unbind_request(zb_uint8_t *ncp_tsn, zb_zdo_bind_req_param_t *bind_param,
                                          zb_bool_t is_bind)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;
  zb_uint16_t ncp_command_id = (is_bind) ? NCP_HL_ZDO_BIND_REQ : NCP_HL_ZDO_UNBIND_REQ;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_bind_unbind_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(ncp_command_id, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, bind_param->req_dst_addr);
    ncp_host_hl_buf_put_u64addr(&body, bind_param->src_address);
    ncp_host_hl_buf_put_u8(&body, bind_param->src_endp);
    ncp_host_hl_buf_put_u16(&body, bind_param->cluster_id);
    ncp_host_hl_buf_put_u8(&body, bind_param->dst_addr_mode);

    if (bind_param->dst_addr_mode == ZB_APS_ADDR_MODE_16_GROUP_ENDP_NOT_PRESENT)
    {
      ncp_host_hl_buf_put_u16(&body, bind_param->dst_address.addr_short);

      /* TODO: Fix expected size in the device and remove these additional 7 bits */
      ncp_host_hl_buf_put_u32(&body, 0);
      ncp_host_hl_buf_put_u16(&body, 0);
      ncp_host_hl_buf_put_u8(&body, 0);
    }
    else if (bind_param->dst_addr_mode == ZB_APS_ADDR_MODE_64_ENDP_PRESENT)
    {
      ncp_host_hl_buf_put_u64addr(&body, bind_param->dst_address.addr_long);
      ncp_host_hl_buf_put_u8(&body, bind_param->dst_endp);
    }

    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_bind_unbind_request, ret %d", (FMT__D, ret));

  return ret;
}


zb_ret_t ncp_host_zdo_mgmt_leave_request(zb_uint8_t *ncp_tsn, zb_zdo_mgmt_leave_param_t *req)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;
  zb_uint8_t flags = 0;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_mgmt_leave_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_ZDO_MGMT_LEAVE_REQ, &body, ncp_tsn))
  {
    if (req->remove_children)
    {
      ZB_SET_BIT_IN_BIT_VECTOR(&flags, NCP_HL_ZDO_LEAVE_FLAG_REMOVE_CHILDREN);
    }

    if (req->rejoin)
    {
      ZB_SET_BIT_IN_BIT_VECTOR(&flags, NCP_HL_ZDO_LEAVE_FLAG_REJOIN);
    }

    ncp_host_hl_buf_put_u16(&body, req->dst_addr);
    TRACE_MSG(TRACE_TRANSPORT3, "dst_addr 0x%x", (FMT__D, req->dst_addr));

    ncp_host_hl_buf_put_u64addr(&body, req->device_address);
    TRACE_MSG(TRACE_TRANSPORT3, "src_addr " TRACE_FORMAT_64, (FMT__A, TRACE_ARG_64(req->device_address)));

    ncp_host_hl_buf_put_u8(&body, flags);
    TRACE_MSG(TRACE_TRANSPORT3, "flags 0x%x", (FMT__D, flags));

    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_mgmt_leave_request, ret %d", (FMT__D, ret));

  return ret;
}


zb_ret_t ncp_host_zdo_rejoin_request(zb_uint8_t *ncp_tsn,
                                     zb_ext_pan_id_t ext_pan_id,
                                     zb_channel_page_t *channels_list,
                                     zb_uint8_t secure_rejoin)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_rejoin_request", (FMT__0));

  if (!ncp_host_get_buf_for_blocking_request(NCP_HL_ZDO_REJOIN, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u64addr(&body, ext_pan_id);
    ncp_host_hl_buf_put_array(&body, (zb_uint8_t*) channels_list, sizeof(*channels_list));
    ncp_host_hl_buf_put_u8(&body, secure_rejoin);

    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_rejoin_request", (FMT__0));

  return ret;
}

zb_ret_t ncp_host_zdo_system_server_discovery_request(zb_uint8_t *ncp_tsn, zb_uint16_t server_mask)
{
  zb_ret_t ret = RET_BUSY;
  ncp_host_hl_tx_buf_handle_t body;

  TRACE_MSG(TRACE_TRANSPORT2, ">> ncp_host_zdo_system_server_discovery_request", (FMT__0));

  if (!ncp_host_get_buf_for_nonblocking_request(NCP_HL_ZDO_SYSTEM_SRV_DISCOVERY_REQ, &body, ncp_tsn))
  {
    ncp_host_hl_buf_put_u16(&body, server_mask);

    ret = ncp_host_hl_send_packet(&body);
  }

  TRACE_MSG(TRACE_TRANSPORT2, "<< ncp_host_zdo_system_server_discovery_request", (FMT__0));

  return ret;
}
