/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* PURPOSE: Common service routines from zdo_app.c
   To be used at both NCP Host and Soc as well as in monolythic build.
*/

#define ZB_TRACE_FILE_ID 5
#include "zb_common.h"
#include "zb_scheduler.h"
#include "zb_bufpool.h"
#include "zb_nwk.h"
#include "zb_nwk_nib.h"
#include "zb_aps.h"
#include "zb_zdo.h"
#include "zdo_common.h"
#include "zb_secur.h"
#include "zb_secur_api.h"
#include "zb_nvram.h"
#include "zb_bdb_internal.h"
#include "zb_watchdog.h"
#include "zb_ncp.h"
#include "zdo_wwah_parent_classification.h"

#if defined ZB_ENABLE_ZLL
#include "zll/zb_zll_common.h"
#include "zll/zb_zll_nwk_features.h"
#endif /* defined ZB_ENABLE_ZLL */

#if defined ZB_ENABLE_SE
#include "zb_se.h"
#endif /* defined ZB_ENABLE_SE */

/*! \addtogroup ZB_ZDO */
/*! @{ */



#ifndef ZB_ALIEN_SCHEDULER
void zdo_main_loop()
{
  while (!ZB_SCHEDULER_IS_STOP())
  {
    zb_sched_loop_iteration();
  }
}


void zcl_main_loop()
{
  zdo_main_loop();
}
#endif


/**
   Initiate ZBOSS start without zb_send_no_autostart_signal auto-start
 */
zb_ret_t zb_zdo_start_no_autostart()
{
  zb_zdo_dev_init();
  zb_buf_get_out_delayed(zb_send_no_autostart_signal);
  return RET_OK;
}


void zb_app_signal_pack_with_data(zb_uint8_t param, zb_uint32_t signal_code, zb_int16_t status)
{
  zb_uint8_t *body;

  ZB_ASSERT(param);

  body = zb_buf_alloc_left(param, sizeof(signal_code));

  ZB_MEMCPY(body, &signal_code, sizeof(signal_code));
  zb_buf_set_status(param, (status == RET_OK ? RET_OK : RET_ERROR));
}

void *zb_app_signal_pack(zb_uint8_t param, zb_uint32_t signal_code, zb_int16_t status, zb_uint8_t data_size)
{
  zb_uint8_t *body;

  ZB_ASSERT(param);
  if (signal_code == ZB_ZDO_SIGNAL_DEFAULT_START && data_size == 0)
  {
    /* Default (compatible with previous API) event has no body */
    body = zb_buf_reuse(param);
    zb_buf_set_status(param, (status == RET_OK ? RET_OK : RET_ERROR));
    ZVUNUSED(body);
    /* don't use tracing in this function - it breaks sleepy ED if ZDO trace subsystem is enabled */
    /* TRACE_MSG(TRACE_ZDO1, "zb_app_signal_pack ZB_ZDO_SIGNAL_DEFAULT_START status %d", (FMT__D, zb_buf_get_status(param))); */
    return NULL;
  }
  else
  {
    /* initial alloc creates aligned pointer. Signal_code size is 4 bytes, so signal data is aligned as well. */
    body = zb_buf_initial_alloc(param, sizeof(signal_code) + data_size);
    ZB_MEMCPY(body, &signal_code, sizeof(signal_code));
    zb_buf_set_status(param, (status == RET_OK ? RET_OK : RET_ERROR));
    /* TRACE_MSG(TRACE_ZDO1, "zb_app_signal_pack sig %ld size %d status %d", (FMT__L_D_D, signal_code, data_size, zb_buf_get_status(param))); */
    return (void*)(body + sizeof(signal_code));
  }
}


zb_zdo_app_signal_type_t zb_get_app_signal(zb_uint8_t param, zb_zdo_app_signal_hdr_t **sg_p)
{
  zb_zdo_app_signal_type_t sig_type;

  if (zb_buf_len(param) == 0)
  {
    sig_type = ZB_ZDO_SIGNAL_DEFAULT_START;
    if (sg_p != NULL)
    {
      *sg_p = NULL;
    }
  }
  else if (zb_buf_len(param) < sizeof(zb_zdo_app_signal_hdr_t))
  {
    sig_type = ZB_ZDO_SIGNAL_ERROR;
    if (sg_p != NULL)
    {
      *sg_p = NULL;
    }
  }
  else
  {
    /* Suppress warning about alignment. Buffer begin is aligned here for sure. */
    void *p = zb_buf_begin(param);
    zb_zdo_app_signal_hdr_t *sg = (zb_zdo_app_signal_hdr_t *)p;
    if (sg_p != NULL)
    {
      *sg_p = sg;
    }
    sig_type = (zb_zdo_app_signal_type_t)sg->sig_type;
  }
  return sig_type;
}


/**
   Send ZB_ZDO_SIGNAL_SKIP_STARTUP signal to zdo_startup_complete()
 */
void zb_send_no_autostart_signal(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "zb_send_no_autostart_signal param %hd", (FMT__H, param));
  (void)zb_app_signal_pack(param, ZB_ZDO_SIGNAL_SKIP_STARTUP, RET_OK, 0);
  ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete, param);
  ZB_ZDO_SCHEDULE_UPDATE_LONG_UPTIME();
}

/*! @} */
