/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*  PURPOSE: Generic ZDO commissioning routines.
 Partially moved from zdo_app.c.
*/

#define ZB_TRACE_FILE_ID 65
#include "zb_common.h"
#include "zb_scheduler.h"
#include "zb_bufpool.h"
#include "zb_nwk.h"
#include "zb_nwk_nib.h"
#include "zb_aps.h"
#include "zb_zdo.h"
#include "zb_secur.h"
#include "zb_secur_api.h"
#include "zb_nvram.h"
#include "zb_bdb_internal.h"
#include "zdo_wwah_stubs.h"
#include "zb_address_internal.h"

#if defined ZB_ENABLE_ZLL
#include "zll/zb_zll_common.h"
#include "zll/zb_zll_nwk_features.h"
#endif /* defined ZB_ENABLE_ZLL */

#if defined ZB_HA_SUPPORT_EZ_MODE
#include "ha/zb_ha_ez_mode_comissioning.h"
#endif

#ifdef ZB_SE_COMMISSIONING
#include "zb_se.h"
#endif
#ifdef NCP_MODE
#include "zb_ncp.h"
#endif

#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
void zdo_restart_association(zb_uint8_t param);
#endif

/*
In case of NCP this file is to be compiled for both SoC side, with
only NCP_MODE defined, and Host side with defined both NCP_MODE and NCP_MODE_HOST,
ZB_BDB_MODE and/or ZB_SE_COMMISSIONING.
 */

/**
   Start specific top-level commissioning.

   Called from zdo_dev_continue_start_after_nwk() (main call) and se_initiate_commissioning() (some trick of falling to BDB from SE for mixed SE/BDB build).
 */
void zdo_commissioning_start(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "> zdo_commissioning_start param %hd comm tyoe %d", (FMT__H_D, param, ZB_COMMISSIONING_TYPE()));
#ifdef ZB_CONFIGURABLE_COMMISSIONING_TYPE
  if (ZB_IN_BDB())
  {
    TRACE_MSG(TRACE_ZDO1, "zdo_commissioning_start: force BDB", (FMT__0));
    COMM_CTX().commissioning_type = ZB_COMMISSIONING_BDB;
  }
#endif
  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      bdb_initiate_commissioning(param);
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
      /* make gcc happy */
      /* FALLTHRU */
    case ZB_COMMISSIONING_CLASSIC:
      zdo_classic_initiate_commissioning(param);
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      se_initiate_commissioning(param);
      break;
#endif
#ifdef NCP_MODE
    case ZB_COMMISSIONING_NCP:
      /* NCP is a special case: after NWL reset and PIB load at SoC do nothing. Let Host define. */
      zb_app_signal_pack(param, ZB_SIGNAL_NWK_INIT_DONE, RET_OK, 0);
      ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
      break;
#endif
    default:
      break;
  }
}

#ifdef ZB_JOIN_CLIENT
void zdo_commissioning_join_via_scanlist(zb_uint8_t param)
{
  zb_uint_t i = 0;
  zb_nlme_network_discovery_confirm_t *cnf = NULL;
  zb_nlme_network_descriptor_t *dsc;
  zb_ext_pan_id_t extended_pan_id;
  zb_ext_pan_id_t use_ext_pan_id;

  ZVUNUSED(param);
  if (COMM_CTX().discovery_ctx.scanlist_ref != 0)
  {
    cnf = (zb_nlme_network_discovery_confirm_t *)zb_buf_begin(COMM_CTX().discovery_ctx.scanlist_ref);
    dsc = (zb_nlme_network_descriptor_t *)(cnf + 1);

#ifdef ZB_TRACE_LEVEL
    if (TRACE_ENABLED(TRACE_ZDO1))
    {
      for (i = 0 ; i < cnf->network_count ; ++i)
      {
        zb_ext_pan_id_t extended_pan_id;

        zb_address_get_pan_id(dsc->panid_ref, extended_pan_id);

        TRACE_MSG(TRACE_ZDO1,
                  "net %hd: xpanid " TRACE_FORMAT_64 ", channel %hd, stack profile %hd, permit_joining %hd, router_cap %hd, ed_cap %hd",
                  (FMT__H_A_H_H_H_H_H,
                   i, TRACE_ARG_64(extended_pan_id),
                   (zb_uint8_t)dsc->logical_channel, (zb_uint8_t)dsc->stack_profile, (zb_uint8_t)dsc->permit_joining,
                   (zb_uint8_t)dsc->router_capacity, (zb_uint8_t)dsc->end_device_capacity));
        dsc++;
      }

      /* Be polite - restore dsc pointer */
      dsc = (zb_nlme_network_descriptor_t *)(cnf + 1);
    }
#endif

    /* Now join thru Association */
    for (i = 0 ; i < cnf->network_count ; ++i)
    {
      if (!dsc->processed)
      {
        if (COMM_CTX().discovery_ctx.scanlist_join_attempt_n < ZB_ZDO_MAX_JOINING_TRY_ATTEMPTS)
        {
          zb_get_use_extended_pan_id(use_ext_pan_id);
          zb_address_get_pan_id(dsc->panid_ref, extended_pan_id);

          if (
            (ZB_EXTPANID_IS_ZERO(use_ext_pan_id)
             || ZB_EXTPANID_CMP(extended_pan_id, use_ext_pan_id)))
          {
            /*
              Now join to the first network or network with desired ext pan id.
            */
#ifdef ZB_NWK_BLACKLIST
            if (zb_nwk_blacklist_is_full())
            {
              zb_nwk_blacklist_reset();
            }
            /* Add the network into the blacklist only once at the first
             * joining attempt */
            if (COMM_CTX().discovery_ctx.scanlist_join_attempt_n == 0)
            {
              zb_nwk_blacklist_add(extended_pan_id);
            }
#endif
            COMM_CTX().discovery_ctx.scanlist_idx = i;
            COMM_CTX().discovery_ctx.scanlist_join_attempt_n++;
            TRACE_MSG(TRACE_ZDO3, "scanlist_idx %d scanlist_join_attempt_n %hd",
                      (FMT__D_H, COMM_CTX().discovery_ctx.scanlist_idx, COMM_CTX().discovery_ctx.scanlist_join_attempt_n));
            zb_buf_get_out_delayed(zdo_join_to_nwk_descr);
            break;
          }
        }
        else
        {
          COMM_CTX().discovery_ctx.scanlist_join_attempt_n = 0;
          dsc->processed = 1;
        }
      }
      dsc++;
    } /* for */
  }
  if (!COMM_CTX().discovery_ctx.scanlist_ref
      || ((cnf != NULL) && (i != 0) && (i == cnf->network_count)))
  {
    ZB_SCHEDULE_CALLBACK(zdo_commissioning_nwk_discovery_failed, 0);
  }
}


/*
  We are here when we could not find a net to join
 */
void zdo_commissioning_nwk_discovery_failed(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ERROR, "Can't find PAN to join to! param %hd", (FMT__H, param));

  /* Use scanlist ref for our internal purposes */
  if (!param)
  {
    param = COMM_CTX().discovery_ctx.scanlist_ref;
  }
  if (!param)
  {
    zb_buf_get_out_delayed(zdo_commissioning_nwk_discovery_failed);
    return;
  }
  zdo_reset_scanlist(ZB_FALSE);
#ifdef ZB_NWK_BLACKLIST
  zb_nwk_blacklist_reset();
#endif

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      TRACE_MSG(TRACE_ZDO2, "nwk discovery failed, let BDB decide ...", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_DISCOVERY_FAILED, param);
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      TRACE_MSG(TRACE_ZDO2, "Calling zb_zdo_startup_complete_int to indicate nwk discovery failure", (FMT__0));
      zb_app_signal_pack(param, ZB_ZDO_SIGNAL_DEFAULT_START, ZB_NWK_STATUS_NOT_PERMITTED, 0);
      ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      se_commissioning_signal(SE_COMM_SIGNAL_NWK_DISCOVERY_FAILED, param);
      break;
#endif
    default:
      break;
  }
}

/*
  We are here when join/rejoin attempt is failed.
 */
void zdo_commissioning_join_failed(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "{re}join failed", (FMT__0));

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      TRACE_MSG(TRACE_ZDO1, "{re}join failed - call BDB machine", (FMT__0));
      /* Note: there was a check for BDB * NFN and call to bdb machine in that case only.
         See originag zb_nlme_join_confirm() code.
         TODO: in BDB machine implement rejoin re-attempt vs association */
      /* NK: Looks like this is already covered in BDB machine - if device is NFN, control will be
       * passed to application (signal with status NO_NETWORK). Not sure if we need to implement
       * some additional logic in BDB machine itself - application may decide if it wants to do
       * rejoin backoff (for example), or siletly go to sleep, or reset to factory defaults
       * etc. */

      if (COMM_CTX().discovery_ctx.scanlist_ref)
      {
        zb_buf_free(param);
        ZB_SCHEDULE_CALLBACK(zdo_retry_joining, 0);
      }
      else
      {
        bdb_commissioning_signal(BDB_COMM_SIGNAL_NWK_JOIN_FAILED, param);
      }
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      if (ZG->zdo.handle.rejoin && ZB_AIB().aps_insecure_join && !ZB_AIB().always_rejoin)
      {
        /* FIXME: Do we really need to mix rejoin - Association - rejoin - Association - ... ?
           Also - if we dropped to association, looks like we need to erase network data before. It
           is really strange case when we do unsuccessful association attempt (no more children
           available), and then successful rejoin attempt to the same network.
         */
        TRACE_MSG(TRACE_ZDO1, "rejoin failed - try nwk search & association", (FMT__0));
        ZB_SCHEDULE_CALLBACK(zdo_restart_association, param);
      }
      else
      {
        if (COMM_CTX().discovery_ctx.scanlist_ref)
        {
          zb_buf_free(param);
          TRACE_MSG(TRACE_ZDO1, "{re}join failed", (FMT__0));
          ZB_SCHEDULE_CALLBACK(zdo_retry_joining, 0);
        }
        else
        {
          TRACE_MSG(TRACE_ZDO1, "{re}join totally failed - inform user", (FMT__0));
          zb_app_signal_pack(param, ZB_ZDO_SIGNAL_DEFAULT_START, ZB_NWK_STATUS_NOT_PERMITTED, 0);
          ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
        }
      }
      break;
#endif

#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      se_commissioning_signal(SE_COMM_SIGNAL_NWK_JOIN_FAILED, param);
      break;
#endif
#if defined NCP_MODE && !defined NCP_MODE_HOST
    case ZB_COMMISSIONING_NCP:
      if (ZG->zdo.handle.rejoin != 0)
      {
        ncp_signal(NCP_SIGNAL_NWK_REJOIN_FAILED, param);
      }
      else
      {
        ncp_signal(NCP_SIGNAL_NWK_JOIN_FAILED, param);
      }
      break;
#endif /* defined NCP_MODE && !defined NCP_MODE_HOST */
    default:
      break;
  } /* switch */
}


void zdo_commissioning_authentication_failed(zb_uint8_t param)
{
  switch (ZB_COMMISSIONING_TYPE())
  {
    /* In BDB, or optionally classic r21, initiate TCLK generation using APS commands */
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      TRACE_MSG(TRACE_ZDO1, "Let BDB handle authentication failure", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NWK_AUTH_FAILED, param);
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
/*
  2.1 Standard Security Mode

  If the joining device
  did not receive the key via the APSME-TRANSPORT-KEY.indication within the
  apsSecurityTimeOutPeriod since receiving the NLME-JOIN.confirm primitive, it
  shall reset and may choose to start the joining procedure again.

  In

  11.29 TP/SEC/BV-29 -I Security NWK Remove II-ZR
  Objective: DUT as ZR removes
  (DUT ZED2 performs silent leave).

  So it is unclear: do we need to just leave or join again?
  Let's leave to satisfy TC.
*/
/* NK: If we are in rejoin procedure, do "silent" rejoin, else - do "silent" leave */
      if (ZG->zdo.handle.rejoin
#ifdef ZB_CERTIFICATION_HACKS
          && !ZB_CERT_HACKS().disable_cyclic_tc_rejoin
#endif
        )
      {
#ifdef ZB_REJOIN_BACKOFF
        if (zb_zdo_rejoin_backoff_is_running())
        {
          /* Go to new rejoin backoff iteration */
          zb_zdo_rejoin_backoff_iteration_done();
          ZB_SCHEDULE_CALLBACK(zb_zdo_rejoin_backoff_continue, 0);
          zb_buf_free(param);
        }
        else
#endif
        {
          zb_ext_pan_id_t ext_pan_id;
          zb_get_extended_pan_id(ext_pan_id);
          /* force unsecure (AKA Trust Center) rejoin to current pan */
          zdo_initiate_rejoin(param, ext_pan_id,
                              ZB_AIB().aps_channel_mask_list, ZB_FALSE);
        }
      }
      else
      {
        TRACE_MSG(TRACE_ERROR, "zdo_authentication_failed - Leave", (FMT__0));
#ifndef ZB_MULTILEAVE_HACK
        zdo_commissioning_leave(param, ZB_FALSE, ZB_FALSE);
        ZB_ZDO_SET_UNAUTH();      /* Call unauth_delayed after leave */
#else
        ZB_JOINED() = 0;
        zb_buf_free(param);
        zb_buf_get_out_delayed_ext(zb_zdo_startup_complete_int_delayed, ZB_NWK_STATUS_NO_KEY);
#endif
      }
      break;
#endif  /* classic */
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      TRACE_MSG(TRACE_SECUR3, "SE - authentication failed", (FMT__0));
      se_commissioning_signal(SE_COMM_SIGNAL_NWK_AUTH_FAILED, param);
      break;
#endif
#if defined NCP_MODE && !defined NCP_MODE_HOST
    case ZB_COMMISSIONING_NCP:
      ncp_signal(NCP_SIGNAL_NWK_JOIN_AUTH_FAILED, param);
      break;
#endif /* defined NCP_MODE && !defined NCP_MODE_HOST */
    default:
      break;
  }
}

#endif  /* ZB_JOIN_CLIENT */


zb_bool_t zb_distributed_security_enabled()
{
#ifdef ZB_SE_COMMISSIONING
  return (zb_bool_t)(ZB_COMMISSIONING_TYPE() != ZB_COMMISSIONING_SE);
#else
  return ZB_TRUE;
#endif
}


#ifdef ZB_JOIN_CLIENT
#if defined ZB_DISTRIBUTED_SECURITY_ON

zb_bool_t zb_tc_is_distributed(void)
{
  return zb_aib_trust_center_address_is_unknown();
}

void zb_sync_distributed(void)
{
  zb_aib_tcpol_set_is_distributed_security(zb_tc_is_distributed());
}

#endif  /* ZB_DISTRIBUTED_SECURITY_ON */


void zdo_commissioning_leave(zb_bufid_t buf, zb_bool_t rejoin, zb_bool_t remove_children)
{
  zb_zdo_mgmt_leave_param_t *req = ZB_BUF_GET_PARAM(buf, zb_zdo_mgmt_leave_param_t);

  TRACE_MSG(TRACE_ZDO2, ">> zdo_commissioning_leave, buf %d, rejoin %d, remove_children %d",
            (FMT__D_D, buf, rejoin, remove_children));

  zb_get_long_address(req->device_address);

  req->dst_addr = zb_get_short_address();
  req->rejoin = rejoin;
  req->remove_children = remove_children;

  zdo_mgmt_leave_req(buf, NULL);

  TRACE_MSG(TRACE_ZDO2, "<< zdo_commissioning_leave", (FMT__0));
}


void zdo_commissioning_handle_dev_annce_sent_event(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO2, ">> zdo_commissioning_handle_dev_annce_sent_event, param %d",
            (FMT__D, param));

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      /* If no r21 mode (BDB or just r21), indicate startup complete only
       * after TCLK update and verify. */
      /* VP: in r21 mode startup will be indicated after node desc response (or tclk exchange if required by device)
         See test TP-R21-BV-10: device shall ask node descriptor from TC after join. */
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NWK_JOIN_DONE, param);
      param = 0;
      break;
#endif /* ZB_BDB_MODE */
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      /* TODO: Implement me correctly!!! */
      if ( !ZB_NIB_SECURITY_LEVEL() ||
           (zb_secur_get_link_key_by_address(ZB_AIB().trust_center_address, ZB_SECUR_ANY_KEY_ATTR)
            && ZG->zdo.handle.rejoin) )
      {
        /* When joins to unsecured network and after secured rejoin - indicate startup immediately;
         * suppose that after secured rejoin we have VERIFIED key (otherwise device will try to obtain new link key with TC). */

        /* TODO: This reboot is also not so well defined  */
        zb_app_signal_pack(param, ZB_SIGNAL_DEVICE_REBOOT, zb_buf_get_status(param), 0);
        ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
        param = 0;
      }
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      /* TODO: Implement me! */
      if ( !ZB_NIB_SECURITY_LEVEL() ||
           (zb_secur_get_link_key_by_address(ZB_AIB().trust_center_address, ZB_SECUR_ANY_KEY_ATTR)
            && ZG->zdo.handle.rejoin))
      {
        zb_app_signal_pack(param, ZB_SIGNAL_DEVICE_REBOOT, zb_buf_get_status(param), 0);
        ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
        param = 0;

        se_commissioning_signal_post(0, SE_COMM_SIGNAL_REJOIN_RECOVERY_OK);
      }

      break;
#endif  /* ZB_SE_COMMISSIONING */
#if defined NCP_MODE && !defined NCP_MODE_HOST
    case ZB_COMMISSIONING_NCP:
      if (ZG->zdo.handle.rejoin != 0)
      {
        ncp_signal(NCP_SIGNAL_NWK_REJOIN_DONE, param);
      }
      else
      {
        ncp_signal(NCP_SIGNAL_NWK_JOIN_DONE, param);
      }
      param = 0;
      break;
#endif /* defined NCP_MODE && !defined NCP_MODE_HOST */
    default:
      TRACE_MSG(TRACE_ZDO1, "Unknown commissioning type", (FMT__0));
      ZB_ASSERT(0);
      break;
  }

  if (param != 0)
  {
    zb_buf_free(param);
  }

  TRACE_MSG(TRACE_ZDO2, "<< zdo_commissioning_handle_dev_annce_sent_event", (FMT__0));
}

void zdo_commissioning_start_router_confirm(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, ">>zdo_commissioning_start_router_confirm %hd", (FMT__H, param));

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NWK_START_ROUTER_CONF, param);
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      if (ZB_NIB_SECURITY_LEVEL() == 0
#ifdef ZB_JOIN_CLIENT
          || !zdo_secur_waiting_for_tclk_update()
#endif
          || ZB_IS_DEVICE_ZC())
      {
        TRACE_MSG(TRACE_ZDO1, "calling zdo_startup_complete", (FMT__0));
        zb_app_signal_pack(param, ZB_ZDO_SIGNAL_DEFAULT_START, zb_buf_get_status(param), 0);
        /* Router start is last step in join. */
        ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
      }
      else
      {
        zb_buf_free(param);
      }
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      se_commissioning_signal(SE_COMM_SIGNAL_NWK_START_ROUTER_CONF, param);
      break;
#endif
#if defined NCP_MODE && !defined NCP_MODE_HOST
    case ZB_COMMISSIONING_NCP:
      if (ZG->zdo.handle.rejoin != 0)
      {
        ncp_signal(NCP_SIGNAL_NWK_REJOIN_DONE, param);
      }
      else
      {
        ncp_signal(NCP_SIGNAL_NWK_JOIN_DONE, param);
      }
      break;
#endif

    default:
      break;
  }
}


void zdo_commissioning_tclk_update_complete(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO3, ">> zdo_commissioning_tclk_update_complete, param %d", (FMT__D, param));

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_TCLK_DONE, param);
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      (void)zb_app_signal_pack(param, ZB_ZDO_SIGNAL_DEFAULT_START, RET_OK, 0);
      ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      ZB_ASSERT(0);
      zb_buf_free(param);
      break;
#endif
#if defined NCP_MODE && !defined NCP_MODE_HOST
    case ZB_COMMISSIONING_NCP:
      ncp_signal(NCP_SIGNAL_SECUR_TCLK_EXCHANGE_COMPLETED, param);
      break;
#endif
    default:
      ZB_ASSERT(0);
      zb_buf_free(param);
      break;
  }

  TRACE_MSG(TRACE_ZDO3, "<< zdo_commissioning_tclk_update_complete", (FMT__0));
}


void zdo_commissioning_tclk_update_failed(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO3, ">> zdo_commissioning_tclk_update_failed, param %d", (FMT__D, param));

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_TCLK_EX_FAILURE, param);
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      zdo_authentication_failed(param);
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      ZB_ASSERT(0);
      zb_buf_free(param);
      break;
#endif
#if defined NCP_MODE && !defined NCP_MODE_HOST
    case ZB_COMMISSIONING_NCP:
      ncp_signal(NCP_SIGNAL_SECUR_TCLK_EXCHANGE_FAILED, param);
      break;
#endif
    default:
      break;
  }

  TRACE_MSG(TRACE_ZDO3, "<< zdo_commissioning_tclk_update_failed", (FMT__0));
}


#endif  /* ZB_JOIN_CLIENT */


#ifndef NCP_MODE_HOST

#ifdef ZB_JOIN_CLIENT

static void zdo_commissioning_authenticated_cont(void)
{
  switch (ZB_COMMISSIONING_TYPE())
  {
    /* In BDB, or optionally classic r21, initiate TCLK generation using APS commands */
#ifdef ZB_BDB_MODE
#ifndef ZB_LITE_BDB_ONLY_COMMISSIONING
    case ZB_COMMISSIONING_BDB:
      /* We want TCLK exchange working on SoC */
      /* TODO: decide if we need to do TCLK exchange based on parameters set by the host
         or initiate it from the host using the special NCP command */
#if defined ZB_NCP_SOC_BDB
    case ZB_COMMISSIONING_NCP:
      /* TODO: Move this part to BDB commissioning machine: have some variations depending of the
       * state when this signal is generated (initialization/TC rejoin/steering/etc) */
#endif /* ZB_NCP_SOC_BDB */
      if (!IS_DISTRIBUTED_SECURITY()
          && ZB_TCPOL().update_trust_center_link_keys_required
          && !ZB_TCPOL().waiting_for_tclk_exchange) /* TCLK is not already in progress */
      {
        zb_uint16_t aps_key_idx = (zb_uint16_t)-1;

        if (!ZB_IEEE_ADDR_IS_ZERO(ZB_AIB().trust_center_address))
        {
          aps_key_idx = zb_aps_keypair_get_index_by_addr(ZB_AIB().trust_center_address, ZB_SECUR_VERIFIED_KEY);
        }
        if (aps_key_idx == (zb_uint16_t)-1)
        {
          TRACE_MSG(TRACE_SECUR3, "BDB & !distributed - get TCLK over APS", (FMT__0));
          zdo_initiate_tclk_gen_over_aps(0);
        }
      }
      break;
#endif
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      if (!IS_DISTRIBUTED_SECURITY()
          && ZB_TCPOL().update_trust_center_link_keys_required)
      {
        TRACE_MSG(TRACE_SECUR3, "!BDB & !distributed & update_trust_center_link_keys_required - get TCLK over APS", (FMT__0));
        zdo_initiate_tclk_gen_over_aps(0);
      }
      else if (ZB_IS_DEVICE_ZED())
      {
        /* If not bdb and I am ZED (no need to start router), I am is started. */
        TRACE_MSG(TRACE_SECUR3, "!BDB ZED - complete start", (FMT__0));
        zb_buf_get_out_delayed_ext(zb_zdo_startup_complete_int_delayed, RET_OK, 0);
      }
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      TRACE_MSG(TRACE_SECUR3, "SE - authenticated - get TCLK over CBKE", (FMT__0));
      se_commissioning_signal(SE_COMM_SIGNAL_NWK_AUTH_OK, 0);
      break;
#endif
    default:
      break;
  }
}

/**
   Reaction on authentication complete - that is, on NWK key recv.

   Next actions depend on mode: BDB/classic/SE/ZLL
 */
void zdo_commissioning_authenticated(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO3, "scheduling device_annce %hd", (FMT__H, param));

  /* According to BDB send Devive_annce immediately after receiving NWK key. */
  /* Call Device_annce before node desc req (according to BDB Figure 13 Trust Center link key exchange procedure sequence chart) */
  if (param != 0)
  {
    ZB_SCHEDULE_CALLBACK(zdo_authenticated_send_device_annce, param);
  }

  zdo_commissioning_authenticated_cont();
}

/*
  Commissioning logic after device started
 */
void zdo_commissioning_dev_annce_sent(zb_uint8_t param)
{
  zb_ret_t status = zb_buf_get_status(param);

  (void)zb_buf_reuse(param);
  /* Indicate startup complete */
  zb_buf_set_status(param, status);

  TRACE_MSG(TRACE_ZDO1, ">> zdo_commissioning_dev_annce_sent param %hd",
            (FMT__H, param));

  TRACE_MSG(TRACE_ZDO3, "sent device_annce, start compl, st %hd  dev type %hd",
            (FMT__H_H, status, zb_get_device_type()));

  /* moved from apsde_data_conf */
#ifdef ZB_USE_NVRAM
  zb_nvram_write_dataset(ZB_NVRAM_COMMON_DATA);
#endif

  if (status != RET_OK)
  {
    /* rejoin to current PAN disregarding current commissioning type */
    ZB_SCHEDULE_CALLBACK(zdo_commissioning_initiate_rejoin, param);
    param = 0;
  }
  else
  {
#if defined ZB_ROUTER_ROLE
    /* FIXME: handle start router in bdb? */
    if (ZB_IS_DEVICE_ZR())
    {
      ZB_SCHEDULE_CALLBACK(zb_zdo_start_router, param);
      param=0;
    }
    else
#endif /* ZB_ROUTER_ROLE */
    {
      zdo_commissioning_handle_dev_annce_sent_event(param);
      param = 0;
    }

    /* Regardless commissioning mode - it's common for all: schedule end device
     * timeout request */

#ifdef ZB_ED_FUNC
#ifndef ZB_LITE_NO_ED_AGING_REQ
    if (ZB_IS_DEVICE_ZED()
#ifdef ZB_CERTIFICATION_HACKS
    /* Used for test purposes in tp_r21_bv-19 for gZED */
        && (!ZB_CERT_HACKS().disable_end_device_timeout_request_at_join)
#endif
      )
    {
      /* send ED timeout request cmd with delay to make sure
         the coordinator have handled previously sent Dev_annce cmd
      */
      /* We do not know our ED timeout mode until resp to ED Timeout Req from ZC/ZR  */
      ZB_SET_KEEPALIVE_MODE(ED_KEEPALIVE_DISABLED);
      ZB_SCHEDULE_ALARM(zb_nwk_ed_send_timeout_req, 0, ZB_ZDO_SEND_ED_TIMEOUT_REQ_DELAY);
    }
#endif
#endif  /* #ifdef ZB_ED_FUNC */
  }

  if (param != 0)
  {
    zb_buf_free(param);
  }
}

#endif /* ZB_JOIN_CLIENT */

#ifdef ZB_ROUTER_ROLE

#ifndef ZB_COORDINATOR_ONLY
void zdo_comm_set_permit_join_after_router_start(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, ">> zdo_commissioning_start_router_confirm %hd", (FMT__H, param));

  zdo_comm_set_permit_join(param, zdo_commissioning_start_router_confirm);

  TRACE_MSG(TRACE_ZDO1, "<< zdo_commissioning_start_router_confirm", (FMT__0));
}
#endif /* #ifndef ZB_COORDINATOR_ONLY */

void zdo_comm_set_permit_join(zb_uint8_t param, zb_callback_t cb)
{
  zb_zdo_mgmt_permit_joining_req_param_t *request;

  request = ZB_BUF_GET_PARAM(param, zb_zdo_mgmt_permit_joining_req_param_t);
  request->dest_addr = ZB_PIBCACHE_NETWORK_ADDRESS();
  request->permit_duration = zb_commissioning_default_permit_duration();

  TRACE_MSG(TRACE_ZDO2, "set permit_duration %d", (FMT__D, request->permit_duration));
  /* Go via zdo mgmt to use cb */
  zb_zdo_mgmt_permit_joining_req(param, cb);
}


zb_int_t zb_commissioning_default_permit_duration(void)
{
  zb_int_t ret = 0;

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      if (ZB_NIB().max_children != 0)
      {
        ret = ZDO_CTX().conf_attr.permit_join_duration;
      }
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      /* No permit join after start in SE */
      ret = 0;
      break;
#endif
    default:
      break;
  }
  return ret;
}

#endif  /* #ifdef ZB_ROUTER_ROLE */


#ifdef ZB_JOIN_CLIENT

void zdo_commissioning_secur_failed(zb_uint8_t param, zb_uint16_t status)
{
#ifndef ZB_COMMISSIONING_CLASSIC_SUPPORT
  ZVUNUSED(status);
#endif

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      TRACE_MSG(TRACE_SECUR1, "Failed TCLK verification - let BDB decide", (FMT__0));
      /* ZB_SCHEDULE_CALLBACK(bdb_commissioning_machine, param); */
      bdb_update_tclk_failed(param);
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      TRACE_MSG(TRACE_APS1, "report error status %hd, zb_app_signal_pack", (FMT__H, status));
      (void)zb_app_signal_pack(param, ZB_ZDO_SIGNAL_DEFAULT_START, status, 0);
      ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      TRACE_MSG(TRACE_SECUR1, "Failed TCLK verification in SE mode - can't be here!", (FMT__0));
      se_commissioning_signal(SE_COMM_SIGNAL_BAD, param);
      break;
#endif
#ifdef NCP_MODE
    case ZB_COMMISSIONING_NCP:
      ncp_signal(NCP_SIGNAL_SECUR_TCLK_EXCHANGE_FAILED, param);
      break;
#endif
    default:
      break;
  }
}


#ifndef ZB_LITE_BDB_ONLY_COMMISSIONING
static void zdo_inform_app_leave(zb_uint8_t leave_type)
{
  /* Notify user about leave with rejoin */
#ifndef ZB_LITE_NO_OLD_CB
  if (ZDO_CTX().app_leave_cb != NULL)
  {
    (ZDO_CTX().app_leave_cb)(leave_type);
  }
  else
#endif  /* ZB_LITE_NO_OLD_CB */
  {
    zb_buf_get_out_delayed_ext(zb_send_leave_signal, leave_type, 0);
  }
}

#endif  /* ZB_LITE_BDB_ONLY_COMMISSIONING */


/*
  Commissioning logic upon reception of Leave Request with rejoin = 1
 */
void zdo_commissioning_leave_with_rejoin(zb_uint8_t param)
{
  ZVUNUSED(param);
  TRACE_MSG(TRACE_ZDO1, ">> zdo_commissioning_leave_with_rejoin param %hd",
            (FMT__H, param));

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      /* [MM]: 01/15/2018: Left this logic unchanged after
       * refactoring. However, explained above is tricky and is required to be
       * checked */

      /* VP: currently we have both bdb join/rejoin and old-style join/rejoin logic and they are mixed.
         In some case BDB device receives nwk leave (with rejoin in our case), then performing usual rejoin (without using BDB machine),
         after join failure/success stack uses BDB commissioning machine to complete join process. This leads to calling bdb_commissioning_machine
         without reset it's logic - then stacl silently completes process without notifying appliation or further join attempts.
         Temporarily fix: use BDB logic for rejoin on receiving leave (locally or remotely).
       */
      zdo_inform_app_leave(ZB_NWK_LEAVE_TYPE_REJOIN);
      ZB_SCHEDULE_CALLBACK(zdo_commissioning_initiate_rejoin, param);
      break;
#endif  /* ZB_BDB_MODE */

#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
#ifdef ZB_USE_NVRAM
      zb_nvram_write_dataset(ZB_NVRAM_COMMON_DATA);
#endif
#ifdef ZB_NWK_BLACKLIST
      zb_nwk_blacklist_reset();
#endif
      zdo_inform_app_leave(ZB_NWK_LEAVE_TYPE_REJOIN);
#ifdef ZB_NSNG
      /* A hack because of unperfect NSNG CSMA/CA mechanism, hack for CCB2255 */
      ZB_SCHEDULE_ALARM(zb_nwk_do_rejoin_after_leave, param, ZB_MILLISECONDS_TO_BEACON_INTERVAL(500));
#else
      zb_nwk_do_rejoin_after_leave(param);
#endif
      break;
#endif  /* ZB_COMMISSIONING_CLASSIC_SUPPORT */

#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
#ifdef ZB_USE_NVRAM
      zb_nvram_write_dataset(ZB_NVRAM_COMMON_DATA);
#endif
      zdo_inform_app_leave(ZB_NWK_LEAVE_TYPE_REJOIN);
      se_commissioning_signal(SE_COMM_SIGNAL_ENTER_REJOIN_ONLY, param);
      break;
#endif

#ifdef NCP_MODE
    case ZB_COMMISSIONING_NCP:
#ifdef ZB_USE_NVRAM
      zb_nvram_write_dataset(ZB_NVRAM_COMMON_DATA);
#endif
#ifdef ZB_NWK_BLACKLIST
      zb_nwk_blacklist_reset();
#endif
      zdo_inform_app_leave(ZB_NWK_LEAVE_TYPE_REJOIN);
#ifdef ZB_NSNG
      /* A hack because of unperfect NSNG CSMA/CA mechanism, hack for CCB2255 */
      ZB_SCHEDULE_ALARM(zb_nwk_do_rejoin_after_leave, param, ZB_MILLISECONDS_TO_BEACON_INTERVAL(500));
#else
      zb_nwk_do_rejoin_after_leave(param);
#endif
      break;
#endif /* NCP_MODE */

    default:
      TRACE_MSG(TRACE_ERROR, "Unknown commissioining type", (FMT__0));
      ZB_ASSERT(0);
      break;
  }


  TRACE_MSG(TRACE_ZDO1, "<< zdo_commissioning_leave_with_rejoin",
            (FMT__0));
}

void zb_secur_rejoin_after_security_failure(zb_uint8_t param)
{
  TRACE_MSG(TRACE_SECUR1, "zb_secur_rejoin_after_security_failure: param %hd;", (FMT__H, param));

  switch (ZB_COMMISSIONING_TYPE())
  {
#if defined ZB_BDB_MODE || defined ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_BDB:
    case ZB_COMMISSIONING_CLASSIC:
    {
      zb_ext_pan_id_t ext_pan_id;
      /* rejoin to current pan */
      zb_get_extended_pan_id(ext_pan_id);
      zdo_initiate_rejoin(param, ext_pan_id,
                          ZB_AIB().aps_channel_mask_list, ZB_FALSE);
      break;
    }
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      /* ZG->aps.authenticated = 0; */
      se_commissioning_signal(SE_COMM_SIGNAL_ENTER_REJOIN_RECOVERY, param);
      break;
#endif
      /* TODO: inform NCP Host */
    default:
      break;
  }
}

void zdo_commissioning_initiate_rejoin(zb_uint8_t param)
{
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
  zb_ext_pan_id_t ext_pan_id;
#endif

  TRACE_MSG(TRACE_ZDO1, "zdo_commissioning_initiate_rejoin param %hd", (FMT__H, param));

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      zb_buf_free(param);
      if (!ZG->zdo.handle.rejoin)
      {
        zdo_rejoin_clear_prev_join();
        ZB_BDB().bdb_force_router_rejoin = ZB_TRUE;
        bdb_start_top_level_commissioning(ZB_BDB_NETWORK_STEERING);
      }
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
#ifdef ZB_NWK_BLACKLIST
      zb_nwk_blacklist_reset();
#endif
      zb_get_extended_pan_id(ext_pan_id);
      zdo_initiate_rejoin(param, ext_pan_id,
                          ZB_AIB().aps_channel_mask_list, zb_zdo_authenticated());
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      se_commissioning_signal(SE_COMM_SIGNAL_ENTER_REJOIN_RECOVERY, param);
      break;
#endif
    default:
      break;
  }
}

void zdo_commissioning_leave_done(zb_uint8_t param)
{
  TRACE_MSG(TRACE_NWK1, ">>zb_zdo_clear_after_leave_cont", (FMT__0));

#if defined ZB_ENABLE_ZLL
  if (    ZLL_TRAN_CTX().transaction_id
          &&  (  ZLL_TRAN_CTX().transaction_task == ZB_ZLL_TRANSACTION_NWK_START_TASK_TGT
                 || ZLL_TRAN_CTX().transaction_task == ZB_ZLL_TRANSACTION_NWK_START_TASK
                 || ZLL_TRAN_CTX().transaction_task == ZB_ZLL_TRANSACTION_JOIN_ROUTER_TASK
                 || ZLL_TRAN_CTX().transaction_task == ZB_ZLL_TRANSACTION_JOIN_ROUTER_TASK_TGT
                 || ZLL_TRAN_CTX().transaction_task == ZB_ZLL_TRANSACTION_JOIN_ED_TASK
                 || ZLL_TRAN_CTX().transaction_task == ZB_ZLL_TRANSACTION_JOIN_ED_TASK_TGT))
  {
    ZG->nwk.leave_context.rejoin_after_leave = ZB_FALSE;
    zb_buf_set_status(param, RET_OK);
    ZB_SCHEDULE_CALLBACK(zll_leave_nwk_confirm, param);
  }
  else
#endif /* defined ZB_ENABLE_ZLL */
  {
    /* Not need anymore */
    zb_buf_free(param);
#ifdef ZB_USE_NVRAM
    zb_nvram_clear();
#endif
    /* Clearing addr map table */
    zb_address_reset();

    switch (ZB_COMMISSIONING_TYPE())
    {
#ifdef ZB_BDB_MODE
      case ZB_COMMISSIONING_BDB:
        /* In r21 all case changed to  bdb_commissioning_leave_done().
           But it is not 100% equivalent.
           There is no zdo_retry_joining() call but is zb_zdo_reset_inform_user_app() call.
           Let's leave r22 behavior for now, but not sure it has no problems which fixed in r21.
        */

        /* VP: Moved from zdo (zb_nlme_reset_confirm) to prevent races between reset_confirm_cb
         * and bdb_commissioning_machine.
         */
        /* FIXME: not very good check, but anyway. We need to differentiate 2 cases:
           1st case: join is not complete
           2nd case: join is complete and control was already passed to application
           If join is not complete (key exchange failed etc), schedule bdb iterations etc.
           If join is already completed, application may already initiate new steering etc, so do not
           touch bdb at all.
        */
        /* TODO: Send special ZDO signal (LEAVE_DONE or so) - it will allow to avoid manual join
         * restarting after leave (from application). */
        if (ZB_BDB().bdb_commissioning_step == ZB_BDB_COMMISSIONING_STOP)
        {
          TRACE_MSG(TRACE_NWK1, "BDB commissioning machine is already finished, do nothing", (FMT__0));
        }
        else
        {
          TRACE_MSG(TRACE_NWK1, "Scheduling BDB after leave step %d status %d", (FMT__D_D, ZB_BDB().bdb_commissioning_step, ZB_BDB().bdb_commissioning_status));
          if (ZB_BDB().bdb_commissioning_status != ZB_BDB_STATUS_NO_NETWORK)
          {
            /* Probably bdb device failed to update tclk exchange */
            bdb_commissioning_signal(BDB_COMM_SIGNAL_LEAVE_DONE, 0);
          }
          else
          {
            /* Will be here after bdb device failed authentication - lets join to
             * other networks */
            ZB_SCHEDULE_CALLBACK(zdo_retry_joining, 0);
          }
        }
        break;
#endif  /* ZB_BDB_MODE */

#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
      case ZB_COMMISSIONING_CLASSIC:
        if (ZB_ZDO_GET_UNAUTH())
        {
          ZB_ZDO_CLEAR_UNAUTH();
          zb_buf_get_out_delayed_ext(zb_zdo_startup_complete_int_delayed, ZB_NWK_STATUS_NO_KEY, 0);
        }
        break;
#endif
#ifdef ZB_SE_COMMISSIONING
      case ZB_COMMISSIONING_SE:
        TRACE_MSG(TRACE_SECUR3, "SE - leave completed", (FMT__0));
        se_commissioning_signal(SE_COMM_SIGNAL_NWK_LEAVE, 0);
        break;
#endif
      default:
        break;
    }
  }

#ifndef ZB_LITE_NO_OLD_CB
  /* FIXME: remove leave signel here?? Note: commented out in r21. */
  if (ZDO_CTX().app_leave_cb != NULL)
  {
    (ZDO_CTX().app_leave_cb)(ZB_NWK_LEAVE_TYPE_RESET);
  }
  else
#endif
  {
    zb_buf_get_out_delayed_ext(zb_send_leave_signal, ZB_NWK_LEAVE_TYPE_RESET, 0);
  }

  TRACE_MSG(TRACE_NWK1, "<<zb_zdo_clear_after_leave_cont", (FMT__0));
}

#endif /* ZB_JOIN_CLIENT */

#if defined ZB_SECURITY_INSTALLCODES && !defined ZB_SECURITY_INSTALLCODES_ONLY
zb_bool_t zdo_secur_must_use_installcode(zb_bool_t is_client)
{
#ifndef ZB_SE_BDB_MIXED
  ZVUNUSED(is_client);
#endif
  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_SECURITY_INSTALLCODES
    case ZB_COMMISSIONING_BDB:
    case ZB_COMMISSIONING_CLASSIC:
    case ZB_COMMISSIONING_NCP:
      return (zb_bool_t)ZB_TCPOL().require_installcodes;
#endif  /* ZB_SECURITY_INSTALLCODES */
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      /*
        See 5.4.1 Joining with Preinstalled Trust Center Link Keys
        Trust Center link keys SHALL be installed in each device prior to joining the utility network


        a new device could be installed by the home owner or installer 2372 communicating the device install code out of band to the coordinator/trust center

        a new device could be installed by the home owner or installer 2372 communicating the device install code out of band to the coordinator/trust center
       */
#ifndef ZB_SE_BDB_MIXED
      return ZB_TRUE;
#else
      /* BDB/SE mixed mode: SE does not allow to join w/o installcode, BDB usually allows.
         If we are here:
         - we are joining device - switch from SE to BDB mode
         - we are parent device - mark somehow that child is not SE (it is not allowed to discover
         Key Establishment etc)
      */
      if (is_client && !(zb_bool_t)ZB_TCPOL().require_installcodes)
      {
        se_switch_to_bdb_commissioning();
      }
      return (zb_bool_t)ZB_TCPOL().require_installcodes;
#endif
#endif
    default:
      return ZB_FALSE;
  }
}
#endif


#if defined ZB_COORDINATOR_ROLE
/**
   After successful secure rejoin in some scenarios need to wait for unique TCLK establishment.

 */
void zdo_commissioning_secure_rejoin_setup_lk_alarm(zb_uint8_t param)
{
  switch (ZB_COMMISSIONING_TYPE())
  {
    /* In BDB, or optionally classic r21, initiate TCLK generation using APS commands */
#ifndef ZB_LITE_NO_TRUST_CENTER_REQUIRE_KEY_EXCHANGE
    case ZB_COMMISSIONING_BDB:
    case ZB_COMMISSIONING_CLASSIC:
    {
      zb_apsme_transport_key_req_t *req = ZB_BUF_GET_PARAM(param, zb_apsme_transport_key_req_t);
      /* r21 doesn't use empty keys
       * In r21 setup alarm waiting for unique TCLK request.
       * Note: param must hold transport key request even while we do not
       * actually send it there!
       */
      /*
         It IS possible to work not in BDB but require TCLK update (use case: certification tests).
         But update_trust_center_link_keys_required flag is always presend and initialized to 1 during BDB init.
       */
      /* Note: is fomewhere in time we put update_trust_center_link_keys_required under ifdef, need to check for ZB_IN_BDB() here.
         Remember ZB_IN_BDB() can be just 1 (BDB-only build). */
#ifdef ZB_COORDINATOR_ROLE
      if (ZB_TCPOL().update_trust_center_link_keys_required
          && !IS_DISTRIBUTED_SECURITY()
          && !zb_secur_get_link_key_by_address(req->dest_address.addr_long, ZB_SECUR_VERIFIED_KEY))
      {
        ZB_SCHEDULE_ALARM(bdb_link_key_refresh_alarm, param, ZB_TCPOL().trust_center_node_join_timeout);
      }
      else
#endif
      {
        zb_buf_free(param);
      }
    }
    break;
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      se_commissioning_signal(SE_COMM_SIGNAL_CHILD_SECURED_REJOIN, param);
      break;
#endif
#ifdef NCP_MODE
    case ZB_COMMISSIONING_NCP:
      zb_buf_free(param);
      break;
#endif
#endif
    default:
      break;
  }
}
#endif  /* ZB_COORDINATOR_ROLE */

#endif /* !defined NCP_MODE_HOST */


void zdo_commissioning_init(void)
{
  COMM_CTX().discovery_ctx.nwk_scan_attempts = ZB_ZDO_NWK_SCAN_ATTEMPTS;
  TRACE_MSG(TRACE_ZDO3, "nwk_scan_attempts %hd", (FMT__H, COMM_CTX().discovery_ctx.nwk_scan_attempts));
  COMM_CTX().discovery_ctx.nwk_time_btwn_scans = ZB_ZDO_NWK_TIME_BTWN_SCANS;
}
