/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*  PURPOSE: BDD specific commissioning
*/

#define ZB_TRACE_FILE_ID 62

#include "zb_common.h"
#include "zb_scheduler.h"
#include "zb_bufpool.h"
#include "zb_nwk.h"
#include "zb_nwk_nib.h"
#include "zb_aps.h"
#include "zb_zdo.h"
#include "zb_secur.h"
#include "zb_secur_api.h"
#include "zb_nvram.h"
#include "zb_bdb_internal.h"
#include "zdo_wwah_stubs.h"

#define BDB_COMMISSIONING_DEBUG

#if defined ZB_BDB_MODE && !defined BDB_OLD

void bdb_commissioning_machine(zb_uint8_t param);
static void bdb_initialization_machine(zb_uint8_t param);
static void bdb_touchlink_machine(zb_uint8_t param);
static void bdb_network_steering_machine(zb_uint8_t param);
static void bdb_network_formation_machine(zb_uint8_t param);
#if defined(ZB_BDB_ENABLE_FINDING_BINDING)
static void bdb_finding_n_binding_machine(zb_uint8_t param);
#endif
void bdb_initialization_procedure(zb_uint8_t param);
static void bdb_precomm_rejoin_over_all_channels(zb_uint8_t param, zb_uint16_t secure);
void bdb_network_steering_on_network(zb_uint8_t param);
#ifdef ZB_JOIN_CLIENT
void bdb_network_steering_not_on_network(zb_uint8_t param);
#endif
void bdb_after_mgmt_permit_joining_cb(zb_uint8_t param);
#ifdef ZB_ZCL_ENABLE_WWAH_SERVER
static void bdb_rejoin_machine(zb_uint8_t param);
#endif

zb_bool_t bdb_not_ever_joined()
{
  zb_ext_pan_id_t curr_ext_pan_id;
  zb_get_extended_pan_id(curr_ext_pan_id);

  return (zb_bool_t)(ZB_EXTPANID_IS_ZERO(curr_ext_pan_id)
                     && !zb_zdo_authenticated()
                     && !zb_zdo_tclk_valid());
}

zb_bool_t bdb_can_do_network_formation()
{
  return (zb_bool_t)(!zb_zdo_joined()
                     && (
#ifdef ZB_DISTRIBUTED_SECURITY_ON
                       (ZB_IS_DEVICE_ZR()
                        && FORMATION_SELECTOR().bdb_formation != NULL
                        /* ZR can do formation in Distributed only */
                        && zb_tc_is_distributed())
#else
                       0
#endif
                       /* Do Formation as ZC */
                       || ZB_IS_DEVICE_ZC())
    );
}

void bdb_commissioning_next_step(zb_uint8_t param)
{
  ZB_BDB().bdb_commissioning_step <<= 1;
  ZB_SCHEDULE_CALLBACK(bdb_commissioning_machine, param);
}

void bdb_initiate_commissioning(zb_uint8_t param)
{
#ifdef ZB_COORDINATOR_ROLE
  zdo_formation_force_link();
#endif
  /* ZB_BDB().bdb_commissioning_step = ZB_BDB_INITIALIZATION; */
  /* TRACE_MSG(TRACE_ZDO1, "bdb mode - initiate bdb commissioning", (FMT__0)); */
  /* ZB_SCHEDULE_CALLBACK(bdb_commissioning_machine, param); */

#if 0//def ZB_BDB_TOUCHLINK
  if (ZB_BDB().bdb_commissioning_mode == ZB_BDB_TOUCHLINK_TARGET)
  {
    ZB_BDB().bdb_commissioning_step = ZB_BDB_TOUCHLINK_TARGET;
    ZB_SCHEDULE_CALLBACK(bdb_touchlink_machine, param);
  }
  else
#endif
  {
    ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_IN_PROGRESS;
    ZB_BDB().v_do_primary_scan = ZB_BDB_JOIN_MACHINE_PRIMARY_SCAN;
    ZB_BDB().v_scan_channels = 0; /* Will be set from scan routines. */
    TRACE_MSG(TRACE_ZDO1, "Start BDB Initialization procedure", (FMT__0));
    ZB_SCHEDULE_CALLBACK(bdb_initialization_procedure, param);
  }
}

zb_bool_t bdb_start_top_level_commissioning(zb_uint8_t mode_mask)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_start_top_level_commissioning mode_mask 0x%hx", (FMT__H, mode_mask));
  if ((ZB_BDB().bdb_commissioning_step != ZB_BDB_INITIALIZATION &&
       ZB_BDB().bdb_commissioning_step != ZB_BDB_COMMISSIONING_STOP) ||
      ZB_BDB().bdb_commissioning_status == ZB_BDB_STATUS_IN_PROGRESS)
  {
    TRACE_MSG(TRACE_ZDO1, "bdb_commissioning in progress (step=%hd), return ZB_FALSE",
                          (FMT__H,ZB_BDB().bdb_commissioning_step));
    return ZB_FALSE;
  }
  ZB_BDB().bdb_commissioning_mode = mode_mask;
  ZB_BDB().bdb_commissioning_step = ZB_BDB_INITIALIZATION;
  /* Do stack init and startup */
  if (!zb_zdo_joined())
  {
    zboss_start_continue();
  }
  else
  {
    ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_IN_PROGRESS;
    ZB_BDB().bdb_commissioning_step = ZB_BDB_INITIALIZATION;
    bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_START, 0);
  }
  return ZB_TRUE;
}

void bdb_commissioning_signal(bdb_commissioning_signal_t sig, zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO3, "bdb_commissioning_signal sig %hd param %hd", (FMT__H_H, sig, param));
  BDB_COMM_CTX().signal = sig;

  if (param)
  {
    ZB_SCHEDULE_CALLBACK(bdb_commissioning_machine, param);
  }
  else
  {
    zb_buf_get_out_delayed(bdb_commissioning_machine);
  }
}

#ifdef ZB_BDB_TOUCHLINK
void bdb_check_fn()
{
  if (!ZB_AIB().aps_use_nvram)
  {
    TRACE_MSG(TRACE_ZDO1, "No nvram - factory new", (FMT__0));
    ZB_ZLL_SET_FACTORY_NEW();
  }
  else
  {
    /* TODO: recheck this condition */
    if (bdb_not_ever_joined())
    {
      TRACE_MSG(TRACE_ZDO1, "Factory new", (FMT__0));
      ZB_ZLL_SET_FACTORY_NEW();
    }
    else
    {
      TRACE_MSG(TRACE_ZDO1, "Not factory new", (FMT__0));
      ZB_ZLL_CLEAR_FACTORY_NEW();
    }
  }
}
#endif


void bdb_init(void)
{
  ZB_BDB().bdb_commissioning_group_id = (zb_uint16_t)-1;
#ifdef ZB_BDB_MODE
  /* ZBOSS "autostart" mode:
     - ZC, not on a network: Formation
     - ZC, on a network: Init + Steering
     - non-D ZR/ZED, not on a network: Steering (not on a network)
     - non-D ZR/ZED, on a network: Init + Steering (on a network)
  */
  ZB_BDB().bdb_commissioning_mode = ZB_BDB_NETWORK_STEERING | ZB_BDB_NETWORK_FORMATION | ZB_BDB_FINDING_N_BINDING;
#endif
  ZB_BDB().bdb_commissioning_time = ZB_BDBC_MIN_COMMISSIONING_TIME_S;
  ZB_BDB().bdb_scan_duration = ZB_DEFAULT_SCAN_DURATION;

  ZB_BDB().bdb_primary_channel_set = ZB_BDB().bdb_secondary_channel_set = ZB_DEFAULT_APS_CHANNEL_MASK;
}

zb_bool_t bdb_joined(void)
{
  zb_bool_t joined;
  zb_ext_pan_id_t ext_pan_id;

  zb_get_extended_pan_id(ext_pan_id);
  joined = (zb_bool_t)(!ZB_EXTPANID_IS_ZERO(ext_pan_id) &&
                       zb_zdo_authenticated());

  if (zb_aib_get_coordinator_version() >= 21)
  {
    joined = (zb_bool_t)(joined && zb_zdo_tclk_valid());
  }

  return joined;
}

void bdb_preinit(void)
{
  zb_bool_t joined;
  zb_ext_pan_id_t ext_pan_id;

  TRACE_MSG(TRACE_ZDO1, ">> bdb_preinit", (FMT__0));

#ifdef ZB_BDB_TOUCHLINK
  bdb_check_fn();
#endif

  ZB_ASSERT(zb_get_device_type() != ZB_NWK_DEVICE_TYPE_NONE);

  zb_get_extended_pan_id(ext_pan_id);
  joined = bdb_joined();

#if defined ZB_DISTRIBUTED_SECURITY_ON
  if (joined)
  {
    zb_sync_distributed();
  }
#endif

  if (joined)
  {
    TRACE_MSG(TRACE_ZDO1, "Not factory new", (FMT__0));
#ifdef ZB_BDB_TOUCHLINK
    /* just in case */
    ZB_ZLL_CLEAR_FACTORY_NEW();
#endif
  }

#ifdef ZB_BDB_TOUCHLINK
  /* 8.7 Touchlink procedure for an initiator:
     The touchlink procedure for an initiator can perform a "normal" channel scan or an "extended"
     channel scan; the latter is used if a reset to factory new is required (see sub-clause 9.2) or
     if the target could be operating on a channel other than those defined in
     bdbcTLPrimaryChannelSet.
 */
  ZB_BDB().bdb_ext_channel_scan = 1; /* ZB_ZLL_IS_FACTORY_NEW() */
#endif
  TRACE_MSG(TRACE_INFO1,
            "dev type %hd, joined %hd, ext_pan_id %hd, authenticated %hd, tclk_valid %hd",
            (FMT__H_H_H_H_H,
             zb_get_device_type(),
             joined,
             !ZB_EXTPANID_IS_ZERO(ext_pan_id),
             zb_zdo_authenticated(),
             zb_zdo_tclk_valid()));
}

/**
   BDB Initialization procedure according to 7.1 Initialization procedure
 */
void bdb_initialization_procedure(zb_uint8_t param)
{
  zb_ext_pan_id_t ext_pan_id;
  zb_get_extended_pan_id(ext_pan_id);

  /* Note: Called twice on startup in r21 (from zb_zdo_dev_init() and here), but looks like it is not
   * a big problem. */
  bdb_preinit();
/* Can do rejoin if ZED and if authenticated: only secured rejoin is
   * allowed by BDB.
   * Seems, ZR must continue its work without a rejoin (hmm?)
   */
  if (bdb_joined())
  {
#if defined ZB_COORDINATOR_ROLE
    if (ZB_IS_DEVICE_ZC())
    {
      TRACE_MSG(TRACE_ZDO2, "Start ZC without formation", (FMT__0));
      ZB_SCHEDULE_CALLBACK(zb_nwk_cont_without_formation, param);
      param = 0;
    }
    else
#endif
#if defined ZB_ROUTER_ROLE
    if (ZB_IS_DEVICE_ZR() && !ZB_BDB().bdb_force_router_rejoin)
    {
      TRACE_MSG(TRACE_ZDO2, "Initiate start router for ZR", (FMT__0));
      ZB_SCHEDULE_CALLBACK(zb_zdo_start_router, param);
      param = 0;
    }
    else
#endif
    {
      /* Perform secure rejoin */
      TRACE_MSG(TRACE_ZDO2, "Perform secure rejoin", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_SECURE_REJOIN, param);
      param = 0;
    }

    /* Reset steering to avoid network opening
     * (permit join request command) */
    ZB_BDB().bdb_commissioning_mode = ZB_BDB_NETWORK_FORMATION | ZB_BDB_FINDING_N_BINDING;

    /* We are NFN - indicate ZB_BDB_SIGNAL_DEVICE_REBOOT */
    ZB_BDB().bdb_application_signal = ZB_BDB_SIGNAL_DEVICE_REBOOT;
  }
  else if (bdb_not_ever_joined())
  {
    /* Nothing to do, just continue */
    TRACE_MSG(TRACE_ERROR, "Nothing to do, just continue", (FMT__0));
    ZB_BDB().bdb_application_signal = ZB_BDB_SIGNAL_DEVICE_FIRST_START;
  }
  else if (!ZB_EXTPANID_IS_ZERO(ext_pan_id)
           && ((zb_aib_get_coordinator_version() < 21) ||
               ((zb_aib_get_coordinator_version() >= 21) && zb_zdo_tclk_valid())))
  {
    /* Perform TC rejoin */
    TRACE_MSG(TRACE_ERROR, "Perform TC rejoin", (FMT__0));
    bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_TC_REJOIN, param);

    ZB_BDB().bdb_application_signal = ZB_BDB_SIGNAL_DEVICE_REBOOT;
    param = 0;
  }
#ifndef ZB_COORDINATOR_ONLY
  else
  {
    TRACE_MSG(TRACE_ERROR, "Strange configuration, leave", (FMT__0));
    zdo_commissioning_leave(param, ZB_FALSE, ZB_FALSE);
    param = 0;
  }
#endif  /* #ifndef ZB_COORDINATOR_ONLY */

  if (param)
  {
    /* No additional actions are needed, finish initialization */
    bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_FINISH, param);
  }
}

static void bdb_post_commissioning_actions(void)
{
#if (defined ZB_ZCL_SUPPORT_CLUSTER_WWAH && defined ZB_ZCL_ENABLE_WWAH_SERVER)
  if (zb_is_wwah_server())
  {
    zb_zcl_wwah_start_periodic_checkin();
    zb_zcl_wwah_start_bad_parent_recovery();
  }
#endif
}

void bdb_commissioning_machine(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_commissioning_machine param %hd signal %hd state %hd",
            (FMT__H_H_H, param, BDB_COMM_CTX().signal, ZB_BDB().bdb_commissioning_step));

  TRACE_MSG(TRACE_ZDO1, ">> bdb_commissioning_machine param %hd step 0x%x bdb_commissioning_mode 0x%x bdb_commissioning_status %d joined %d",
            (FMT__H_D_D_D_D, param, ZB_BDB().bdb_commissioning_step, ZB_BDB().bdb_commissioning_mode, ZB_BDB().bdb_commissioning_status, zb_zdo_joined()));

  /* Run machines specific to the current SE device state */
  switch (ZB_BDB().bdb_commissioning_step)
  {
    case ZB_BDB_INITIALIZATION:
      ZB_SCHEDULE_CALLBACK(bdb_initialization_machine, param);
      break;

    case ZB_BDB_TOUCHLINK_COMMISSIONING:
      /* FIXME: Do we really need separate step for target? Do not touch it for now, but need to
       * refactor during Touchlink commissioning refactoring/extension. */
    case ZB_BDB_TOUCHLINK_TARGET:
      ZB_SCHEDULE_CALLBACK(bdb_touchlink_machine, param);
      break;

    case ZB_BDB_NETWORK_STEERING:
      ZB_SCHEDULE_CALLBACK(bdb_network_steering_machine, param);
      break;

    case ZB_BDB_NETWORK_FORMATION:
      ZB_SCHEDULE_CALLBACK(bdb_network_formation_machine, param);
      break;

#if defined(ZB_BDB_ENABLE_FINDING_BINDING)
    case ZB_BDB_FINDING_N_BINDING:
      ZB_SCHEDULE_CALLBACK(bdb_finding_n_binding_machine, param);
      break;
#endif

#ifdef ZB_ZCL_ENABLE_WWAH_SERVER
    case ZB_BDB_REJOIN:
      ZB_SCHEDULE_CALLBACK(bdb_rejoin_machine, param);
      break;
#endif /* ZB_ZCL_ENABLE_WWAH_SERVER */

    case ZB_BDB_COMMISSIONING_STOP:
      if (BDB_COMM_CTX().signal == BDB_COMM_SIGNAL_FINISH)
      {
        /* commissioning done */
        if (ZB_BDB().bdb_commissioning_status == ZB_BDB_STATUS_IN_PROGRESS)
        {
          ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_SUCCESS;
          /* Reset "force rejoin" flag */
          ZB_BDB().bdb_force_router_rejoin = ZB_FALSE;
        }
        TRACE_MSG(TRACE_INFO1, "COMMISSIONING_STOP: app signal %hd comm status %hd", (FMT__H_H, ZB_BDB().bdb_application_signal, ZB_BDB().bdb_commissioning_status));
        zb_app_signal_pack(param, ZB_BDB().bdb_application_signal, -ZB_BDB().bdb_commissioning_status, 0);
        ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);

        /* FIXME: Not sure when to start WWAH and other activities */
        if (ZB_BDB().bdb_commissioning_status == ZB_BDB_STATUS_SUCCESS &&
            (ZB_BDB().bdb_application_signal == ZB_BDB_SIGNAL_DEVICE_FIRST_START ||
             ZB_BDB().bdb_application_signal == ZB_BDB_SIGNAL_DEVICE_REBOOT ||
             ZB_BDB().bdb_application_signal == ZB_BDB_SIGNAL_STEERING))
        {
          bdb_post_commissioning_actions();
        }
      }
      /* else ignore anything we can receive (for instance, LEAVE_DONE etc) */
      else
      {
        TRACE_MSG(TRACE_ZDO1, "commissioning is not in progress, ignore", (FMT__0));
        zb_buf_free(param);
      }
      break;

    default:
      TRACE_MSG(TRACE_ERROR, "Unknown commissioning state %d", (FMT__D, ZB_BDB().bdb_commissioning_mode));
#ifdef BDB_COMMISSIONING_DEBUG
      ZB_ASSERT(0);
#else
      if (param)
      {
        zb_buf_free(param);
      }
#endif
      break;
  }
}

static void bdb_initialization_machine(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_initialization_machine param %hd signal %hd state %hd",
            (FMT__H_H_H, param, BDB_COMM_CTX().signal, ZB_BDB().bdb_commissioning_step));
  switch (BDB_COMM_CTX().signal)
  {
    case BDB_COMM_SIGNAL_INIT_START:
    {
#ifdef ZB_BDB_TOUCHLINK
      if (ZB_BDB().bdb_commissioning_mode == ZB_BDB_TOUCHLINK_TARGET)
      {
        bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_FINISH, param);
      }
      else
#endif
      {
        if (bdb_joined() && zb_zdo_joined())
        {
          TRACE_MSG(TRACE_ZDO1, "Already inited, skip initialization", (FMT__0));
          bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_FINISH, param);
        }
        else
        {
          ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_IN_PROGRESS;
          ZB_BDB().v_do_primary_scan = ZB_BDB_JOIN_MACHINE_PRIMARY_SCAN;
          ZB_BDB().v_scan_channels = 0; /* Will be set from scan routines. */
          TRACE_MSG(TRACE_ZDO1, "Start BDB Initialization procedure", (FMT__0));
          ZB_SCHEDULE_CALLBACK(bdb_initialization_procedure, param);
        }
      }
    }
    break;

    case BDB_COMM_SIGNAL_INIT_SECURE_REJOIN:
      bdb_precomm_rejoin_over_all_channels(param, 1);
      break;

    case BDB_COMM_SIGNAL_INIT_TC_REJOIN:
      bdb_precomm_rejoin_over_all_channels(param, 0);
      break;

    case BDB_COMM_SIGNAL_NWK_AUTH_FAILED:
      /* ZG->aps.authenticated = 0; */
      /* FALLTHROUGH */
    case BDB_COMM_SIGNAL_NWK_JOIN_FAILED:
      TRACE_MSG(TRACE_ZDO1, "rejoin failed, finish", (FMT__0));
      ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_NO_NETWORK;
      bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_NWK_JOIN_DONE:
      TRACE_MSG(TRACE_ZDO1, "Device is joined", (FMT__0));
#ifdef ZB_JOIN_CLIENT
      if (!zdo_secur_waiting_for_tclk_update())
#endif
      {
        /* Join is finished - can reset the blacklist. */
#ifdef ZB_NWK_BLACKLIST
        zb_nwk_blacklist_reset();
#endif
      }
      bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_NWK_FORMATION_OK:
      TRACE_MSG(TRACE_ZDO1, "Network is started OK", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_NWK_START_ROUTER_CONF:
      TRACE_MSG(TRACE_ZDO1, "Network is started OK (router)", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_LEAVE_DONE:
      TRACE_MSG(TRACE_ZDO1, "Left from the network", (FMT__0));
      /* ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_NO_NETWORK; */
      /* We left from the network (probably because of wrong network configuration).
         Try to continue commissioning as FN device. */
      ZB_BDB().bdb_application_signal = ZB_BDB_SIGNAL_DEVICE_FIRST_START;
      bdb_commissioning_signal(BDB_COMM_SIGNAL_INIT_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_INIT_FINISH:
      TRACE_MSG(TRACE_ZDO1, "Initialization finished, next step - Touchlink", (FMT__0));
      /* Go to next commissioning step - Touchlink procedure */
      ZB_BDB().bdb_commissioning_step = ZB_BDB_TOUCHLINK_COMMISSIONING;
      bdb_commissioning_signal(BDB_COMM_SIGNAL_TOUCHLINK_START, param);
      break;

    case BDB_COMM_SIGNAL_NETWORK_STEERING_DISCOVERY_FAILED:

      break;


    default:
      TRACE_MSG(TRACE_ERROR, "Unknown signal %d", (FMT__D, BDB_COMM_CTX().signal));
#ifdef BDB_COMMISSIONING_DEBUG
      ZB_ASSERT(0);
#else
      if (param)
      {
        zb_buf_free(param);
      }
#endif
      break;
  }
}

static void bdb_precomm_rejoin_over_all_channels(zb_uint8_t param, zb_uint16_t secure)
{
  if (!param)
  {
    zb_buf_get_out_delayed_ext(bdb_precomm_rejoin_over_all_channels, secure, 0);
  }
  else
  {
    zb_channel_list_t rejoin_ch_list;
    zb_ext_pan_id_t ext_pan_id;


    TRACE_MSG(TRACE_ZDO1, "bdb_precomm_rejoin_over_all_channels %hd", (FMT__H, param));
    zb_channel_list_init(rejoin_ch_list);
#if defined ZB_SUBGHZ_BAND_ENABLED
    zb_channel_page_list_copy(rejoin_ch_list, ZB_AIB().aps_channel_mask_list);
#else
    zb_channel_page_list_set_2_4GHz_mask(rejoin_ch_list,
                                         zb_aib_channel_page_list_get_2_4GHz_mask());
#endif /* !ZB_SUBGHZ_BAND_ENABLED */
    zb_get_extended_pan_id(ext_pan_id);
    zdo_initiate_rejoin(param, ext_pan_id,
                        rejoin_ch_list,
                        (zb_bool_t)secure);
  }
}

#ifdef ZB_BDB_TOUCHLINK
static void bdb_touchlink_target_timeout(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_touchlink_target_timeout: finish Touchlink Target", (FMT__0));
  ZB_BDB().bdb_application_signal = ZB_BDB_SIGNAL_TOUCHLINK_TARGET_FINISHED;
  ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_SUCCESS;
  bdb_commissioning_signal(BDB_COMM_SIGNAL_TOUCHLINK_FINISH, param);
}
#endif

/* TODO: Split on Touchlink Target and Touchlink Initiator */
static void bdb_touchlink_machine(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_touchlink_machine param %hd signal %hd state %hd",
            (FMT__H_H_H, param, BDB_COMM_CTX().signal, ZB_BDB().bdb_commissioning_step));
  switch (BDB_COMM_CTX().signal)
  {
    case BDB_COMM_SIGNAL_TOUCHLINK_START:
    {
#ifdef ZB_BDB_TOUCHLINK
      if (ZB_BDB().bdb_commissioning_mode & ZB_BDB_TOUCHLINK_TARGET)
      {
        /* TODO: Fix this correctly after discussion about signals. */
        if (ZB_BDB().bdb_commissioning_mode == ZB_BDB_TOUCHLINK_TARGET)
        {
          ZB_BDB().bdb_application_signal = ZB_BDB_SIGNAL_TOUCHLINK_TARGET;
        }

        TRACE_MSG(TRACE_ZDO1, "Start Touchlink Target", (FMT__0));
        /* Now if touchlink compiled, force formation compile (need for distributed). */
        /* TODO: a) check for touchlink enabled; b) call touchlink indirectly */
#ifdef ZB_FORMATION
        zdo_formation_force_link();
#endif
        ZB_SCHEDULE_CALLBACK(bdb_touchlink_target_start, param);
      }
      else if (ZB_BDB().bdb_commissioning_mode & ZB_BDB_TOUCHLINK_COMMISSIONING)
      {
        /* TODO: Fix this correctly after discussion about signals. */
        if (ZB_BDB().bdb_commissioning_mode == ZB_BDB_TOUCHLINK_COMMISSIONING)
        {
          ZB_BDB().bdb_application_signal = ZB_BDB_SIGNAL_TOUCHLINK;
        }

        TRACE_MSG(TRACE_ZDO1, "Start Touchlink initiator", (FMT__0));
        TRACE_MSG(TRACE_ZDO1, "Running BDB Touchlink initiator v_do_primary_scan %hd", (FMT__H, ZB_BDB().v_do_primary_scan));
        ZB_SCHEDULE_CALLBACK(bdb_touchlink_initiator, param);
      }
      else
#endif
      {
        TRACE_MSG(TRACE_ZDO1, "Skip Touchlink", (FMT__0));
        bdb_commissioning_signal(BDB_COMM_SIGNAL_TOUCHLINK_FINISH, param);
      }
    }
    break;

#ifdef ZB_BDB_TOUCHLINK
    case BDB_COMM_SIGNAL_TOUCHLINK_INITIATOR_DONE:
      TRACE_MSG(TRACE_ZDO1, "Touchlink initiator succeeded, continue", (FMT__0));
      ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_SUCCESS;
      bdb_commissioning_signal(BDB_COMM_SIGNAL_TOUCHLINK_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_TOUCHLINK_INITIATOR_FAILED:
      TRACE_MSG(TRACE_ZDO1, "Touchlink initiator failed", (FMT__0));
      ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_TARGET_FAILURE;
      bdb_commissioning_signal(BDB_COMM_SIGNAL_TOUCHLINK_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_NETWORK_STEERING_DISCOVERY_FAILED:
      if (ZB_BDB().bdb_commissioning_mode & ZB_BDB_TOUCHLINK_TARGET)
      {
        /* Touchlink Target: We are asked to start the network, no other networks discovered -
         * continue */
        /* TODO: Manipulating with status this way only because of bdb_touchlink implementation,
         * remove when it will be fixed! */
        ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_NO_NETWORK;
        ZB_SCHEDULE_CALLBACK(bdb_touchlink_target_start, param);
      }
      else
      {
        ZB_ASSERT(0);           /* Debug */
      }
      break;

    case BDB_COMM_SIGNAL_NWK_JOIN_FAILED:
      if (ZB_BDB().bdb_commissioning_mode & ZB_BDB_TOUCHLINK_COMMISSIONING &&
          !(ZB_BDB().bdb_commissioning_mode & ZB_BDB_TOUCHLINK_TARGET))
      {
        TRACE_MSG(TRACE_ZDO1, "Touchlink Initiator: Failed to rejoin to the formed network", (FMT__0));
        ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_NO_NETWORK;
        bdb_commissioning_signal(BDB_COMM_SIGNAL_TOUCHLINK_FINISH, param);
      }
      else
      {
        ZB_ASSERT(0);           /* Debug */
      }
      break;

    case BDB_COMM_SIGNAL_NWK_START_ROUTER_CONF:
    {
      TRACE_MSG(TRACE_ZDO1, "Touchlink Target: successfully formed the network", (FMT__0));

      /* Inform the application about Touchlink Target formation */
      zb_app_signal_pack(param, ZB_BDB_SIGNAL_TOUCHLINK_NWK,
                         ZB_BDB_STATUS_SUCCESS,
                         0);
      ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);

      ZB_SCHEDULE_ALARM_CANCEL(bdb_touchlink_target_timeout, ZB_ALARM_ANY_PARAM);
      bdb_touchlink_target_timeout(0);
    }
    break;

    case BDB_COMM_SIGNAL_NWK_JOIN_DONE:
      TRACE_MSG(TRACE_ZDO1, "Touchlink Initiator: successfully joined the network", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_TOUCHLINK_FINISH, 0);
      break;

    case BDB_COMM_SIGNAL_TOUCHLINK_NOTIFY_TASK_RESULT:
    {
      zb_ret_t status = zb_buf_get_status(param);
      if (ZB_BDB().bdb_commissioning_mode & ZB_BDB_TOUCHLINK_TARGET)
      {
        if (status == ZB_ZLL_TASK_STATUS_OK)
        {
          /* Touchlink Target started */

          /* FIXME: BDB spec does not define the timeout for Touchlink Target. Let's assume device
           * will be in the Target mode for 1 min by default. */
          /* TODO: Stop Touchlink Target if device finishes Touchlink Target procedure before
           * timeout. */
          TRACE_MSG(TRACE_ZDO1, "Touchlink target started, wait for Touchlink Target timeout", (FMT__0));
          ZB_SCHEDULE_ALARM(bdb_touchlink_target_timeout, 0, ZB_TIME_ONE_SECOND * 60);

          /* Inform the application that Touchlink Target is started */
          zb_app_signal_pack(param, ZB_BDB_SIGNAL_TOUCHLINK_TARGET, ZB_BDB_STATUS_SUCCESS, 0);
          ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
        }
        else
        {
          TRACE_MSG(TRACE_ZDO1, "Touchlink target start failed", (FMT__0));
          ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_TARGET_FAILURE;
          bdb_commissioning_signal(BDB_COMM_SIGNAL_TOUCHLINK_FINISH, param);
        }
      }
      else /* Touchlink Initiator */
      {
        if (status == ZB_ZLL_TASK_STATUS_FINISHED)
        {
          TRACE_MSG(TRACE_ZDO1, "Some device discovered: continue with touchlink_initiator", (FMT__0));
          /* TODO: Refactor this function! Similar to bdb_commissioning_machine() there is a lot of
           * unclear manipulations with flags there etc. */
          ZB_SCHEDULE_CALLBACK(bdb_touchlink_initiator, param);
        }
        else
        {
          ZB_ASSERT(0);         /* debug */
        }
      }
    }
    break;
#endif /* ZB_BDB_TOUCHLINK */
    case BDB_COMM_SIGNAL_TOUCHLINK_FINISH:
      if (ZB_BDB().bdb_commissioning_status != ZB_BDB_STATUS_IN_PROGRESS)
      {
        ZB_BDB().bdb_commissioning_step = ZB_BDB_COMMISSIONING_STOP;
        TRACE_MSG(TRACE_ZDO3, "ZB_BDB_COMMISSIONING_STOP", (FMT__0));
        bdb_commissioning_signal(BDB_COMM_SIGNAL_FINISH, param);
      }
      else
      {
        TRACE_MSG(TRACE_ZDO1, "Touchlink finished, next step - Network Steering", (FMT__0));
        /* Go to next commissioning step - Network Steering procedure */
        ZB_BDB().bdb_commissioning_step = ZB_BDB_NETWORK_STEERING;
        bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_START, param);
      }
      break;

    default:
      TRACE_MSG(TRACE_ERROR, "Unknown signal %d", (FMT__D, BDB_COMM_CTX().signal));
#ifdef BDB_COMMISSIONING_DEBUG
      ZB_ASSERT(0);
#else
      if (param)
      {
        zb_buf_free(param);
      }
#endif
      break;
  }
}

#ifndef ZB_COORDINATOR_ONLY
void bdb_network_steering_auth_failure(zb_uint8_t param)
{
  ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_NO_NETWORK;
#ifdef ZB_CERTIFICATION_HACKS
  if (ZB_CERT_HACKS().stay_on_network_after_auth_failure && zb_zdo_joined())
  {
    /* special hack for bdb tests - do not leave network if authentication failed */
    ZB_BDB().bdb_commissioning_step = ZB_BDB_COMMISSIONING_STOP;
    TRACE_MSG(TRACE_ZDO3, "ZB_BDB_COMMISSIONING_STOP", (FMT__0));
    bdb_commissioning_signal(BDB_COMM_SIGNAL_FINISH, param);
  }
  else
#endif
  {
    TRACE_MSG(TRACE_ZDO1, "Authentication failure. Leave.", (FMT__0));
    zdo_commissioning_leave(param, ZB_FALSE, ZB_FALSE);
  }
}

void bdb_network_steering_tclk_ex_failure(zb_uint8_t param)
{
  ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_TCLK_EX_FAILURE;

  /* Joined but failed to update it's TC link key - leave network and complete steering
   * with status ZB_BDB_STATUS_TCLK_EX_FAILURE */

  TRACE_MSG(TRACE_ZDO1, "TC link key update failure. Leave.", (FMT__0));
  /* ZB_BDB().bdb_commissioning_step = ZB_BDB_LAST_COMMISSIONING_STEP; */

#ifdef ZB_CERTIFICATION_HACKS
  if (ZB_CERT_HACKS().stay_on_network_after_auth_failure)
  {
    /* special hack for bdb tests - do not leave network if tclk update failed */
    bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH, param);
  }
  else
#endif
  {
    zdo_commissioning_leave(param, ZB_FALSE, ZB_FALSE);
  }
}
#endif /* !ZB_COORDINATOR_ONLY */

void bdb_network_steering_start(zb_uint8_t param)
{
#ifdef ZB_JOIN_CLIENT
  if (zdo_secur_waiting_for_tclk_update())
  {
    TRACE_MSG(TRACE_ZDO1, "We are waiting for TCLK update. What to do? Lets free the buf at least. buf %hd", (FMT__H, param));
    zb_buf_free(param);
  }
  else
#endif
  {
    if (zb_zdo_joined())
    {
      /*
        The node then sets bdbCommissioningStatus to SUCCESS. If the node supports
        touchlink, it sets the values of the aplFreeNwkAddrRangeBegin,
        aplFreeNwkAddrRangeEnd, aplFreeGroupID-RangeBegin and
        aplFreeGroupIDRangeEnd attributes all to 0x0000 (indicating the node 1045 having
        joined the network using MAC association).
      */
      TRACE_MSG(TRACE_ZDO1, "Start BDB network steering when on network", (FMT__0));
      ZB_SCHEDULE_CALLBACK(bdb_network_steering_on_network, param);
    }
    else
    {
      /* If we are NFN and join failed - finish and pass it up to application. */
      if (!zb_bdb_is_factory_new())
      {
        TRACE_MSG(TRACE_ZDO1, "Not ZB_JOINED and NFN - finish BDB", (FMT__0));
        ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_NO_NETWORK;
        bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH, param);
      }
      else
      {
        TRACE_MSG(TRACE_ZDO1, "Start BDB network steering when NOT on network", (FMT__0));
#ifdef ZB_JOIN_CLIENT
        ZB_SCHEDULE_CALLBACK(bdb_network_steering_not_on_network, param);
#endif
      }
    }
  }
}

#ifdef ZB_JOIN_CLIENT
void bdb_network_steering_start_scan(zb_uint8_t param)
{
  zb_nlme_network_discovery_request_t *req = ZB_BUF_GET_PARAM(param, zb_nlme_network_discovery_request_t);

  ZB_BZERO(req, sizeof(zb_nlme_network_discovery_request_t));

  TRACE_MSG(TRACE_ZDO3, "bdb_network_steering_start_scan param %hd", (FMT__H, param));
  /* MMDEVSTUBS */

#if defined ZB_SUB_GHZ_ZB30_SUPPORT
  {
    zb_uint8_t used_page;

    used_page = zb_aib_channel_page_list_get_first_filled_page();
    zb_channel_page_list_set_mask(req->scan_channels_list, used_page, ZB_BDB().v_scan_channels);
  }
#else
  zb_channel_page_list_set_2_4GHz_mask(req->scan_channels_list, ZB_BDB().v_scan_channels);
#endif /* !ZB_SUB_GHZ_ZB30_SUPPORT */

  req->scan_duration = ZB_BDB().bdb_scan_duration;
  COMM_CTX().discovery_ctx.disc_count = COMM_CTX().discovery_ctx.nwk_scan_attempts;
  /* That finishes in zdo_app.c zb_nlme_network_discovery_confirm */
  ZB_SCHEDULE_CALLBACK(zb_nlme_network_discovery_request, param);
}


void bdb_network_steering_not_on_network(zb_uint8_t param)
{
  if (ZB_BDB().v_do_primary_scan == ZB_BDB_JOIN_MACHINE_PRIMARY_SCAN
      && ZB_BDB().bdb_primary_channel_set != 0
      && !ZB_IS_DEVICE_ZC())
  {
    ZB_BDB().v_do_primary_scan = ZB_BDB_JOIN_MACHINE_SECONDARY_SCAN_START;
    ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_IN_PROGRESS;

#ifdef ZB_REJOIN_BACKOFF
    if (zb_zdo_rejoin_backoff_is_running() && (ZDO_CTX().zdo_rejoin_backoff.rjb_cnt == 1) && !ZB_BDB().v_scan_channels)
    {
      /* the first attempt - do secure rejoin using the current channel */
      ZB_BDB().v_scan_channels = (1l << ZB_PIBCACHE_CURRENT_CHANNEL());
      /* TRICKY: Set primary_scan again - we will do 2 PRIMARY_SCANs (first on current channel,
       * second on bdb_primary_channel_set), then SECONDARY_SCAN.  */
      ZB_BDB().v_do_primary_scan = ZB_BDB_JOIN_MACHINE_PRIMARY_SCAN;
    }
    else
    {
      /* We done rejoin_backoff specific attempt, now reset v_scan_channels and use usual scheme
       * (bdb_primary_channel_set, then bdb_secondary_channel_set). */
      ZB_BDB().v_scan_channels = 0;

      /* the second and further attempts - do secure rejoin using ALL channels */

      /* 01/27/2017 EE Why having already 2 channel sets we introduce
         another one - rejoin backoff channels mask? Can't we go thru
         primary/secondary like we do for usual start? ok, after first attempt at
         the current channel.
         Do we need default_channel_mask parameter in zb_zdo_rejoin_backoff_start?
         Isn't product_cfg.aps_channel_mask == bdb_primary_channel_set?
      */
      /* NK: Seems like it is ok. Lets update bdb_channel_mask on rejoin_backoff start instead of
       * using rjb_chan_mask (in the case when application wants to use custom channel mask for rjb).
       *
       * Currently will not remove rjb_chan_mask because it is used in legacy (non-BDB)
       * variant. It is a hack because in this variant we overwrite original
       * ZB_AIB().aps_channel_mask on 1st rjb attempt, then restore it from rjb_chan_mask (in all known
       * cases production_mask == ZB_AIB().aps_channel_mask == rjb_chan_mask).
       * Not sure about supporting legacy (non-BDB) rejoin_backoff, for me it looks like it may be
       * completely removed.
       */

      /* Implement insecure rejoin via BDB if needed. */
    }
#endif

    if (!ZB_BDB().v_scan_channels)
    {
      /* 01/30/2017 EE  In case of rejoin backoff you substitute primary
      channel set by current channel, then use secondary channel set.
      Not sure this is what you want to do.
      Probably need to go in 2 steps: current, primary, secondary.
      As an alternative (simpler, but a bit further from BDB : current, primary | secondary.
      BTW I do not know why they ever invented primary and secondary...
      */
      ZB_BDB().v_scan_channels = ZB_BDB().bdb_primary_channel_set;
    }
    TRACE_MSG(TRACE_ZDO1, "Doing primary scan channel mask 0x%lx", (FMT__L, ZB_BDB().v_scan_channels));
    ZB_SCHEDULE_CALLBACK(bdb_network_steering_start_scan, param);
  }
  else if (!ZB_IS_DEVICE_ZC()
           && ((ZB_BDB().v_do_primary_scan == ZB_BDB_JOIN_MACHINE_SECONDARY_SCAN_START
                && ZB_BDB().bdb_secondary_channel_set != 0)
               ||
               (ZB_BDB().v_do_primary_scan == ZB_BDB_JOIN_MACHINE_PRIMARY_SCAN
                && ZB_BDB().bdb_secondary_channel_set != 0
                && ZB_BDB().bdb_primary_channel_set == 0)))
  {
    ZB_BDB().v_do_primary_scan = ZB_BDB_JOIN_MACHINE_SECONDARY_SCAN_DONE;
    ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_IN_PROGRESS;
    ZB_BDB().v_scan_channels = ZB_BDB().bdb_secondary_channel_set;
    TRACE_MSG(TRACE_ZDO1, "Doing secondary scan channel mask 0x%lx", (FMT__L, ZB_BDB().v_scan_channels));
    ZB_SCHEDULE_CALLBACK(bdb_network_steering_start_scan, param);
  }
  else
  {
    ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_NO_NETWORK;
    TRACE_MSG(TRACE_ZDO1, "BDB: Could not join or authorize - run next bdb machine step", (FMT__0));
    ZB_BDB().v_do_primary_scan = ZB_BDB_JOIN_MACHINE_PRIMARY_SCAN;
    bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH, param);
  }
}
#endif  /* ZB_JOIN_CLIENT */

void bdb_network_steering_finish(zb_uint8_t param)
{
  bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH, param);
}

void bdb_network_steering_on_network(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO3, "bdb_network_steering_on_network param %hd", (FMT__H, param));
  /*
    Send Permit Joining.

    If I am ZR or ZC, open myself for connects (permit joining).
   */
  {
    /*
  Continues EZ-Mode Nwk steering after network has been formed or device has joined to the nwk:
  invokes permit_joining routines for ZC & ZR, pushes the results of Nwk steering up
  via user callback for ZED
*/
    /* Send mgmt permit joining broadcast */
    zb_zdo_mgmt_permit_joining_req_param_t *req_param = ZB_BUF_GET_PARAM(param, zb_zdo_mgmt_permit_joining_req_param_t);

    ZB_BZERO(req_param, sizeof(zb_zdo_mgmt_permit_joining_req_param_t));
    req_param->dest_addr = ZB_NWK_BROADCAST_ROUTER_COORDINATOR;
    req_param->permit_duration = ZB_BDBC_MIN_COMMISSIONING_TIME_S;
    req_param->tc_significance = 1;
    zb_zdo_mgmt_permit_joining_req(param, bdb_after_mgmt_permit_joining_cb);
  }
}

/*
  Called when broadcasting PermitJoining finished
  during EZ-Mode Nwk Steering for ZC or ZR:
  schedules local nlme.permit_joining request
*/
void bdb_after_mgmt_permit_joining_cb(zb_uint8_t param)
{

  TRACE_MSG(TRACE_ZDO2, "> bdb_after_mgmt_permit_joining_cb %hd", (FMT__H, param));

  if (zb_buf_get_status(param) == RET_OK)
  {
#ifdef ZB_ROUTER_ROLE
    if (ZB_IS_DEVICE_ZC_OR_ZR())
    {
      zb_zdo_mgmt_permit_joining_req_param_t *req = ZB_BUF_GET_PARAM(param,
                                                                           zb_zdo_mgmt_permit_joining_req_param_t);

      ZB_BZERO(req, sizeof(zb_zdo_mgmt_permit_joining_req_param_t));
      req->permit_duration = ZB_BDBC_MIN_COMMISSIONING_TIME_S;
      req->dest_addr = ZB_PIBCACHE_NETWORK_ADDRESS();

      zb_zdo_mgmt_permit_joining_req(param, bdb_network_steering_finish);
      param = 0;
    }
#endif
  }
  else
  {
    ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_NOT_PERMITTED;
  }

  if (param)
  {
    bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH, param);
  }

  TRACE_MSG(TRACE_ZCL2, "< bdb_after_mgmt_permit_joining_cb", (FMT__0));
}

static void bdb_network_steering_machine(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_network_steering_machine param %hd signal %hd state %hd",
            (FMT__H_H_H, param, BDB_COMM_CTX().signal, ZB_BDB().bdb_commissioning_step));
  switch (BDB_COMM_CTX().signal)
  {
    case BDB_COMM_SIGNAL_NETWORK_STEERING_START:
    {
      if (ZB_BDB().bdb_commissioning_mode & ZB_BDB_NETWORK_STEERING)
      {
        /* TODO: Fix this correctly after discussion about signals. */
        if (ZB_BDB().bdb_commissioning_mode == ZB_BDB_NETWORK_STEERING)
        {
          ZB_BDB().bdb_application_signal = ZB_BDB_SIGNAL_STEERING;
        }

        /* For ZC steering means "open net for join". Meaningful only after formation complete. */
        if (!ZB_IS_DEVICE_ZC() || zb_zdo_joined())
        {
          TRACE_MSG(TRACE_ZDO1, "Start Network Steering", (FMT__0));
          ZB_SCHEDULE_CALLBACK(bdb_network_steering_start, param);
        }
        else
        {
          TRACE_MSG(TRACE_ZDO1, "Skip Network Steering: meaningless before formation", (FMT__0));
          bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH, param);
        }
      }
      else
      {
        TRACE_MSG(TRACE_ZDO1, "Skip Network Steering", (FMT__0));
        bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH, param);
      }
    }
    break;

    case BDB_COMM_SIGNAL_NETWORK_STEERING_DISCOVERY_FAILED:
    case BDB_COMM_SIGNAL_NWK_JOIN_FAILED:
      TRACE_MSG(TRACE_ZDO1, "No networks found", (FMT__0));
      ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_NO_NETWORK;
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH, param);
      break;

#ifndef ZB_COORDINATOR_ONLY
    case BDB_COMM_SIGNAL_NWK_AUTH_FAILED:
      ZB_SCHEDULE_CALLBACK(bdb_network_steering_auth_failure, param);
      break;

    case BDB_COMM_SIGNAL_NETWORK_STEERING_TCLK_EX_FAILURE:
      ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_TCLK_EX_FAILURE;
      ZB_SCHEDULE_CALLBACK(bdb_network_steering_tclk_ex_failure, param);
      break;
#endif /* !ZB_COORDINATOR_ONLY */

    case BDB_COMM_SIGNAL_NWK_START_ROUTER_CONF:
    case BDB_COMM_SIGNAL_NWK_JOIN_DONE:
#ifdef ZB_JOIN_CLIENT
      if (!zdo_secur_waiting_for_tclk_update())
      {
        /* Join is finished - can reset the blacklist. */
        TRACE_MSG(TRACE_ZDO1, "Router started/ED join done, TCLK is not needed - continue", (FMT__0));
        ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_IN_PROGRESS;
        /* Join is finished - can reset the blacklist. */
#ifdef ZB_NWK_BLACKLIST
        zb_nwk_blacklist_reset();
#endif
        /* TRICKY: Run steering one more time - to broadcast Permit Joining. This time it will be
         * steering for device on network. */
        bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_START, param);
      }
      else
#endif
      {
        TRACE_MSG(TRACE_ZDO1, "Router started/ED join done, wait for TCLK", (FMT__0));
        zb_buf_free(param);
      }
      break;

    case BDB_COMM_SIGNAL_LEAVE_DONE:
      if (ZB_BDB().bdb_commissioning_status != ZB_BDB_STATUS_NO_NETWORK)
      {
        /* Probably bdb device failed to update tclk exchange */
        bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH, param);
      }
      else
      {
        /* Will be here after bdb device failed authentication - lets join to
         * other networks */
#ifdef ZB_JOIN_CLIENT
        ZB_SCHEDULE_CALLBACK(zdo_retry_joining, 0);
#endif
      }
      break;

    case BDB_COMM_SIGNAL_NETWORK_STEERING_TCLK_DONE:
      TRACE_MSG(TRACE_ZDO1, "TCLK succeeded", (FMT__0));
      ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_IN_PROGRESS;
      /* Join is finished - can reset the blacklist. */
#ifdef ZB_NWK_BLACKLIST
      zb_nwk_blacklist_reset();
#endif
      /* TRICKY: Run steering one more time - to broadcast Permit Joining. This time it will be
       * steering for device on network. */
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_STEERING_START, param);
      break;

    case BDB_COMM_SIGNAL_NETWORK_STEERING_FINISH:
      /* Do not stop even if steering is failed - may continue with Formation if it is allowed
       * (Distributed ZR case) */

      /* if (ZB_BDB().bdb_commissioning_status != ZB_BDB_STATUS_IN_PROGRESS) */
      /* { */
      /*   ZB_BDB().bdb_commissioning_step = ZB_BDB_COMMISSIONING_STOP; */
      /*   bdb_commissioning_signal(BDB_COMM_SIGNAL_FINISH, param); */
      /* } */
      /* else */
      {
        /* Go to next commissioning step - Network Steering procedure */
        TRACE_MSG(TRACE_ZDO1, "Network Steering finished, next step - Network Formation", (FMT__0));
        ZB_BDB().bdb_commissioning_step = ZB_BDB_NETWORK_FORMATION;
        bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_FORMATION_START, param);
      }
      break;

    default:
      TRACE_MSG(TRACE_ERROR, "Unknown signal %d", (FMT__D, BDB_COMM_CTX().signal));
#ifdef BDB_COMMISSIONING_DEBUG
      ZB_ASSERT(0);
#else
      if (param)
      {
        zb_buf_free(param);
      }
#endif
      break;
  }
}


static void bdb_network_formation_machine(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_network_formation_machine param %hd signal %hd state %hd",
            (FMT__H_H_H, param, BDB_COMM_CTX().signal, ZB_BDB().bdb_commissioning_step));
  switch (BDB_COMM_CTX().signal)
  {
    case BDB_COMM_SIGNAL_NETWORK_FORMATION_START:
    {
      if (ZB_BDB().bdb_commissioning_mode & ZB_BDB_NETWORK_FORMATION)
      {
        /* TODO: Fix this correctly after discussion about signals. */
        if (ZB_BDB().bdb_commissioning_mode == ZB_BDB_NETWORK_FORMATION)
        {
          ZB_BDB().bdb_application_signal = ZB_BDB_SIGNAL_FORMATION;
        }

        if (bdb_can_do_network_formation())
        {
          TRACE_MSG(TRACE_ZDO1, "Start Network Formation", (FMT__0));
          TRACE_MSG(TRACE_ZDO1, "Start BDB formation for device type %d", (FMT__D, zb_get_device_type()));
          ZB_ASSERT(FORMATION_SELECTOR().bdb_formation);
          ZB_SCHEDULE_CALLBACK(FORMATION_SELECTOR().bdb_formation, param);
        }
        else
        {
          TRACE_MSG(TRACE_ZDO1, "Can not do Network Formation, skip", (FMT__0));
          bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_FORMATION_FINISH, param);
        }
      }
      else
      {
        TRACE_MSG(TRACE_ZDO1, "Skip Network Formation", (FMT__0));
        bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_FORMATION_FINISH, param);
      }
    }
    break;

    case BDB_COMM_SIGNAL_NWK_FORMATION_OK:
      TRACE_MSG(TRACE_ZDO1, "Network is successfully formed", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NETWORK_FORMATION_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_NETWORK_FORMATION_FINISH:
    default:
      if (ZB_BDB().bdb_commissioning_status != ZB_BDB_STATUS_IN_PROGRESS)
      {
        ZB_BDB().bdb_commissioning_step = ZB_BDB_COMMISSIONING_STOP;
        TRACE_MSG(TRACE_ZDO3, "ZB_BDB_COMMISSIONING_STOP", (FMT__0));
        bdb_commissioning_signal(BDB_COMM_SIGNAL_FINISH, param);
      }
      else
      {
        TRACE_MSG(TRACE_ZDO1, "Network Formation finished, next step - Finding & Binding", (FMT__0));
        /* Go to next commissioning step - finish */
#if defined(ZB_BDB_ENABLE_FINDING_BINDING)
        ZB_BDB().bdb_commissioning_step = ZB_BDB_FINDING_N_BINDING;
        bdb_commissioning_signal(BDB_COMM_SIGNAL_FINDING_N_BINDING_START, param);
#endif
      }
      break;
  }
}

#if defined(ZB_BDB_ENABLE_FINDING_BINDING)
static void bdb_finding_n_binding_machine(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_finding_n_binding_machine param %hd signal %hd state %hd",
            (FMT__H_H_H, param, BDB_COMM_CTX().signal, ZB_BDB().bdb_commissioning_step));
  switch (BDB_COMM_CTX().signal)
  {
    case BDB_COMM_SIGNAL_FINDING_N_BINDING_START:
    {
      /* if (ZB_BDB().bdb_commissioning_mode & ZB_BDB_FINDING_N_BINDING) */
      /* { */
      /*   TRACE_MSG(TRACE_ZDO1, "TODO: Start Finding & Binding", (FMT__0)); */
      /* } */
      /* else */
      {
        TRACE_MSG(TRACE_ZDO1, "Skip Finding & Binding", (FMT__0));
        bdb_commissioning_signal(BDB_COMM_SIGNAL_FINDING_N_BINDING_FINISH, param);
      }
    }
    break;

    case BDB_COMM_SIGNAL_FINDING_N_BINDING_FINISH:
      TRACE_MSG(TRACE_ZDO1, "Finding & Binding finished, finish BDB commissioning", (FMT__0));
      /* Go to next commissioning step - finish */
      ZB_BDB().bdb_commissioning_step = ZB_BDB_COMMISSIONING_STOP;
      TRACE_MSG(TRACE_ZDO3, "ZB_BDB_COMMISSIONING_STOP", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_LEAVE_DONE:
      TRACE_MSG(TRACE_ZDO1, "Leave done, abort BDB commissioning", (FMT__0));

      switch (BDB_COMM_CTX().state)
      {
        case ZB_BDB_COMM_IDLE:
          /* Do nothing in this case, BDB_COMM_CTX should be already reset */
          break;

        case ZB_BDB_COMM_FINDING_AND_BINDING:
          zb_bdb_finding_binding_initiator_cancel();
          break;

        case ZB_BDB_COMM_FINDING_AND_BINDING_TARGET:
          zb_bdb_finding_binding_target_cancel();
          break;

        default:
          /* NOTE: other COMM states are not currently used, so handle them when it will be needed */
          ZB_ASSERT(ZB_FALSE);
          break;
      }

      break;

    default:
      TRACE_MSG(TRACE_ERROR, "Unknown signal %d", (FMT__D, BDB_COMM_CTX().signal));

      TRACE_MSG(TRACE_ERROR, "Ignore (F&B implemented separately)", (FMT__0));
      if (param)
      {
        zb_buf_free(param);
      }
      break;
  }
}
#endif

#ifdef ZB_ZCL_ENABLE_WWAH_SERVER
static zb_bool_t bdb_ds_filter_cb(zb_uint_t i)
{
  switch (i)
  {
    case ZB_NVRAM_COMMON_DATA:
      return ZB_TRUE;
    default:
      return ZB_FALSE;
  }
}

/* Savepoint logic for SE R&R and WWAH Rejoin could be unified and moved into a separate module.
   For NCP case it could be put at SoC side */
static void bdb_savepoint_before_rr()
{
  TRACE_MSG(TRACE_ZDO1, "bdb_savepoint_before_rr", (FMT__0));

  BDB_COMM_CTX().rejoin.rr_sv_authenticated = ZG->aps.authenticated;
  BDB_COMM_CTX().rejoin.rr_sv_device_type = zb_get_device_type();

  /* Save parent info. We do not want to reload it from nvram because neighbor table load supposes clear neighbor table. */
  BDB_COMM_CTX().rejoin.rr_sv_parent_short = 0xffff;
  if (ZG->nwk.handle.parent != (zb_uint8_t)-1)
  {
    zb_neighbor_tbl_ent_t *nent = NULL;
    zb_uint16_t addr = BDB_COMM_CTX().rejoin.rr_sv_parent_short;
    zb_address_by_ref(BDB_COMM_CTX().rejoin.rr_sv_parent_long, &addr, ZG->nwk.handle.parent);
    if (zb_nwk_neighbor_get(ZG->nwk.handle.parent, ZB_FALSE, &nent) == RET_OK)
    {
      BDB_COMM_CTX().rejoin.rr_sv_parent_nent = *nent;
    }
  }
  /* All other information is now in our nvram. We can restore it from there. */
}

static void bdb_restore_cont(zb_uint8_t param)
{
  /* Undo effect of zdo_rejoin_clear_prev_join(). */
  /* If ZED, start poll again */
  if (ZB_IS_DEVICE_ZED())
  {
    if (!ZB_PIBCACHE_RX_ON_WHEN_IDLE())
    {
      zb_zdo_pim_start_poll(0);
    }
    zb_buf_free(param);
  }
  else
  {
    BDB_COMM_CTX().rejoin.rr_ignore_start_router_conf = 1;
    ZB_SCHEDULE_CALLBACK(zb_zdo_start_router, param);
  }
}

static void bdb_restore_saved_rr(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_restore_saved_rr param %hd", (FMT__H, param));

#ifdef ZB_USE_NVRAM
  /* get common dataset from nvram */
  ZB_NVRAM().ds_filter_cb = bdb_ds_filter_cb;
  zb_nvram_load();
  ZB_NVRAM().ds_filter_cb = NULL;
#endif

  if (BDB_COMM_CTX().rejoin.rr_sv_parent_short != 0xffff)
  {
    /* Create and lock parent */
    zb_address_update(BDB_COMM_CTX().rejoin.rr_sv_parent_long, BDB_COMM_CTX().rejoin.rr_sv_parent_short, ZB_FALSE, &ZG->nwk.handle.parent);
  }
  else
  {
    /* Restore ZG->nwk.handle.parent from BDB_COMM_CTX().rejoin.rr_sv_parent_long */
    zb_address_by_ieee(BDB_COMM_CTX().rejoin.rr_sv_parent_long, ZB_FALSE, ZB_FALSE, &ZG->nwk.handle.parent);
  }

  {
    zb_neighbor_tbl_ent_t *nent = NULL;
    if (zb_nwk_neighbor_get(ZG->nwk.handle.parent, ZB_TRUE, &nent) == RET_OK)
    {
      nent->relationship = ZB_NWK_RELATIONSHIP_PARENT;
    }
  }

  {
    zb_address_ieee_ref_t ref;
    zb_address_update(ZB_PIBCACHE_EXTENDED_ADDRESS(),
                      ZB_PIBCACHE_NETWORK_ADDRESS(),
                      ZB_TRUE, &ref);
  }
  ZB_NIB().device_type = (zb_nwk_device_type_t)BDB_COMM_CTX().rejoin.rr_sv_device_type;
  ZG->aps.authenticated = BDB_COMM_CTX().rejoin.rr_sv_authenticated;
  ZB_JOINED() = 1;
#ifdef ZB_USE_NVRAM
  zb_nvram_store_addr_n_nbt();
#endif

  /* Now pibcache is filled by right values, but values are not pushed to MAC. Switch MAC channel etc. */
  zb_nwk_sync_pibcache_with_mac(param, bdb_restore_cont);
}
#endif  /* #ifdef ZB_ZCL_ENABLE_WWAH_SERVER */

/* WWAH Rejoin procedure:
   1. Store existing network settings.
   [Continue if Sleepy ED and join was NOT caused by Poll Control Check-in failure, else go to step
   5]
   2. Secure Rejoin on the Current Channel.
   3. TC Rejoin on the Current Channel.
   4. (optional, now skip) TC Rejoin on preferred channel list
   [<<<]
   [Continue if device is using unique TCLK, else go to step 8]
   5. TC Rejoin on all channels
   6. Back to previous network settings and
      Wait for X seconds or until additional triggers (example: user button push)
      a. Non-sleepy ED: X seconds will be equal to the Keepalive cluster value or 24 hours if
      Keepalive cluster is not Implemented
      b. Sleepy ED: if WWAH rejoin algorithm is implemented then hte wait will be determined by the
      WWAH rejoin algorithm, else 15 minutes
   7. Non-sleepy ED or ZR - back to step 3.
   [These steps are only for Sleepy ED]
   8. Secure Rejoin on all channels.
   9. Back to previous network settings and
      Wait for X seconds or until additional triggers
   10. Go to step 1.

   *When performing a rejoin on multiple channels, the device shall first perform a beacon survey on
   the given set of channels, sort the resulting beacons using the WWAH prioritized parent
   algorithm and prefer beacons with a higher network update ID when more than one share the same
   WWAH parent priority.
 */
/* TODO: implement all rejoin triggers for ZR and ZED */
#ifdef ZB_ZCL_ENABLE_WWAH_SERVER
static void bdb_retry_rejoin_recovery(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_retry_rejoin_recovery param %hd", (FMT__H, param));
  if (!param)
  {
    zb_buf_get_out_delayed(bdb_retry_rejoin_recovery);
  }
  else
  {
    BDB_COMM_CTX().rejoin.waiting = 0;
    bdb_commissioning_signal(BDB_COMM_CTX().rejoin.next_rejoin_signal, param);
  }
}

static void bdb_rejoin_machine(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "bdb_rejoin_machine param %hd signal %hd state %hd",
            (FMT__H_H_H, param, BDB_COMM_CTX().signal, ZB_BDB().bdb_commissioning_step));

  switch (BDB_COMM_CTX().signal)
  {
    case BDB_COMM_SIGNAL_REJOIN_START:
    {
      TRACE_MSG(TRACE_ZDO1, "Start Rejoin", (FMT__0));

      /* Double-check this should be stopped during Rejoin */
      zb_zcl_wwah_stop_periodic_checkin();
      zb_zcl_wwah_stop_bad_parent_recovery();

      if (!BDB_COMM_CTX().rejoin.rr_skip_savepoint)
      {
        bdb_savepoint_before_rr();
      }

      BDB_COMM_CTX().rejoin.rr_have_unique_tclk = (zb_secur_get_link_key_by_address(ZB_PIBCACHE_EXTENDED_ADDRESS(), ZB_SECUR_VERIFIED_KEY) != NULL);

#ifdef ZB_NWK_BLACKLIST
      zb_nwk_blacklist_reset();
#endif

      if (!ZB_PIBCACHE_RX_ON_WHEN_IDLE() &&
          !BDB_COMM_CTX().rejoin.rejoin_by_checkin_failure)
      {
        bdb_commissioning_signal(BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_CURRENT_CHANNEL, 0);
        BDB_COMM_CTX().rejoin.last_rejoin_signal = BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_CURRENT_CHANNEL;
      }
      else
      {
        bdb_commissioning_signal(BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_ALL_CHANNELS, 0);
        BDB_COMM_CTX().rejoin.last_rejoin_signal = BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_ALL_CHANNELS;
      }

      /* Inform application that rejoin is started */
      zb_app_signal_pack(param, ZB_BDB_SIGNAL_WWAH_REJOIN_STARTED,
                         RET_OK,
                         0);
      ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
    }
    break;

    case BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_CURRENT_CHANNEL:
    case BDB_COMM_SIGNAL_REJOIN_TRY_TC_REJOIN_ON_CURRENT_CHANNEL:
    case BDB_COMM_SIGNAL_REJOIN_TRY_TC_REJOIN_ON_ALL_CHANNELS:
    case BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_ALL_CHANNELS:
    {
      zb_channel_list_t channel_list;
      zb_bool_t current_channel =
        (zb_bool_t)(BDB_COMM_CTX().signal == BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_CURRENT_CHANNEL ||
         BDB_COMM_CTX().signal == BDB_COMM_SIGNAL_REJOIN_TRY_TC_REJOIN_ON_CURRENT_CHANNEL);
      zb_bool_t secure =
        (zb_bool_t)(BDB_COMM_CTX().signal == BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_CURRENT_CHANNEL ||
         BDB_COMM_CTX().signal == BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_ALL_CHANNELS);

      if (current_channel)
      {
        TRACE_MSG(TRACE_ZDO1, "Try rejoin over current page %hd channel %hd",
                  (FMT__H_H, ZB_PIBCACHE_CURRENT_PAGE(), ZB_PIBCACHE_CURRENT_CHANNEL()));
        zb_channel_list_init(channel_list);
        zb_channel_page_list_set_logical_channel(channel_list, ZB_PIBCACHE_CURRENT_PAGE(),
                                                 ZB_PIBCACHE_CURRENT_CHANNEL());
      }
      else
      {
        /* FIXME: do we need to use preferred channels here? */
        TRACE_MSG(TRACE_ZDO1, "Try rejoin over all channels", (FMT__0));
        zb_channel_page_list_copy(channel_list, ZB_AIB().aps_channel_mask_list);
      }

      if (secure)
      {
        ZG->aps.authenticated = 1;
      }

      TRACE_MSG(TRACE_ZDO1, "rejoin secure %d", (FMT__D, secure));
      BDB_COMM_CTX().rejoin.last_rejoin_signal = BDB_COMM_CTX().signal;

#if defined ZB_DISTRIBUTED_SECURITY_ON
      if (!secure && IS_DISTRIBUTED_SECURITY())
      {
        /* WWAH spec, 11.4 Parent and Hub Connectivity Recovery Procedure
           Please note that for Distributed security networks, TC Rejoin is not supported. As a
           result, for a device rejoining a Distributed security network, all steps involving TC
           Rejoin shall be skipped.
         */
        /* Skip rejoin by imitation that current step is failed  */
        bdb_commissioning_signal(BDB_COMM_SIGNAL_NWK_JOIN_FAILED, param);
      }
      else
#endif
      {
        zb_ext_pan_id_t ext_pan_id;
        zb_get_extended_pan_id(ext_pan_id);

        zdo_initiate_rejoin(param, ext_pan_id,
                              channel_list,
                              (zb_bool_t)secure);
      }
    }
    break;

    case BDB_COMM_SIGNAL_NWK_AUTH_FAILED:
      TRACE_MSG(TRACE_ZDO1, "Failed NWK authentication", (FMT__0));
      /* FALLTHROUGH */
    case BDB_COMM_SIGNAL_LEAVE_DONE: /* Is it possible? */
    case BDB_COMM_SIGNAL_NWK_JOIN_FAILED:
    {
      zb_bool_t need_to_wait = ZB_FALSE;

      TRACE_MSG(TRACE_ZDO1, "Failed rejoin", (FMT__0));

      switch (BDB_COMM_CTX().rejoin.last_rejoin_signal)
      {
        case BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_CURRENT_CHANNEL: /* Step 2 */
          /* -> Step 3 */
          BDB_COMM_CTX().rejoin.next_rejoin_signal =
            BDB_COMM_SIGNAL_REJOIN_TRY_TC_REJOIN_ON_CURRENT_CHANNEL;
          break;
        case BDB_COMM_SIGNAL_REJOIN_TRY_TC_REJOIN_ON_CURRENT_CHANNEL: /* Step 3 */
        {
          if (BDB_COMM_CTX().rejoin.rr_have_unique_tclk)
          {
            /* Have unique TCLK: -> Step 5 */
            BDB_COMM_CTX().rejoin.next_rejoin_signal =
              BDB_COMM_SIGNAL_REJOIN_TRY_TC_REJOIN_ON_ALL_CHANNELS;
          }
          else
          {
            /* Do not have unique TCLK */
            need_to_wait = ZB_TRUE;

            if (ZB_PIBCACHE_RX_ON_WHEN_IDLE())
            {
              /* RX-ON: Wait, than back to step 3 */
              BDB_COMM_CTX().rejoin.next_rejoin_signal =
                BDB_COMM_SIGNAL_REJOIN_TRY_TC_REJOIN_ON_CURRENT_CHANNEL;
            }
            else
            {
              /* Non-sleepy ED: Continue to step 8 */
              BDB_COMM_CTX().rejoin.next_rejoin_signal =
                BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_ALL_CHANNELS;
            }
          }
        }
        break;
        case BDB_COMM_SIGNAL_REJOIN_TRY_TC_REJOIN_ON_ALL_CHANNELS:  /* Step 5 */
        {
          /* -> Step 6 */
          need_to_wait = ZB_TRUE;

          if (ZB_PIBCACHE_RX_ON_WHEN_IDLE())
          {
            /* Non-sleepy ED: Back to step 3 */
            BDB_COMM_CTX().rejoin.next_rejoin_signal =
              BDB_COMM_SIGNAL_REJOIN_TRY_TC_REJOIN_ON_CURRENT_CHANNEL;
          }
          else
          {
            /* Non-sleepy ED: Continue to step 8 */
            BDB_COMM_CTX().rejoin.next_rejoin_signal =
              BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_ALL_CHANNELS;
          }
        }
        break;
        case BDB_COMM_SIGNAL_REJOIN_TRY_SECURE_REJOIN_ON_ALL_CHANNELS:  /* Step 8 */
        {
          /* -> Step 9-10  */
          need_to_wait = ZB_TRUE;
          BDB_COMM_CTX().rejoin.next_rejoin_signal =
            BDB_COMM_SIGNAL_REJOIN_START;
        }
        break;
        default:
          ZB_ASSERT(0);         /* debug */
          break;
      }

      if (need_to_wait)
      {
        zb_time_t time_to_wait = 0;

        TRACE_MSG(TRACE_ZDO1, "skip savepoint %hd", (FMT__H, !BDB_COMM_CTX().rejoin.rr_skip_savepoint));

        if (ZB_PIBCACHE_RX_ON_WHEN_IDLE())
        {
          /* For Routers and Non-Sleepy End Devices, X seconds will be equal to the Keepalive
           * cluster value of 24 hours if Keepalive cluster is not implemented. */
          if (WWAH_CTX().periodic_checkins.poll_method == ZB_ZCL_WWAH_PERIODIC_CHECKINS_READ_KEEPALIVE)
          {
            time_to_wait = (WWAH_CTX().periodic_checkins.keepalive_base +
                            ZB_RANDOM_VALUE(WWAH_CTX().periodic_checkins.keepalive_jitter)) * ZB_TIME_ONE_SECOND;
          }
          else
          {
            time_to_wait = ZB_ZCL_WWAH_REJOIN_NON_SLEEPY_ED_ROUTER_TIMEOUT;
          }
        }
        else
        {
          ++BDB_COMM_CTX().rejoin.rr_global_retries;

#ifdef ZB_JOIN_CLIENT
          if (zb_zcl_wwah_get_rejoin_tmo(BDB_COMM_CTX().rejoin.rr_global_retries, &time_to_wait) == RET_EXIT)
          {
            /* Wait for next Full Rejoin cycle */
            BDB_COMM_CTX().rejoin.rr_global_retries = 0;
          }
#endif
        }

        if (!BDB_COMM_CTX().rejoin.rr_skip_savepoint)
        {
          ZB_SCHEDULE_CALLBACK(bdb_restore_saved_rr, param);
        }
        TRACE_MSG(TRACE_ZDO1, "retry rejoin, time_to_wait=%ld", (FMT__L, time_to_wait));
        BDB_COMM_CTX().rejoin.waiting = 1;
        ZB_SCHEDULE_ALARM(bdb_retry_rejoin_recovery, 0, time_to_wait);
      }
      else
      {
        ZB_SCHEDULE_CALLBACK(bdb_retry_rejoin_recovery, param);
      }
    }
    break;

    case BDB_COMM_SIGNAL_NWK_START_ROUTER_CONF:
      if (BDB_COMM_CTX().rejoin.waiting)
      {
        /* Ignore START_ROUTER_CONF when it is triggered by bdb_restore_saved_rr(). */
        break;
      }
      /* FALLTHROUGH */
    case BDB_COMM_SIGNAL_NWK_JOIN_DONE:
      TRACE_MSG(TRACE_ZDO1, "Device is joined", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_REJOIN_FINISH, param);
      break;

    case BDB_COMM_SIGNAL_REJOIN_FINISH:
      TRACE_MSG(TRACE_ZDO1, "Rejoin finished, finish BDB commissioning", (FMT__0));
      /* Go to next commissioning step - finish */
      ZB_BDB().bdb_commissioning_step = ZB_BDB_COMMISSIONING_STOP;
      TRACE_MSG(TRACE_ZDO3, "ZB_BDB_COMMISSIONING_STOP", (FMT__0));
      bdb_commissioning_signal(BDB_COMM_SIGNAL_FINISH, param);
      break;

    default:
      TRACE_MSG(TRACE_ERROR, "Unknown signal %d", (FMT__D, BDB_COMM_CTX().signal));

      TRACE_MSG(TRACE_ERROR, "Ignore (F&B implemented separately)", (FMT__0));
      if (param)
      {
        zb_buf_free(param);
      }
      break;
  }
}

#endif /* ZB_ZCL_ENABLE_WWAH_SERVER */

void bdb_start_rejoin_recovery(zb_uint8_t param, zb_uint16_t user_param)
{
#ifdef ZB_ZCL_ENABLE_WWAH_SERVER
  TRACE_MSG(TRACE_ZDO1, ">> bdb_start_rejoin_recovery: param %hd user_param %d", (FMT__H_D, param, user_param));
  if (!param)
  {
    zb_buf_get_out_delayed((zb_callback_t)bdb_start_rejoin_recovery);
  }
  else
  {
    zb_uint8_t rejoin_reason = ZB_GET_LOW_BYTE(user_param);

    BDB_COMM_CTX().rejoin.rejoin_by_checkin_failure =
      (rejoin_reason == BDB_COMM_REJOIN_REASON_POLL_CONTROL_CHECK_IN);

    BDB_COMM_CTX().rejoin.rr_global_retries = 0;

    ZB_BDB().bdb_commissioning_step = ZB_BDB_REJOIN;
    bdb_commissioning_signal(BDB_COMM_SIGNAL_REJOIN_START, param);
  }
  TRACE_MSG(TRACE_ZDO1, "<< bdb_start_rejoin_recovery", (FMT__0));
#else
  ZVUNUSED(user_param);
  TRACE_MSG(TRACE_ERROR, "bdb_start_rejoin_recovery: not supported!", (FMT__0));
  if (param)
  {
    zb_buf_free(param);
  }
#endif
}

#endif /* ifndef BDB_OLD */
