/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*  PURPOSE: Classiz ZB commissioning: defined in Core specification and HA
*/


#define ZB_TRACE_FILE_ID 67
#include "zb_common.h"

#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
#include "zb_scheduler.h"
#include "zb_bufpool.h"
#include "zb_nwk.h"
#include "zb_nwk_nib.h"
#include "zb_aps.h"
#include "zb_zdo.h"
#include "zb_secur.h"
#include "zb_secur_api.h"
#include "zb_nvram.h"
#include "zb_bdb_internal.h"


void zdo_classic_initiate_commissioning(zb_uint8_t param)
{
  zb_ext_pan_id_t use_ext_pan_id;

  TRACE_MSG(TRACE_ZDO1, "zdo_classic_initiate_commissioning param %hd device type %d",
            (FMT__H_D_D, param, zb_get_device_type()));

  ZB_ASSERT(zb_get_device_type() != ZB_NWK_DEVICE_TYPE_NONE);

  zb_get_use_extended_pan_id(use_ext_pan_id);

  /* a. Maybe Formation */
  if (ZB_IS_DEVICE_ZC()
#if defined ZB_ROUTER_ROLE && defined ZB_DISTRIBUTED_SECURITY_ON
      /* Not joned ZR with TC addr -1 means: start Distributed formation.*/
      || (zb_tc_is_distributed()
          && ZB_IS_DEVICE_ZR()
          && !zb_zdo_joined())
#endif
    )
  {
#if defined ZB_FORMATION
    ZB_ASSERT(FORMATION_SELECTOR().start_formation);
    ZB_SCHEDULE_CALLBACK(FORMATION_SELECTOR().start_formation, param);
#else
    TRACE_MSG(TRACE_MAC1, "Coordinator role is not supported", (FMT__0));
    zb_buf_free(param);
#endif
  }
  /* Rejoin after reboot? */
#ifdef ZB_JOIN_CLIENT
  else if (!ZB_EXTPANID_IS_ZERO(ZB_NIB_EXT_PAN_ID()))
  {
    /* try to rejoin the current network:
       use apsChannelMask as channel mask and do unsecure rejoin
       if we are not authenticated.
    */
    zdo_initiate_rejoin(param, ZB_NIB_EXT_PAN_ID(),
                        ZB_AIB().aps_channel_mask_list,
                        zb_zdo_authenticated());
  }
  /* Rejoin to the net defined by aps_use_extended_pan_id */
  else if (!ZB_EXTPANID_IS_ZERO(use_ext_pan_id))
  {
    /* If the device is not the designated coordinator and apsUseExtendedPANID
       has a non-zero value, the device should attempt to rejoin the network
       specified in apsUseExtendedPANID. To do this, it should use NLME-JOIN.request
       with the ExtendedPANID parameter equal to the value of apsUseExtendedPANID, the
       ScanChannels parameter of the primitive equal to the value of the apsChannelMask
       configuration attribute. The RejoinNetwork parameter of the NLME-JOIN.request
       primitive should have a value of 0x02 indicating rejoin. */
    zdo_initiate_rejoin(param, use_ext_pan_id,
                        ZB_AIB().aps_channel_mask_list,
                        ZB_FALSE);
  }
  else
  {
    /* ZR: discovery, then join */
    zb_nlme_network_discovery_request_t *req = ZB_BUF_GET_PARAM(param, zb_nlme_network_discovery_request_t);

    zb_channel_page_list_copy(req->scan_channels_list, ZB_AIB().aps_channel_mask_list);
    req->scan_duration = ZB_DEFAULT_SCAN_DURATION; /* TODO: configure it somehow? */
    TRACE_MSG(TRACE_APS1, "discovery, then join by association, scan attempts %hd", (FMT__H, COMM_CTX().discovery_ctx.nwk_scan_attempts));

    COMM_CTX().discovery_ctx.disc_count = COMM_CTX().discovery_ctx.nwk_scan_attempts;
    ZB_SCHEDULE_CALLBACK(zb_nlme_network_discovery_request, param);
  }
#endif  /* ZB_JOIN_CLIENT */
}

#ifdef ZB_JOIN_CLIENT
void zdo_restart_association(zb_uint8_t param)
{
  zb_nlme_network_discovery_request_t *req = ZB_BUF_GET_PARAM(param, zb_nlme_network_discovery_request_t);
  TRACE_MSG(TRACE_ERROR, "rejoin failed - try association", (FMT__0));

  ZB_ASSERT(zb_get_device_type() != ZB_NWK_DEVICE_TYPE_NONE);

  TRACE_MSG(TRACE_ZDO1, "Dev type %hd", (FMT__H, zb_get_device_type()));
  ZG->zdo.handle.rejoin = 0;

  zb_channel_page_list_copy(req->scan_channels_list, ZB_AIB().aps_channel_mask_list);
  req->scan_duration = ZB_DEFAULT_SCAN_DURATION; /* TODO: configure it somehow? */
  COMM_CTX().discovery_ctx.disc_count = COMM_CTX().discovery_ctx.nwk_scan_attempts;
  ZB_SCHEDULE_CALLBACK(zb_nlme_network_discovery_request, param);
}
#endif  /* ZB_JOIN_CLIENT */

#endif  /* #ifndef ZB_LITE_BDB_ONLY_COMMISSIONING */
