/*
 * ZBOSS Zigbee software protocol stack
 *
 * Copyright (c) 2012-2020 DSR Corporation, Denver CO, USA.
 * http://www.dsr-zboss.com
 * http://www.dsr-corporation.com
 * All rights reserved.
 *
 *
 * Use in source and binary forms, redistribution in binary form only, with
 * or without modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 2. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 3. This software must only be used in or with a processor manufactured by Nordic
 *    Semiconductor ASA, or in or with a processor manufactured by a third party that
 *    is used in combination with a processor manufactured by Nordic Semiconductor.
 *
 * 4. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* PURPOSE: Formation code moved there from zdo_app.c
   For NCP this file is compiled in SoC and Host-side mode with different defines.
*/

#define ZB_TRACE_FILE_ID 53
#include "zb_common.h"
#include "zb_scheduler.h"
#include "zb_bufpool.h"
#include "zb_nwk.h"
#include "zb_nwk_nib.h"
#include "zb_aps.h"
#include "zb_zdo.h"
#include "zdo_common.h"
#include "zb_secur.h"
#include "zb_secur_api.h"
#include "zb_nvram.h"
#include "zb_bdb_internal.h"
#include "zb_watchdog.h"
#include "zb_ncp.h"
#include "zdo_wwah_parent_classification.h"

#if defined ZB_ENABLE_ZLL
#include "zll/zb_zll_common.h"
#include "zll/zb_zll_nwk_features.h"
#endif /* defined ZB_ENABLE_ZLL */

#if defined ZB_HA_SUPPORT_EZ_MODE
#include "ha/zb_ha_ez_mode_comissioning.h"
#endif

#if defined ZB_ENABLE_SE
#include "zb_se.h"
#endif /* defined ZB_ENABLE_SE */

/*! \addtogroup ZB_ZDO */
/*! @{ */

#ifdef ZB_FORMATION

static void zdo_start_formation(zb_uint8_t param);

#ifdef ZB_BDB_MODE

#if !defined NCP_MODE || defined NCP_MODE_HOST || defined ZB_CONFIGURABLE_COMMISSIONING_TYPE
static void bdb_formation(zb_uint8_t param);
static zb_uint32_t bdb_commissioning_formation_channels_mask(void);

#endif /* !NCP_MODE || NCP_MODE_HOST */

#endif /* ZB_BDB_MODE */

#if !defined NCP_MODE || defined NCP_MODE_HOST || defined ZB_CONFIGURABLE_COMMISSIONING_TYPE
static void zdo_commissioning_formation_channels_mask(zb_channel_list_t list);
#endif

static void zdo_comm_set_permit_join_after_formation(zb_uint8_t param);

void zdo_formation_force_link(void)
{
  TRACE_MSG(TRACE_ZDO1, "zdo_formation_force_link", (FMT__0));

#ifndef NCP_MODE_HOST
  /* Link NWK-level formation if this is not NCP host */
  zb_nwk_formation_force_link();

  APS_SELECTOR().authhenticate_child_directly = secur_authenticate_child_directly;
#endif /* !defined NCP_MODE_HOST */

#if !defined NCP_MODE || defined NCP_MODE_HOST || defined ZB_CONFIGURABLE_COMMISSIONING_TYPE

#ifdef ZB_BDB_MODE
  FORMATION_SELECTOR().bdb_formation = bdb_formation;
#endif /* ZB_BDB_MODE */

  FORMATION_SELECTOR().start_formation = zdo_start_formation;

#endif  /* !defined NCP_MODE || defined NCP_MODE_HOST */
}

/*
  Called directly in monolythic build, via serial proto at NCP host.
 */
void zb_nlme_network_formation_confirm(zb_uint8_t param)
{
  TRACE_MSG(TRACE_INFO1, "zb_nlme_network_formation_confirm %hd status %hd",
            (FMT__H_H, param, zb_buf_get_status(param)));

  if (zb_buf_get_status(param) != RET_OK)
  {
    ZB_SCHEDULE_CALLBACK(zdo_commissioning_formation_failed, param);
  }
  else
  {
    zb_address_ieee_ref_t addr_ref;

    /* Be sure my address exists in the translation table */
    zb_address_by_short(ZB_PIBCACHE_NETWORK_ADDRESS(), ZB_TRUE, ZB_FALSE, &addr_ref);

    secur_tc_init();

#ifdef ZB_USE_NVRAM
    zb_nvram_write_dataset(ZB_NVRAM_COMMON_DATA);
#endif

    TRACE_MSG(TRACE_ZDO3, "Schedule zdo_send_parent_annce from zb_nlme_network_formation_confirm", (FMT__0));
    ZB_SCHEDULE_CALLBACK(zdo_send_parent_annce_at_formation, 0);
#if (defined ZB_PARENT_CLASSIFICATION) && !(defined ZB_ED_ROLE)
    nwk_set_tc_connectivity(1);
#endif
    ZB_SCHEDULE_CALLBACK(zdo_comm_set_permit_join_after_formation, param);
  }
}


static void zdo_comm_set_permit_join_after_formation(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO2, ">>set_permit_join_after_formation %hd max_children %hd",
            (FMT__H_H, param, zb_get_max_children()));

  zdo_comm_set_permit_join(param, zdo_commissioning_formation_done);

  TRACE_MSG(TRACE_ZDO2, "<<set_permit_join_after_formation", (FMT__0));
}


void zdo_commissioning_formation_done(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO1, "zdo_commissioning_formation_done param %hd", (FMT__H, param));

  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      /* continue formation */
      bdb_commissioning_signal(BDB_COMM_SIGNAL_NWK_FORMATION_OK, param);
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      zb_app_signal_pack(param, ZB_ZDO_SIGNAL_DEFAULT_START, zb_buf_get_status(param), 0);
      ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      /* continue formation */
      se_commissioning_signal(SE_COMM_SIGNAL_NWK_FORMATION_OK, param);
      break;
#endif
#if defined NCP_MODE && !defined NCP_MODE_HOST
    case ZB_COMMISSIONING_NCP:
      /* NCP is a special case: after NWL reset and PIB load do nothing. Let Host define. */
      ncp_signal(NCP_SIGNAL_NWK_FORMATION_OK, param);
      break;
#endif
    default:
      break;
  }
}


void zdo_commissioning_formation_failed(zb_uint8_t param)
{
  switch (ZB_COMMISSIONING_TYPE())
  {
#ifdef ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      TRACE_MSG(TRACE_ZDO1, "Formation failed - let BDB decide", (FMT__0));
      ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_FORMATION_FAILURE;
      ZB_SCHEDULE_CALLBACK(bdb_commissioning_machine, param);
      break;
#endif
#ifdef ZB_COMMISSIONING_CLASSIC_SUPPORT
    case ZB_COMMISSIONING_CLASSIC:
      zb_app_signal_pack(param, ZB_ZDO_SIGNAL_DEFAULT_START, zb_buf_get_status(param), 0);
      ZB_SCHEDULE_CALLBACK(zb_zdo_startup_complete_int, param);
      break;
#endif
#ifdef ZB_SE_COMMISSIONING
    case ZB_COMMISSIONING_SE:
      TRACE_MSG(TRACE_SECUR3, "SE - formation failed", (FMT__0));
      se_commissioning_signal(SE_COMM_SIGNAL_NWK_FORMATION_FAILED, param);
      break;
#endif
#if defined NCP_MODE && !defined NCP_MODE_HOST
    case ZB_COMMISSIONING_NCP:
      /* NCP is a special case: after NWL reset and PIB load do nothing. Let Host define. */
      ncp_signal(NCP_SIGNAL_NWK_FORMATION_FAILED, param);
      break;
#endif
    default:
      break;
  }
}


/* High-level commissioning is at Host side of NCP */
#if !defined NCP_MODE || defined NCP_MODE_HOST || defined ZB_CONFIGURABLE_COMMISSIONING_TYPE

static void zdo_start_formation(zb_uint8_t param)
{
  /*! [zb_nlme_network_formation_request] */
  /* will start as coordinator: Formation */
  zb_nlme_network_formation_request_t *req = ZB_BUF_GET_PARAM(param, zb_nlme_network_formation_request_t);
  zb_ext_pan_id_t ext_pan_id;
  zb_ext_pan_id_t use_ext_pan_id;

  TRACE_MSG(TRACE_ZDO2, "channel mask list:", (FMT__0));
#if defined TRACE_LEVEL && !defined NCP_MODE_HOST
  if (TRACE_ENABLED(TRACE_ZDO2))
  {
    zb_uint_t i;

    for (i = 0; i < ZB_CHANNEL_PAGES_NUM; i++)
    {
      TRACE_MSG(TRACE_ZDO2,
                "page %hd, mask 0x%lx",
                (FMT__H_L,
                 zb_aib_channel_page_list_get_page(i),
                 zb_aib_channel_page_list_get_mask(i)));
    }
  }
#endif

  zb_get_extended_pan_id(ext_pan_id);
  if (!ZB_EXTPANID_IS_ZERO(ext_pan_id) && zb_get_current_channel() != 0)
  {
    TRACE_MSG(TRACE_ERROR, "Pan ID = " TRACE_FORMAT_64, (FMT__A, TRACE_ARG_64(ext_pan_id)));
    TRACE_MSG(TRACE_ERROR, "Start already created network on %hd channel",
              (FMT__H, zb_get_current_channel()));
    /* Call nwk directly or call NCP SoC */
#if defined ZB_DISTRIBUTED_SECURITY_ON
    zb_sync_distributed();
#endif  /* ZB_DISTRIBUTED_SECURITY_ON */
    zb_nwk_cont_without_formation(param);
  }
  else
  {
    /* we must set nwkExtendedPanID to aspUseExtendedPanID if any. Else use our
   * MAC address. */
    /* aps_use_extended_pan_id is assigned by zb_set_extended_pan_id() even at NCP Host */
    zb_get_use_extended_pan_id(use_ext_pan_id);
    if (!ZB_EXTPANID_IS_ZERO(use_ext_pan_id))
    {
      /* For NCP Host: added ext panid into req and remove that
       * assignment from here - better do it in
       * zb_nlme_network_formation_request */
      ZB_IEEE_ADDR_COPY(req->extpanid, use_ext_pan_id);
    }
    else
    {
      ZB_IEEE_ADDR_ZERO(req->extpanid);
    }

    /* 1st start case - do network formation */
    TRACE_MSG(TRACE_INFO1, "First start, perform network formation", (FMT__0));

    /* FIXME: BDB assigns v_scan_channels but we do not use it! */
    /* TODO: in BDB mode use first primary, then secondary channels mask */

    zdo_commissioning_formation_channels_mask(req->scan_channels_list);
    req->scan_duration = ZB_DEFAULT_SCAN_DURATION; /* TODO: configure it somehow? */
    /* timeout for every channel is
       ((1l << duration) + 1) * 15360 / 1000000

       For duration 8 ~ 4s
       For duration 5 ~0.5s
       For duration 2 ~0.08s
       For duration 1 ~0.05s
    */
#if defined ZB_DISTRIBUTED_SECURITY_ON
    zb_sync_distributed();
    if (IS_DISTRIBUTED_SECURITY())
    {
      TRACE_MSG(TRACE_ERROR, "ZR - do distributed netwok formation", (FMT__0));
      req->distributed_network = 1;
      req->distributed_network_address = zb_get_short_address();
    }
#endif  /* ZB_DISTRIBUTED_SECURITY_ON */
    /* Possible NCP call.. */
    ZB_SCHEDULE_CALLBACK(zb_nlme_network_formation_request, param);
  }
}


static void zdo_commissioning_formation_channels_mask(zb_channel_list_t list)
{
#if defined ZB_BDB_MODE
  zb_uint32_t mask;
#endif /* ZB_BDB_MODE */

#if defined ZB_SUB_GHZ_ZB30_SUPPORT
  zb_uint8_t used_page;
#endif /* ZB_SUB_GHZ_ZB30_SUPPORT */

  switch (ZB_COMMISSIONING_TYPE())
  {
#if defined ZB_BDB_MODE
    case ZB_COMMISSIONING_BDB:
      mask = bdb_commissioning_formation_channels_mask();
      zb_channel_list_init(list);

#if defined ZB_SUB_GHZ_ZB30_SUPPORT
      used_page = zb_aib_channel_page_list_get_first_filled_page();
      zb_channel_page_list_set_mask(list, used_page, mask);
#else
      zb_channel_page_list_set_mask(list, ZB_CHANNEL_LIST_PAGE0_IDX, mask);
#endif /* !ZB_SUB_GHZ_ZB30_SUPPORT */
      break;
#endif /* ZB_BDB_MODE */

#ifndef ZB_LITE_BDB_ONLY_COMMISSIONING
    case ZB_COMMISSIONING_CLASSIC:
      /* Classic is actually 2.4-only also, but can just copy vaue from AIB. */
      zb_channel_page_list_copy(list, ZB_AIB().aps_channel_mask_list);
      break;
#ifdef ZB_SE_COMMISSIONING

    case ZB_COMMISSIONING_SE:
      se_commissioning_formation_channels_mask(list);
      break;
#endif /* ZB_SE_COMMISSIONING */
#endif /* !ZB_LITE_BDB_ONLY_COMMISSIONING */

    default:
      break;
  }
}


#ifdef ZB_BDB_MODE
static void bdb_formation(zb_uint8_t param)
{
  TRACE_MSG(TRACE_ZDO3, "bdb_formation param %hd", (FMT__H, param));

  /* See Figure 5 - Network formation procedure */

  if (ZB_BDB().v_do_primary_scan == ZB_BDB_JOIN_MACHINE_PRIMARY_SCAN
      && ZB_BDB().bdb_primary_channel_set != 0)
  {
    ZB_BDB().v_do_primary_scan = ZB_BDB_JOIN_MACHINE_SECONDARY_SCAN_START;
    ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_IN_PROGRESS;
    ZB_BDB().v_scan_channels = ZB_BDB().bdb_primary_channel_set;
    TRACE_MSG(TRACE_ZDO1, "Doing formation channel mask 0x%lx", (FMT__L, ZB_BDB().v_scan_channels));
    ZB_SCHEDULE_CALLBACK(zdo_start_formation, param);
  }
  else if ((ZB_BDB().v_do_primary_scan == ZB_BDB_JOIN_MACHINE_SECONDARY_SCAN_START
            && ZB_BDB().bdb_secondary_channel_set != 0)
           ||
           (ZB_BDB().v_do_primary_scan == ZB_BDB_JOIN_MACHINE_PRIMARY_SCAN
            && ZB_BDB().bdb_secondary_channel_set != 0
            && ZB_BDB().bdb_primary_channel_set == 0))
  {
    ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_IN_PROGRESS;
    ZB_BDB().v_do_primary_scan = ZB_BDB_JOIN_MACHINE_SECONDARY_SCAN_DONE;
    ZB_BDB().v_scan_channels = ZB_BDB().bdb_secondary_channel_set;
    TRACE_MSG(TRACE_ZDO1, "Doing formation channel mask 0x%lx", (FMT__L, ZB_BDB().v_scan_channels));
    ZB_SCHEDULE_CALLBACK(zdo_start_formation, param);
  }
  else
  {
    ZB_BDB().bdb_commissioning_status = ZB_BDB_STATUS_FORMATION_FAILURE;
    TRACE_MSG(TRACE_ZDO1, "BDB: Formation failed run next bdb machine step", (FMT__0));
    ZB_BDB().v_do_primary_scan = ZB_BDB_JOIN_MACHINE_PRIMARY_SCAN;
    ZB_BDB().bdb_commissioning_step <<= 1;
    ZB_SCHEDULE_CALLBACK(bdb_commissioning_machine, param);
  }
}

static zb_uint32_t bdb_commissioning_formation_channels_mask(void)
{
  return ZB_BDB().v_scan_channels;
}

#endif  /* ZB_BDB_MODE */

#endif  /* !defined NCP_MODE || defined NCP_MODE_HOST */



void zb_enable_distributed(void)
{
  zdo_formation_force_link();
}

#endif  /* ZB_FORMATION */

/*! @} */
